// OptionList.cpp: implementation of the OptionList class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "DLaunch.h"
#include "OptionList.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

OptionList::OptionList()
{
	memset(&root, 0, sizeof(root));
	root.next = &root;
	root.prev = &root;
	numOptions = 0;
}

OptionList::~OptionList()
{
	Clear();
}

int OptionList::Number()
{
	return numOptions;
}

option_t *OptionList::Add()
{
	option_t *newone = new option_t;
	memset(newone, 0, sizeof(*newone));
	// Add to the end.
	newone->prev = root.prev;
	newone->next = &root;
	root.prev->next = newone;
	root.prev = newone;
	numOptions++;
	return newone;
}

int ReadTrimmed(CStdioFile &file, CString &str)
{
	int ret = file.ReadString(str);
	str.TrimLeft();
	str.TrimRight();
	return ret;
}

// A recursive reader.
void OptionList::ReadOptionFile(CStdioFile &file, option_t *parentPtr)
{
	CString		buff;
	option_t	*opt = NULL;	// The current option.

	while(ReadTrimmed(file, buff))
	{
		if(buff == "option")
		{
			// A new option begins.
			opt = Add();
			opt->parent = parentPtr;
			// The next line is the name of the option.
			ReadTrimmed(file, buff);
			strcpy(opt->text, buff);
			continue;
		}

		// Non-option data.
		if(buff == "title")
		{
			ReadTrimmed(file, title);
			continue;
		}
		if(buff == "logo")
		{
			ReadTrimmed(file, logoBmp);
			continue;
		}
		if(buff == "editfile")
		{
			ReadTrimmed(file, buff);
			editFiles.AddTail(buff);
		}

		// The following tokens are only valid inside option definitions.
		if(!opt) continue;

		if(buff == "begin")
		{
			ReadOptionFile(file, opt);
		}
		else if(buff == "end") break;
		else if(buff == "integer")
		{
			opt->flags |= OPF_INT_VALUE;
			ReadTrimmed(file, buff);
			opt->value = (float) atoi(buff);
		}
		else if(buff == "string")
		{
			opt->flags |= OPF_STR_VALUE;
			ReadTrimmed(file, buff);
			strcpy(opt->strValue, buff);
		}
		else if(buff == "cmdline")
		{
			ReadTrimmed(file, buff);
			strcpy(opt->option, buff);
		}
		else if(buff == "help")
		{
			ReadTrimmed(file, buff);
			strcpy(opt->description, buff);
		}
		else if(buff == "enabled") opt->flags |= OPF_ENABLED;
		else if(buff == "disabled") opt->flags |= OPF_DISABLED;
		else if(buff == "file") opt->flags |= OPF_FILE;
		else if(buff == "nocmdline") opt->flags |= OPF_NOT_IN_CMDLINE;
		else if(buff == "executable") opt->flags |= OPF_EXECUTABLE;
		else if(buff == "workingdir") opt->flags |= OPF_WORKING_DIR;
		else if(buff == "list") opt->flags |= OPF_LIST;
		else if(buff == "integer choice")
		{
			opt->flags |= OPF_CHOICE | OPF_INT_VALUE;
			ReadTrimmed(file, buff);
			opt->value = (float) atoi(buff); // Current selection (choice index).
			ReadTrimmed(file, buff);
			opt->numChoices = atoi(buff); // Number of choices.
			opt->choices = new choice_t[opt->numChoices];
			for(int i=0; i<opt->numChoices; i++)
			{
				choice_t *ch = opt->choices + i;
				ReadTrimmed(file, buff);
				ch->value = (float) atoi(buff);
				ReadTrimmed(file, buff);
				strcpy(ch->text, buff);
			}
		}
		else if(buff == "string choice")
		{
			opt->flags |= OPF_CHOICE | OPF_STR_VALUE;
			ReadTrimmed(file, buff);
			opt->value = (float) atoi(buff); // Current selection (choice index).
			ReadTrimmed(file, buff);
			opt->numChoices = atoi(buff); // Number of choices.
			opt->choices = new choice_t[opt->numChoices];
			for(int i=0; i<opt->numChoices; i++)
			{
				choice_t *ch = opt->choices + i;
				ReadTrimmed(file, buff);
				strcpy(ch->strValue, buff);
				ReadTrimmed(file, buff);
				strcpy(ch->text, buff);
			}
		}
	}
}

void writeStr(CStdioFile &file, const CString &str)
{
	file.WriteString(str);
}

// This is called recursively.
void OptionList::WriteOptionFile(CStdioFile &file, option_t *parent, int level)
{
	CString tab, str;
	int i;

	// Construct the indentation.
	for(i=0; i<level*4; i++) tab += " ";
	// Check everything.
	for(option_t *iter=root.next; iter != &root; iter = iter->next)
	{
		if(iter->parent != parent) continue; // Not to be written now.
		
		// Write this option.
		writeStr(file, "\n" + tab + "option\n");
		writeStr(file, tab + iter->text + "\n");

		if(strcmp(iter->option, "")) 
			writeStr(file, tab + "cmdline\n" + tab + iter->option + "\n");

		if(iter->flags & OPF_CHOICE)
		{
			writeStr(file, tab + (iter->flags & OPF_INT_VALUE? "integer choice" 
				: iter->flags & OPF_STR_VALUE? "string choice" 
				: "choice") + "\n");
			str.Format("%d\n", (int) iter->value);
			writeStr(file, tab + str);
			str.Format("%d\n", iter->numChoices);
			writeStr(file, tab + str);
			for(i=0; i<iter->numChoices; i++)
			{
				if(iter->flags & OPF_INT_VALUE)
					str.Format("%d\n", (int) iter->choices[i].value);
				else
					str.Format("%s\n", iter->choices[i].strValue);
				writeStr(file, tab + str + tab + iter->choices[i].text + "\n");		
			}
		}
		else
		{
			if(iter->flags & OPF_INT_VALUE)
			{
				str.Format("%d\n", (int) iter->value);
				writeStr(file, tab + "integer\n" + tab + str);
			}
			if(iter->flags & OPF_STR_VALUE)
			{
				str.Format("%s\n", iter->strValue);
				writeStr(file, tab + "string\n" + tab + str);
			}
		}
		if(iter->flags & OPF_ENABLED) writeStr(file, tab + "enabled\n");
		if(iter->flags & OPF_DISABLED) writeStr(file, tab + "disabled\n");
		if(iter->flags & OPF_FILE) writeStr(file, tab + "file\n");
		if(iter->flags & OPF_LIST) writeStr(file, tab + "list\n");
		if(iter->flags & OPF_NOT_IN_CMDLINE) writeStr(file, tab + "nocmdline\n");
		if(iter->flags & OPF_EXECUTABLE) writeStr(file, tab + "executable\n");
		if(iter->flags & OPF_WORKING_DIR) writeStr(file, tab + "workingdir\n");
		if(strcmp(iter->description, ""))
			writeStr(file, tab + "help\n" + tab + iter->description + "\n");
		
		// Check the children.
		if(HasChildren(iter))
		{
			writeStr(file, tab + "begin\n");
			WriteOptionFile(file, iter, level+1);
			writeStr(file, tab + "end\n");
		}
	}
}

int OptionList::SaveTo(const CString &filename)
{
	CStdioFile	file;
	
	if(!file.Open(filename, CFile::modeCreate | CFile::modeWrite))
		return false;
	
	if(title != "")	writeStr(file, "title\n" + title + "\n");
	if(logoBmp != "") writeStr(file, "logo\n" + logoBmp + "\n");

	for(POSITION pos = editFiles.GetHeadPosition(); pos;)
		writeStr(file, "editfile\n" + editFiles.GetNext(pos) + "\n");

	WriteOptionFile(file);
	// Success!
	return true;
}

int OptionList::LoadFrom(const CString &filename)
{
	CStdioFile	file;
	
	// First clear all old data.
	Clear();
	// Open the new file.
	if(!file.Open(filename, CFile::modeRead))
	{
		return 0;
	}
	ReadOptionFile(file);
	return numOptions;
}

void OptionList::Clear()
{
	for(option_t *iter = root.next; iter != &root;)
	{
		option_t *the_next_one = iter->next;
		if(iter->choices) delete [] iter->choices;
		delete iter;
		iter = the_next_one;
	}
	root.next = &root;
	root.prev = &root;
	numOptions = 0;
	logoBmp = "";
	title = "";
	editFiles.RemoveAll();
}

option_t *OptionList::GetMask(int flags)
{
	for(option_t *iter=root.next; iter != &root; iter=iter->next)
		if((iter->flags & flags) == flags)
			return iter;
	return NULL;
}

void OptionList::DeleteChildrenOf(option_t *par)
{
	for(option_t *iter=root.next; iter != &root; iter = iter->next)
	{
		if(iter->parent == par)
		{
			// Unlink and destroy.
			iter->prev->next = iter->next;
			iter->next->prev = iter->prev;
			option_t *temp = iter->prev;
			delete iter;
			iter = temp;
			numOptions--;
		}
	}
}

option_t *OptionList::Get(int idx)
{
	int i;
	option_t *iter;
	for(i=0, iter=root.next; iter != &root; i++, iter=iter->next)
		if(i == idx) return iter;
	return NULL;
}

int OptionList::GetIndex(option_t *ptr)
{
	int c;
	option_t *iter;
	for(c=0, iter=root.next; iter != &root; c++, iter=iter->next)
		if(iter == ptr) return c;
	// Not found.
	return -1;
}

bool OptionList::HasChildren(option_t *opt, int flags)
{
	for(option_t *iter=root.next; iter != &root; iter=iter->next)
	{
		if(iter->parent != opt) continue;
		if((iter->flags & flags) == flags) return true;
	}
	return false;
}

// src isn't modified during the operation, only *this is.
void OptionList::UpdateWith(OptionList &src)
{
	OptionList result;	// The end result that will replace *this.
	option_t *i;

	// We are basically updating the new data using the old.
	result = src;

	for(i=result.root.next; i != &result.root; i=i->next)
	{
		// Is there an equivalent option for this one?
		option_t *eq = FindEquivalent(result.GetOptionPath(i));
		if(!eq) continue;	// Nothing for this one.
		// All right, a potential match. The types must also be the same.
		if((i->flags & OPF_TYPE_FLAGS) != (eq->flags & OPF_TYPE_FLAGS)) continue;
		// A complete match! Copy over all the data.
		i->flags &= ~(OPF_ENABLED | OPF_DISABLED);
		i->flags |= eq->flags & (OPF_ENABLED | OPF_DISABLED);
		i->value = eq->value;
		strcpy(i->strValue, eq->strValue);
		if(i->numChoices)
		{
			i->value = 0;
			// Try to find a matching choice in the new list.
			for(int a=0; a<i->numChoices; a++)
			{
				if(eq->choices[(int)eq->value].value == i->choices[a].value)
				{
					i->value = (float) a;
					break;
				}
			}
		}
		if(i->flags & OPF_LIST)
		{
			result.DeleteChildrenOf(i);
			for(option_t *ptr=root.next; ptr != &root; ptr=ptr->next)
			{
				if(ptr->parent != eq) continue;
				// Add this one.
				option_t *New = result.Add();
				New->SafeCopy(*ptr);
				New->parent = i;
			}
		}
	}

	// Now we're done, replace *this with the new data.
	*this = result;
}

OptionList& OptionList::operator =(OptionList &other)
{
	option_t *opt, *i;

	Clear();
	title = other.title;
	logoBmp = other.logoBmp;
	for(POSITION pos = other.editFiles.GetHeadPosition(); pos;)
		editFiles.AddTail(other.editFiles.GetNext(pos));
	for(i=other.root.next; i != &other.root; i=i->next)
	{
		// Create a copy of each option. The parent pointers will 
		// be updated separately.
		opt = Add();
		opt->SafeCopy(*i);
	}
	for(i=root.next; i != &root; i=i->next)
	{
		if(i->parent) i->parent = Get(other.GetIndex(i->parent));
	}	
	return *this;
}

CString OptionList::GetOptionPath(option_t *opt)
{
	CString path = opt->text;

	while(opt->parent)
	{
		opt = opt->parent;
		path = CString(opt->text) + "\\" + path;
	}
	return path;
}

// Tries to find a path that matches the given one.
option_t *OptionList::FindEquivalent(const CString &path)
{
	for(option_t *i=root.next; i != &root; i=i->next)
	{
		if(!GetOptionPath(i).CompareNoCase(path))
			return i;
	}
	return NULL;
}

void option_t::SafeCopy(option_t &i)
{
	parent = i.parent;
	strcpy(text, i.text);
	strcpy(option, i.option);
	flags = i.flags;
	value = i.value;
	strcpy(strValue, i.strValue);
	strcpy(description, i.description);
	numChoices = i.numChoices;
	if(numChoices)
	{
		choices = new choice_t[numChoices];
		memcpy(choices, i.choices, sizeof(choice_t) * numChoices);
	}
}
