// OptionList.h: interface for the OptionList class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_OPTIONLIST_H__3014D38F_0889_11D4_9FA8_A87B741ED15E__INCLUDED_)
#define AFX_OPTIONLIST_H__3014D38F_0889_11D4_9FA8_A87B741ED15E__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


#define OPF_ENABLED			0x1		// Used.
#define OPF_DISABLED		0x2		// Not in use.
#define OPF_INT_VALUE		0x4		// Has a value of type integer.
#define OPF_FLOAT_VALUE		0x8		
#define OPF_STR_VALUE		0x10
#define OPF_FILE			0x20	// When changing, browse for a file.
#define OPF_CHOICE			0x40	// The choices define a valid list.
#define OPF_LIST			0x80	// A user-editable list w/items that can be enabled.
#define OPF_NOT_IN_CMDLINE	0x100
#define OPF_EXECUTABLE		0x10000	// Doomsday.exe, for instance.
#define OPF_WORKING_DIR		0x20000	// CD to the dir specified by this filename before launching.

#define OPF_TYPE_FLAGS		0x300dc

struct choice_t
{
	char		text[60];
	float		value;
	char		strValue[256];
};

struct option_t
{
	option_t	*next, *prev;		// Pointer to the next and previous options in the list.
	option_t	*parent;			// Can be NULL.
	HTREEITEM	handle;
	char		text[256];
	char		option[256];		// The actual option (like -allowsoftware).
	int			flags;		
	float		value;
	char		strValue[256];
	char		description[256];	// A short info text.
	choice_t	*choices;
	int			numChoices;
public:
	void SafeCopy(option_t& other);
};

class OptionList  
{
public:
	OptionList();
	virtual ~OptionList();

	OptionList& operator = (OptionList& other);

	option_t *FindEquivalent(const CString &path);
	CString GetOptionPath(option_t *opt);
	void UpdateWith(OptionList &src);
	bool HasChildren(option_t *opt, int flags=0);
	option_t *Add();
	void DeleteIndex(int idx);
	void DeleteChildrenOf(option_t *opt);
	int GetIndex(option_t *ptr);
	void Clear();
	int LoadFrom(const CString &filename);
	int SaveTo(const CString &filename);
	int Number();
	option_t *Get(int idx);
	option_t *GetMask(int flags);
	void ReadOptionFile(CStdioFile &file, option_t *parentPtr=NULL);
	void WriteOptionFile(CStdioFile &file, option_t *parent=NULL, int level=0);

	// Public data.
	CString		title;	// For the launcher window.
	CString		logoBmp;
	CStringList	editFiles;

protected:
	option_t	root;
	int			numOptions;
};

#endif // !defined(AFX_OPTIONLIST_H__3014D38F_0889_11D4_9FA8_A87B741ED15E__INCLUDED_)
