// DOOM console stuff.

#include <stdlib.h>

#include "doomdef.h"
#include "doomstat.h"
#include "d_main.h"
#include "d_config.h"
#include "d_netjd.h"
#include "g_game.h"
#include "r_local.h"
#include "w_wad.h"
#include "s_sound.h"
#include "sounds.h"
#include "hu_stuff.h"
#include "m_menu.h"

// Macros -----------------------------------------------------------------

#define DEFCC(name)		int name(int argc, char **argv)

// External Functions -----------------------------------------------------

DEFCC(CCmdCycleSpy);
DEFCC(CCmdMenuAction);
DEFCC(CCmdCD);
DEFCC(CCmdMidi);
DEFCC(CCmdCrosshair);
DEFCC(CCmdCheat);
DEFCC(CCmdCheatGod);
DEFCC(CCmdCheatNoClip);
DEFCC(CCmdCheatWarp);
DEFCC(CCmdCheatReveal);
DEFCC(CCmdCheatGive);
DEFCC(CCmdCheatMassacre);
DEFCC(CCmdBeginChat);

// Public Functions -------------------------------------------------------

DEFCC(CCmdScreenShot);
DEFCC(CCmdViewSize);
DEFCC(CCmdPlayDemo);
DEFCC(CCmdRecordDemo);
DEFCC(CCmdStopDemo);
DEFCC(CCmdPause);
DEFCC(CCmdDoomFont);

// External Data ----------------------------------------------------------

extern char *chat_macros[];
extern int showMessages;

// Public Data ------------------------------------------------------------

int consoleFlat = 10;
float consoleZoom = 1;

// Console variables.
cvar_t gameCVars[] =
{
	"i_mouseSensiX",	CVF_NO_MAX,	CVT_INT,	&cfg.mouseSensiX,	0, 25, "Mouse X axis sensitivity.",
	"i_mouseSensiY",	CVF_NO_MAX,	CVT_INT,	&cfg.mouseSensiY,	0, 25, "Mouse Y axis sensitivity.",
	"i_jlookInvY",		0,			CVT_INT,	&cfg.jlookInverseY, 0, 1,	"1=Inverse joystick look Y axis.",
	"i_mlookInvY",		0,			CVT_INT,	&cfg.mlookInverseY, 0, 1,	"1=Inverse mouse look Y axis.",
	"fps",			CVF_NO_ARCHIVE,	CVT_INT,	&cfg.showFPS,		0, 1,	"1=Show the frames per second counter.",
	"echomsg",			0,			CVT_BYTE,	&cfg.echoMsg,		0, 1,	"1=Echo all messages to the console.",

	"lookspeed",		0,			CVT_INT,	&cfg.lookSpeed,	1, 5,	"The speed of looking up/down.",
	"dclickuse",		0,			CVT_INT,	&cfg.dclickuse,	0, 1,	"1=Doubleclick forward/strafe equals use key.",
//	"quakefly",			0,			CVT_INT,	&quakeFly,		0, 1,	"1=Use Quake's flight model (fly where you're looking).",
	"bgflat",		CVF_NO_MAX,		CVT_INT,	&consoleFlat,	0, 0,	"The number of the flat to use for the console background.",
	"bgzoom",			0,			CVT_FLOAT,	&consoleZoom,	0.1f, 100.0f, "Zoom factor for the console background.",
	"povlook",			0,			CVT_BYTE,	&cfg.povLookAround,	0, 1,	"1=Look around using the POV hat.",

	"i_mlook",			0,			CVT_INT,	&cfg.usemlook,	0, 1,	"1=Mouse look active.",
	"i_jlook",			0,			CVT_INT,	&cfg.usejlook,	0, 1,	"1=Joystick look active.",
	"alwaysrun",		0,			CVT_INT,	&cfg.alwaysRun,	0, 1,	"1=Always run.",
	"lookspring",		0,			CVT_INT,	&cfg.lookSpring,0, 1,	"1=Lookspring active.",
	"noautoaim",		0,			CVT_INT,	&cfg.noAutoAim,	0, 1,	"1=Autoaiming disabled.",
	"d_viewsize",	CVF_PROTECTED,	CVT_INT,	&screenblocks,	3, 11,	"View window size (3-11).",
	"d_sbsize",		CVF_PROTECTED,	CVT_INT,	&cfg.sbarscale,	1, 20,	"Status bar size (1-20).",

	"flash_R",			0,			CVT_FLOAT,	&cfg.flashcolor[0],			0, 1,	"Menu selection flash color, red component.",
	"flash_G",			0,			CVT_FLOAT,	&cfg.flashcolor[1],			0, 1,	"Menu selection flash color, green component.",
	"flash_B",			0,			CVT_FLOAT,	&cfg.flashcolor[2],			0, 1,	"Menu selection flash color, blue component.",
	"flash_speed",		0,			CVT_INT,	&cfg.flashspeed,			0, 50,	"Menu selection flash speed.",
	"turningSkull",		0,			CVT_BYTE,	&cfg.turningSkull,			0, 1,	"1=Menu skull turns at slider items.",
	"hud_health",		0,			CVT_BYTE,	&cfg.hudShown[HUD_HEALTH],	0, 1,	"1=Show health in HUD.",
	"hud_armor",		0,			CVT_BYTE,	&cfg.hudShown[HUD_ARMOR],	0, 1,	"1=Show armor in HUD.",
	"hud_ammo",			0,			CVT_BYTE,	&cfg.hudShown[HUD_AMMO],	0, 1,	"1=Show ammo in HUD.",
	"hud_keys",			0,			CVT_BYTE,	&cfg.hudShown[HUD_KEYS],	0, 1,	"1=Show keys in HUD.",
	"hud_frags",		0,			CVT_BYTE,	&cfg.hudShown[HUD_FRAGS],	0, 1,	"1=Show deathmatch frags in HUD.",
	"hud_scale",		0,			CVT_FLOAT,	&cfg.hudScale,				0.1f, 10, "Scaling for HUD info.",
	"hud_R",			0,			CVT_FLOAT,	&cfg.hudColor[0],			0, 1,	"HUD info color.",
	"hud_G",			0,			CVT_FLOAT,	&cfg.hudColor[1],			0, 1,	"HUD info color.",
	"hud_B",			0,			CVT_FLOAT,	&cfg.hudColor[2],			0, 1,	"HUD info color.",

	"xhair", CVF_NO_MAX|CVF_PROTECTED, CVT_INT,	&cfg.xhair,			0, 0,	"The current crosshair.",
	"xhairR",			0,			CVT_BYTE,	&cfg.xhairColor[0], 0, 255,	"Red crosshair color component.",
	"xhairG",			0,			CVT_BYTE,	&cfg.xhairColor[1], 0, 255, "Green crosshair color component.",
	"xhairB",			0,			CVT_BYTE,	&cfg.xhairColor[2], 0, 255, "Blue crosshair color component.",
	"xhairSize",	CVF_NO_MAX,		CVT_INT,	&cfg.xhairSize,		0, 0,	"Crosshair size: 1=Normal.",

	//"s_channels",	CVF_PROTECTED,	CVT_INT,	&snd_Channels,	1, 16,	"The number of sounds that can be played simultaneously.",
//	"sounddebug",	CVF_NO_ARCHIVE,	CVT_INT,	&DebugSound,	0, 1,	"1=Display sound debug information.",
	
	"s_3D",				0,			CVT_BYTE,	&cfg.snd_3D,			0, 1,	"1=Play sounds in 3D.",
	"s_ReverbVol",		0,			CVT_BYTE,	&cfg.snd_ReverbVolume,	0, 100, "General reverb strength (0-100).",
	"reverbdebug",	CVF_NO_ARCHIVE,	CVT_BYTE,	&cfg.reverbDebug,		0, 1,	"1=Reverb debug information in the console.",

	"messages",			0,			CVT_INT,	&showMessages,	0, 1,	"1=Show messages.",

	"chatmacro0",		0,			CVT_CHARPTR, &chat_macros[0], 0, 0, "Chat macro 1.",
	"chatmacro1",		0,			CVT_CHARPTR, &chat_macros[1], 0, 0, "Chat macro 2.",
	"chatmacro2",		0,			CVT_CHARPTR, &chat_macros[2], 0, 0, "Chat macro 3.",
	"chatmacro3",		0,			CVT_CHARPTR, &chat_macros[3], 0, 0, "Chat macro 4.",
	"chatmacro4",		0,			CVT_CHARPTR, &chat_macros[4], 0, 0, "Chat macro 5.",
	"chatmacro5",		0,			CVT_CHARPTR, &chat_macros[5], 0, 0, "Chat macro 6.",
	"chatmacro6",		0,			CVT_CHARPTR, &chat_macros[6], 0, 0, "Chat macro 7.",
	"chatmacro7",		0,			CVT_CHARPTR, &chat_macros[7], 0, 0, "Chat macro 8.",
	"chatmacro8",		0,			CVT_CHARPTR, &chat_macros[8], 0, 0, "Chat macro 9.",
	"chatmacro9",		0,			CVT_CHARPTR, &chat_macros[9], 0, 0, "Chat macro 10.",

	// Game settings for servers.
	"nomonsters",		0,			CVT_BYTE,	&cfg.netNomonsters,	0, 1,	"1=No monsters.",
	"respawn",			0,			CVT_BYTE,	&cfg.netRespawn,	0, 1,	"1= -respawn was used.",
	"n_skill",			0,			CVT_BYTE,	&cfg.netSkill,		0, 4,	"Skill level in multiplayer games.",
	"n_map",			0,			CVT_BYTE,	&cfg.netMap,		1, 31,	"Map to use in multiplayer games.",
	"n_episode",		0,			CVT_BYTE,	&cfg.netEpisode,	1, 6,	"Episode to use in multiplayer games.",
	"n_slot",			0,			CVT_BYTE,	&cfg.netSlot,		0, 6,	"The savegame slot to start from. 0=none.",
	"n_jump",			0,			CVT_BYTE,	&cfg.netJumping,	0, 1,	"1=Allow jumping in multiplayer games.",
	"deathmatch",		0,			CVT_BYTE,	&cfg.netDeathmatch,	0, 2,	"Start multiplayers games as deathmatch.",

	// Player data.
	"n_color",			0,			CVT_BYTE,	&cfg.netColor,		0, 3,	"Player color: 0=green, 1=gray, 2=brown, 3=red.",

/*	"ringfilter",		0,			CVT_INT,	&ringFilter,	1, 2,	"Ring effect filter. 1=Brownish, 2=Blue.",*/
	//"testcvar",	CVF_NO_ARCHIVE|CVF_NO_MIN|CVF_NO_MAX, CVT_INT,	&testcvar,		0, 0,	"",
	"allowjump",		0,			CVT_INT,	&cfg.jumpEnabled,	0, 1,	"1=Allow jumping.",
	"fastMonsters",		0,			CVT_BYTE,	&fastparm,			0, 1,	"1=Fast monsters in non-demo single player.",
	"zclip",			0,			CVT_BYTE,	&cfg.moveCheckZ,	0, 1,	"1=Allow mobjs to move under/over each other.",
	"jumpPower",		0,			CVT_FLOAT,	&cfg.jumpPower,		0, 100,	"Jump power.",
	"autoswitch",		0,			CVT_BYTE,	&cfg.weaponAutoSwitch, 0, 1, "1=Change weapon automatically when picking one up.",
	"secretMsg",		0,			CVT_INT,	&cfg.secretMsg,		0, 1,	"1=Announce the discovery of secret areas.",

	NULL
};

ccmd_t gameCCmds[] =
{
	"spy",			CCmdCycleSpy,		"Spy mode: cycle player views in co-op.",
	"screenshot",	CCmdScreenShot,		"Takes a screenshot. Saved to DOOMnn.TGA.",
	"viewsize",		CCmdViewSize,		"View size adjustment.",
	"sbsize",		CCmdViewSize,		"Status bar size adjustment.",
	"playdemo",		CCmdPlayDemo,		"Play a demo.",
	"recorddemo",	CCmdRecordDemo,		"Record a demo in the current level or in the specified one.",
	"stopdemo",		CCmdStopDemo,		"Stop demo playback/recording.",
	"pause",		CCmdPause,			"Pause the game.",
	"cd",			CCmdCD,				"CD control.",
	"midi",			CCmdMidi,			"Music control.",
	"crosshair",	CCmdCrosshair,		"Crosshair setup.",
	"cheat",		CCmdCheat,			"Issue a cheat code using the original Doom cheats.",
	"god",			CCmdCheatGod,		"God mode.",
	"noclip",		CCmdCheatNoClip,	"No movement clipping (walk through walls).",
	"warp",			CCmdCheatWarp,		"Warp to another map.",
	"reveal",		CCmdCheatReveal,	"Map cheat.",
	"give",			CCmdCheatGive,		"Gives you weapons, ammo, power-ups, etc.",
	"kill",			CCmdCheatMassacre,	"Kill all the monsters on the level.",
	"doomfont",		CCmdDoomFont,		"Use the Doom font in the console.",
	"beginChat",	CCmdBeginChat,		"Begin chat mode.",

	// Menu Actions
	"HelpScreen",	CCmdMenuAction,		"Show the Help screens.",
	"SaveGame",		CCmdMenuAction,		"Open the save game menu.",
	"LoadGame",		CCmdMenuAction,		"Open the load game menu.",
	"SoundMenu",	CCmdMenuAction,		"Open the sound settings menu.",
	"QuickSave",	CCmdMenuAction,		"Quicksave the game.",
	"EndGame",		CCmdMenuAction,		"End the game.",
	"ToggleMsgs",	CCmdMenuAction,		"Messages on/off.",
	"QuickLoad",	CCmdMenuAction,		"Load the quicksaved game.",
	"quit",			CCmdMenuAction,		"Quit the game and return to the OS.",
	"ToggleGamma",	CCmdMenuAction,		"Cycle gamma correction levels.",
	NULL
};

// Private Data -----------------------------------------------------------

static char playDemoName[10];

// Code -------------------------------------------------------------------

void D_ConsoleRegistration()
{
	int		i;
	
	for(i=0; gameCVars[i].name; i++) gi.AddVariable(gameCVars+i);
	for(i=0; gameCCmds[i].name; i++) gi.AddCommand(gameCCmds+i);
	for(i=0; netCCmds[i].name; i++) gi.AddCommand(netCCmds+i);
}

void D_ConsoleBg(int *width, int *height)
{
	extern int consoleFlat;
	extern float consoleZoom;
	
	gi.GL_SetFlat(consoleFlat);
	*width = 64*consoleZoom;
	*height = 64*consoleZoom;
}

int CCmdScreenShot(int argc, char **argv)
{
	G_ScreenShot();
	return true;
}

int CCmdViewSize(int argc, char **argv)
{
	int	min=3, max=11, *val = &screenblocks;

	if(argc != 2)
	{
		gi.conprintf( "Usage: %s (size)\n", argv[0]);
		gi.conprintf( "Size can be: +, -, (num).\n");
		return true;
	}
	if(!stricmp(argv[0], "sbsize"))
	{
		min = 1;
		max = 20;
		val = &cfg.sbarscale;
	}
	if(!stricmp(argv[1], "+"))
	{
		(*val)++;
		S_StartSound(NULL, sfx_stnmov);
	}
	else if(!stricmp(argv[1], "-"))
	{
		(*val)--;
		S_StartSound(NULL, sfx_stnmov);
	}
	else
		*val = strtol(argv[1], NULL, 0);

	if(*val < min) *val = min;
	if(*val > max) *val = max;

	// Update the view size if necessary.
	R_SetViewSize(screenblocks, 0);
	return true;
}

int CCmdPlayDemo(int argc, char **argv)
{	
	extern int demosequence;

	if(argc != 2)
	{
		gi.conprintf( "Usage: playdemo (name)\n");
		return true;
	}
	if(W_CheckNumForName(argv[1]) == -1)
	{
		gi.conprintf( "There is a no lump named '%s'.\n", argv[1]);
		return false;
	}
	demosequence = -1;
	strcpy(playDemoName, argv[1]);
	G_DeferedPlayDemo(playDemoName);
	return true;
}

int CCmdRecordDemo(int argc, char **argv)
{
	int episode = gameepisode;
	int	map = gamemap;

	if(demorecording)
	{
		gi.conprintf( "A demo is already being recorded!\n");
		return false;
	}
	if(argc == 1 || argc > 4)
	{
		gi.conprintf( "Usage: recorddemo (name) (episode) (map)\n");
		gi.conprintf( "The demo will be saved in a file named (name).lmp.\n");
		gi.conprintf( "If you don't specify a map the current one is used.\n");
		return true;
	}
	if(argc >= 3) episode = atoi(argv[2]);
	if(argc == 4) map = atoi(argv[3]);
	G_InitNew(gameskill, episode, map);	
	G_RecordDemo(argv[1]);
	G_BeginRecording();
	return true;
}

int CCmdStopDemo(int argc, char **argv)
{
	if(!demoplayback && !demorecording)
	{
		gi.conprintf( "No demo being played or recorded.\n");
		return false;
	}
	G_CheckDemoStatus();
	D_StartTitle();	
	return true;
}

int CCmdPause(int argc, char **argv)
{
	extern boolean sendpause;

	if(!menuactive) sendpause = true;
	return true;
}

void ConTextOut(char *text, int x, int y)
{
	M_WriteText2(x, y, text, hu_font_a, -1, -1, -1);
}

int ConTextWidth(char *text)
{
	return M_StringWidth(text, hu_font_a);
}

void ConTextFilter(char *text)
{
	strupr(text);
}

int CCmdDoomFont(int argc, char **argv)
{
	ddfont_t	cfont;

	cfont.flags = DDFONT_WHITE;
	cfont.height = 8;
	cfont.sizeX = 1.5f;
	cfont.sizeY = 2;
	cfont.TextOut = ConTextOut;
	cfont.Width = ConTextWidth;
	cfont.Filter = ConTextFilter;
	gi.SetConsoleFont(&cfont);
	return true;
}
