#include "doomdef.h"
#include "doomstat.h"
#include "st_stuff.h"
#include "wi_stuff.h"
#include "hu_stuff.h"
#include "f_finale.h"
#include "am_map.h"
#include "w_wad.h"
#include "r_local.h"
#include "d_config.h"
#include "x_hair.h"

// Macros -------------------------------------------------------------------

#define viewheight	gi.Get(DD_VIEWWINDOW_HEIGHT)

// External Data ------------------------------------------------------------

extern char MobjLightOffsets[NUMMOBJTYPES];

// Code ---------------------------------------------------------------------

//
// R_InitTranslationTables
// Creates the translation tables to map
//  the green color ramp to gray, brown, red.
// Assumes a given structure of the PLAYPAL.
// Could be read from a lump instead.
//
void R_Init (void)
{
	byte	*translationtables = (byte*) gi.Get(DD_TRANSLATIONTABLES_ADDRESS);
    int		i;
  
    // translate just the 16 green colors
    for (i=0 ; i<256 ; i++)
    {
		if (i >= 0x70 && i<= 0x7f)
		{
			// map green ramp to gray, brown, red
			translationtables[i] = 0x60 + (i&0xf);
			translationtables[i+256] = 0x40 + (i&0xf);
			translationtables[i+512] = 0x20 + (i&0xf);
		}
		else
		{
			// Keep all other colors as is.
			translationtables[i] = translationtables[i+256] 
				= translationtables[i+512] = i;
		}
    }
}


//
// D_PageDrawer
//
void D_PageDrawer (void)
{
	extern char *pagename;
	gi.GL_SetNoTexture();
	gi.GL_DrawPatch(0, 0, W_GetNumForName(pagename));
}


void R_DrawSpecialFilter()
{
	player_t *player = &players[displayplayer];

	if(player->powers[pw_invulnerability])
	{
		float max = 30;
		float str, r, g, b;
		int t = player->powers[pw_invulnerability];
		if(t < max)
			str = t/max;
		else if(t < 4*32 && !(t&8))
			str = .7f;
		else if(t > INVULNTICS-max)
			str = (INVULNTICS-t)/max;
		else
			str = 1; // Full inversion.
		// Draw an inversing filter. 
		gl.Disable(DGL_TEXTURING);
		gl.Func(DGL_BLENDING, DGL_ONE_MINUS_DST_COLOR, DGL_ONE_MINUS_SRC_COLOR);
		r = str*2;
		g = str*2 - .4;
		b = str*2 - .8;
		if(r < 0) r = 0;
		if(r > 1) r = 1;
		if(g < 0) g = 0;
		if(g > 1) g = 1;
		if(b < 0) b = 0;
		if(b > 1) b = 1;
		gi.GL_DrawRect(0, 0, 320, 200, r, g, b, 1);

		/*if(player->powers[pw_invulnerability] > 4*32
			|| (player->powers[pw_invulnerability]&8))
		{
			

		}*/
		// Restore the normal rendering state.
		gl.Func(DGL_BLENDING, DGL_SRC_ALPHA, DGL_ONE_MINUS_SRC_ALPHA);
		gl.Enable(DGL_TEXTURING);
	}
}

//
// D_Display
//  draw current display, possibly wiping it from the previous
//

// wipegamestate can be set to -1 to force a wipe on the next draw
gamestate_t		wipegamestate = GS_DEMOSCREEN;
extern boolean	setsizeneeded;
extern int		setblocks;
extern boolean	inhelpscreens;
extern float	lookOffset;

//void R_ExecuteSetViewSize (void);
void R_SetAllDoomsdayFlags();

void D_Display (void)
{
    static  boolean		viewactivestate = false;
    static  boolean		menuactivestate = false;
    static  boolean		inhelpscreensstate = false;
    static  boolean		fullscreen = false;
    static  gamestate_t		oldgamestate = -1;
    static  int			borderdrawcount;
//    int				nowtime;
//    int				tics;
//    int				wipestart;
    int				y;
//    boolean			done;
//    boolean			wipe;
    boolean			redrawsbar;
	player_t		*player = &players[displayplayer];

    if (nodrawers) return;                  // for comparative timing / profiling
		
    redrawsbar = false;

    // Change the view size if needed
	if(setsizeneeded)
	{
		setsizeneeded = false;
		if(setblocks > 10)
		{
			// Full screen.
			gi.ViewWindow(0, 0, 320, 200);
		}
		else
		{
			int w = setblocks*32;
			int h = setblocks*(200-ST_HEIGHT*cfg.sbarscale/20)/10;
			gi.ViewWindow(160-(w>>1), (200-ST_HEIGHT*cfg.sbarscale/20-h)>>1, w, h);
		}
	}

    // save the current screen if about to wipe
    /*if(gamestate != wipegamestate)
    {
		wipe = true;
		wipe_StartScreen(0, 0, SCREENWIDTH, SCREENHEIGHT);
    }
    else
		wipe = false;

	if(gamestate == GS_LEVEL && gametic)
	HU_Erase();*/
    
	// do buffered drawing
	switch (gamestate)
    {
	case GS_LEVEL:
		if(IS_CLIENT && !gi.Get(DD_GAME_READY)) 
			break;
		if(!gametic) break;
		if(automapactive) 
			AM_Drawer ();
		else
		{
			// Draw the player view.
			R_SetAllDoomsdayFlags();
			players[displayplayer].plr->mo->ddflags |= DDMF_DONTDRAW;
			// The view angle offset.
			gi.Set(DD_VIEWANGLE_OFFSET, ANGLE_MAX * -lookOffset);
			if(IS_CLIENT)
			{
				// Clients get the color filter from the server.
				gi.GL_SetFilter(players[displayplayer].plr->filter);
			}
			// How about fullbright?
			gi.Set(DD_FULLBRIGHT, (player->powers[pw_infrared] > 4*32)
				|| (player->powers[pw_infrared]&8) 
				|| player->powers[pw_invulnerability] > 30);
			gi.RenderPlayerView(players[displayplayer].plr);
			R_DrawSpecialFilter(); 
			X_Drawer();
		}
		if(/*wipe || */(viewheight != 200/* && fullscreen*/))
			redrawsbar = true;
		if(inhelpscreensstate && !inhelpscreens)
			redrawsbar = true;              // just put away the help screen
		ST_Drawer(viewheight == 200, redrawsbar);
		fullscreen = viewheight == 200;
		HU_Drawer();
		// Need to update the borders?
		if(oldgamestate != GS_LEVEL 
			|| (!automapactive 
			&& (gi.Get(DD_VIEWWINDOW_WIDTH) != 320 || menuactive || cfg.sbarscale < 20)))
		{
			// Update the borders.
			gi.Update(DDUF_BORDER);
		}
		break;

	case GS_INTERMISSION:
		WI_Drawer ();
		break;

	case GS_FINALE:
		F_Drawer ();
		break;

	case GS_DEMOSCREEN:
		D_PageDrawer ();
		break;
	}
	gi.Update(DDUF_FULLSCREEN);
    
    // draw buffered stuff to screen
    //I_UpdateNoBlit ();
    
    // draw the view directly
/*    if(gamestate == GS_LEVEL && !automapactive && gametic)
	{
		R_SetDoomsdayFlags();
		players[displayplayer].plr->mo->ddflags |= DDMF_DONTDRAW;
		gi.RenderPlayerView(players[displayplayer].plr);
	}
    if(gamestate == GS_LEVEL && gametic)
		HU_Drawer();
    */
    // clean up border stuff
    /*if(gamestate != oldgamestate && gamestate != GS_LEVEL)
		I_SetPalette (W_CacheLumpName ("PLAYPAL",PU_CACHE));*/

    // see if the border needs to be initially drawn
/*    if(gamestate == GS_LEVEL && oldgamestate != GS_LEVEL)
    {
		viewactivestate = false;        // view was not active
		//R_FillBackScreen ();    // draw the pattern into the back screen
    }*/

    // see if the border needs to be updated to the screen
/*    if(gamestate == GS_LEVEL && !automapactive && scaledviewwidth != 320)
    {
		if(menuactive || menuactivestate || !viewactivestate)
			borderdrawcount = 3;
		if(borderdrawcount)
		{
			R_DrawViewBorder ();    // erase old menu stuff
			borderdrawcount--;
		}
    }*/

    menuactivestate = menuactive;
    viewactivestate = viewactive;
    inhelpscreensstate = inhelpscreens;
    oldgamestate = wipegamestate = gamestate;
    
    // draw pause pic
    if(paused)
    {
		if(automapactive)
			y = 4;
		else
			y = viewwindowy+4;
	
		gi.GL_DrawPatch(/*viewwindowx + (scaledviewwidth-68)/2*/126, 
			y, W_GetNumForName("M_PAUSE"));
    }


    // menus go directly to the screen
    //M_Drawer ();          // menu is drawn even on top of everything
    //NetUpdate ();         // send out any new accumulation


    /*// normal update
    if (!wipe)
    {
	I_FinishUpdate ();              // page flip or blit buffer
	return;
    }*/
    
    // wipe update
    //wipe_EndScreen(0, 0, SCREENWIDTH, SCREENHEIGHT);

    //wipestart = I_GetTime () - 1;

    /*do
    {
		do
		{
			nowtime = I_GetTime ();
			tics = nowtime - wipestart;
		} while (!tics);
		wipestart = nowtime;
		done = wipe_ScreenWipe(wipe_Melt
			       , 0, 0, SCREENWIDTH, SCREENHEIGHT, tics);
	I_UpdateNoBlit ();
	M_Drawer ();                            // menu is drawn even on top of wipes
	I_FinishUpdate ();                      // page flip or blit buffer
    } while (!done);*/


}


void R_SetAllDoomsdayFlags()
{
	int			isClient = IS_CLIENT;
	int			i, off;
	sector_t	*sec = sectors;
	mobj_t		*iter;

	// Only visible things are in the sector thinglists, so this is good.
	for(i=0; i<numsectors; i++, sec++)
		for(iter=sec->thinglist; iter; iter=iter->snext)
		{
			// Client mobjs can't be set here.
			if(isClient && iter->ddflags & DDMF_REMOTE) continue;
			
			// Reset the flags for a new frame.
			if(server)
			{
				// Servers must not clear the Sent flag.
				iter->ddflags &= DDMF_SENT;	
			}
			else
				iter->ddflags = 0;
			
			// Local objects aren't sent to clients.
			if(iter->flags & MF_LOCAL) iter->ddflags |= DDMF_LOCAL;
			if(iter->flags & MF_MISSILE) 
			{
				iter->ddflags |= DDMF_MISSILE;
			}

			// The torches often go into the ceiling. This'll prevent
			// them from 'jumping'.
			if(iter->type == MT_MISC41 || iter->type == MT_MISC42 
				|| iter->type == MT_MISC43 // tall torches
				|| iter->type == MT_MISC44 || iter->type == MT_MISC45
				|| iter->type == MT_MISC46) // short torches
				iter->ddflags |= DDMF_NOFITBOTTOM;

			if(iter->flags & MF_BRIGHTSHADOW)
				iter->ddflags |= DDMF_BRIGHTSHADOW;
			else if(iter->flags & MF_SHADOW)
				iter->ddflags |= DDMF_SHADOW;

			if((iter->flags & MF_VIEWALIGN && !(iter->flags & MF_MISSILE)) 
				|| iter->flags & MF_FLOAT 
				|| (iter->flags & MF_MISSILE && !(iter->flags & MF_VIEWALIGN)))
				iter->ddflags |= DDMF_VIEWALIGN;

			iter->ddflags |= iter->flags & MF_TRANSLATION;

			// An offset for the light emitted by this object.
			off = MobjLightOffsets[iter->type];
			if(off < 0) off = 8-off;
			// off must now be in range 0-15.
			iter->ddflags |= off << DDMF_LIGHTOFFSETSHIFT;			
		}
}


