// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:  none
//
//-----------------------------------------------------------------------------


static const char
rcsid[] = "$Id: g_game.c,v 1.8 1997/02/03 22:45:09 b1 Exp $";

#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "doomdef.h" 
#include "doomstat.h"
#include "d_action.h"
#include "d_config.h"

#include "z_zone.h"
#include "f_finale.h"
#include "m_argv.h"
#include "m_misc.h"
#include "m_menu.h"
#include "m_random.h"
#include "i_system.h"

#include "p_local.h"
#include "p_setup.h"
#include "p_saveg.h"
#include "p_tick.h"

#include "d_main.h"

#include "d_netjd.h"

#include "wi_stuff.h"
#include "hu_stuff.h"
#include "st_stuff.h"
#include "am_map.h"

// Needs access to LFB.
#include "v_video.h"

#include "w_wad.h"

#include "p_local.h" 

#include "s_sound.h"

// Data.
#include "dstrings.h"
#include "sounds.h"

// SKY handling - still the wrong place.
#include "r_data.h"
#include "r_sky.h"



#include "g_game.h"


#define SAVEGAMESIZE	0x2c000
#define SAVESTRINGSIZE	24



boolean	G_CheckDemoStatus (void); 
void	G_ReadDemoTiccmd (player_t *pl); 
void	G_WriteDemoTiccmd (ticcmd_t* cmd); 
void	G_PlayerReborn (int player); 
void	G_InitNew (skill_t skill, int episode, int map); 
 
void	G_DoReborn (int playernum); 
 
void	G_DoLoadLevel (void); 
void	G_DoNewGame (void); 
void	G_DoLoadGame (void); 
void	G_DoPlayDemo (void); 
void	G_DoCompleted (void); 
void	G_DoVictory (void); 
void	G_DoWorldDone (void); 
void	G_DoSaveGame (void); 
 
 
gameaction_t    gameaction; 
gamestate_t     gamestate; 
skill_t         gameskill; 
boolean		respawnmonsters;
int             gameepisode; 
int             gamemap; 
 
boolean         paused; 
boolean         sendpause;             	// send a pause event next tic 
boolean         sendsave;             	// send a save event next tic 
boolean         usergame;               // ok to save / end game 
 
boolean         timingdemo;             // if true, exit with report on completion 
boolean         nodrawers;				// for comparative timing purposes 
boolean         noblit;                 // for comparative timing purposes 
int             starttime;          	// for comparative timing purposes  	 
 
boolean         viewactive; 
 
boolean         deathmatch;           	// only if started as net death 
//boolean         netgame;                // only true if packets are broadcast 
//boolean         playeringame[MAXPLAYERS]; 
player_t        players[MAXPLAYERS]; 
 
//int             consoleplayer;          // player taking events and displaying 
//int             displayplayer;          // view being displayed 
//int             gametic; 
int             levelstarttic;          // gametic at level start 
int             totalkills, totalitems, totalsecret;    // for intermission 
 
char            demoname[32]; 
boolean         demorecording; 
boolean         demoplayback; 
boolean		netdemo; 
int			maxDemoSize;
byte*		demobuffer;
byte*		demo_p;
//byte*		demoend; 
boolean         singledemo;            	// quit after playing a demo from cmdline 
 
boolean         precache = true;        // if true, load all graphics at start 
 
wbstartstruct_t wminfo;               	// parms for world map / intermission 
 
//short		consistancy[MAXPLAYERS][BACKUPTICS]; 
 
byte*		savebuffer;
 
 
// 
// controls (have defaults) 
// 
/*int             key_right;
int		key_left;

int		key_up;
int		key_down; 
int             key_strafeleft;
int		key_straferight; 
int             key_fire;
int		key_use;
int		key_strafe;
int		key_speed; 
 
int             mousebfire; 
int             mousebstrafe; 
int             mousebforward; 
 
int             joybfire; 
int             joybstrafe; 
int             joybuse; 
int             joybspeed; 
 */


jdoom_config_t cfg;

// Looking around.
int		povangle = -1;				// -1 means centered (really 0 - 7).
float	targetLookOffset=0;
float	lookOffset=0;

 
 
#define MAXPLMOVE		(forwardmove[1]) 
 
#define TURBOTHRESHOLD	0x32

fixed_t		forwardmove[2] = {0x19, 0x32}; 
fixed_t		sidemove[2] = {0x18, 0x28}; 
fixed_t		angleturn[3] = {640, 1280, 320};	// + slow turn 

#define SLOWTURNTICS	6 
 
#define NUMKEYS		256 

boolean         gamekeydown[NUMKEYS]; 
int             turnheld;				// for accelerative turning 
 
boolean		mousearray[4]; 
boolean*	mousebuttons = &mousearray[1];		// allow [-1]

// mouse values are used once 
int     mousex;
int		mousey;         

int     dclicktime;
int		dclickstate;
int		dclicks; 
int     dclicktime2;
int		dclickstate2;
int		dclicks2;

// joystick values are repeated 
int     joyxmove;
int		joyymove;
boolean joyarray[5]; 
boolean* joybuttons = &joyarray[1];		// allow [-1] 
 
int		savegameslot; 
char	savedescription[32]; 
 
 
#define	BODYQUESIZE	32

mobj_t*	bodyque[BODYQUESIZE]; 
int		bodyqueslot; 
 
void*	statcopy = NULL;				// for statistics driver
 
 
 
/*int G_CmdChecksum (ticcmd_t* cmd) 
{ 
    int		i;
    int		sum = 0; 
	 
    for (i=0 ; i< sizeof(*cmd)/4 - 1 ; i++) 
	sum += ((int *)cmd)[i]; 
		 
    return sum; 
} */
 

// Offset is in 'angles', where 110 corresponds 85.
// The delta has higher precision with small offsets.
char G_MakeLookDelta(float offset)
{
	boolean minus = offset < 0;

	offset = sqrt(fabs(offset)) * DELTAMUL;
	if(minus) offset = -offset;
	// It's only a char...
	if(offset > 127) offset = 127;
	if(offset < -128) offset = -128;
	return (signed char) offset;
}

void G_AdjustAngle(player_t *player, int turn)
{
	if(!player->plr->mo) return;	// Sorry, can't help you, pal.

	// Turn around.
	if(player->playerstate != PST_DEAD)
		player->plr->clAngle += (turn << FRACBITS);
}

void G_AdjustLookDir(player_t *player, int look)
{
	ddplayer_t *ddplr = player->plr;

	if(look)
	{
		if(look == TOCENTER)
		{
			player->centering = true;
		}
		else
		{
			int spd = cfg.lookSpeed;
			if(demoplayback || demorecording) // Must be backwards compatible.
				spd = 5;
			ddplr->clLookDir += spd*look;
		}
	}
	if(player->centering)
	{
		if(ddplr->clLookDir > 0)
		{
			ddplr->clLookDir -= 8;
		}
		else if(ddplr->clLookDir < 0)
		{
			ddplr->clLookDir += 8;
		}
		if(abs(ddplr->clLookDir) < 8)
		{
			ddplr->clLookDir = 0;
			player->centering = false;
		}
	}
}

void G_SetCmdViewAngles(ticcmd_t *cmd, player_t *pl)
{
	// These will be sent to the server (or P_MovePlayer).
	cmd->angle = pl->plr->clAngle >> 16;
	// 110 corresponds 85 degrees.
	if(pl->plr->clLookDir > 110) pl->plr->clLookDir = 110;
	if(pl->plr->clLookDir < -110) pl->plr->clLookDir = -110;
	cmd->lookdir = pl->plr->clLookDir/110 * DDMAXSHORT;
}

//
// G_BuildTiccmd
// Builds a ticcmd from all of the available inputs
// or reads it from the demo buffer. 
// If recording a demo, write it out 
// 
void G_BuildTiccmd (ticcmd_t* cmd) 
{ 
	boolean pausestate = paused || (!netgame && menuactive && !demoplayback);
    int		i; 
    boolean	strafe;
    boolean	bstrafe; 
    int		speed;
    int		tspeed; 
    int		forward;
    int		side;
	int		turn = 0;
	player_t *cplr = &players[consoleplayer];
    
	memset(cmd, 0, sizeof(*cmd));
	
//    cmd->consistancy = consistancy[consoleplayer][maketic%BACKUPTICS]; 
 
    strafe = actions[DA_STRAFE].on; 
    speed = actions[DA_SPEED].on;
	if(cfg.alwaysRun) speed = !speed; // Walk -> run, run -> walk.
 
    forward = side = 0;
    
    // use two stage accelerative turning
    // on the keyboard and joystick
    if (joyxmove < 0 || joyxmove > 0 || actions[DA_TURNRIGHT].on 
		|| actions[DA_TURNLEFT].on)
		turnheld += ticdup; 
    else 
		turnheld = 0; 

    if (turnheld < SLOWTURNTICS) 
		tspeed = 2;             // slow turn 
    else 
		tspeed = speed;
    
    // let movement keys cancel each other out
    if (strafe) 
    { 
		if(actions[DA_TURNRIGHT].on) 
		{
			// fprintf(stderr, "strafe right\n");
			side += sidemove[speed]; 
		}
		if(actions[DA_TURNLEFT].on) 
		{
		    //	fprintf(stderr, "strafe left\n");
		    side -= sidemove[speed]; 
		}
		if(joyxmove > 0) 
			side += sidemove[speed]; 
		if(joyxmove < 0) 
			side -= sidemove[speed]; 
    } 
    else 
    { 
		if(actions[DA_TURNRIGHT].on) 
		    turn -= angleturn[tspeed]; 
		if(actions[DA_TURNLEFT].on) 
			turn += angleturn[tspeed]; 
		if(joyxmove > 0) 
		    turn -= angleturn[tspeed]; 
		if(joyxmove < 0) 
			turn += angleturn[tspeed]; 
	} 
 
    if(actions[DA_FORWARD].on) 
    {
		// fprintf(stderr, "up\n");
		forward += forwardmove[speed]; 
    }
    if(actions[DA_BACKWARD].on) 
    {
		// fprintf(stderr, "down\n");
		forward -= forwardmove[speed]; 
    }
    if(joyymove < 0) 
		forward += forwardmove[speed]; 
    if(joyymove > 0) 
		forward -= forwardmove[speed]; 
    if(actions[DA_STRAFERIGHT].on) 
		side += sidemove[speed]; 
    if(actions[DA_STRAFELEFT].on) 
		side -= sidemove[speed];
    
    // buttons
    //cmd->chatchar = HU_dequeueChatChar(); 
 
    if(actions[DA_FIRE].on) cmd->buttons |= BT_ATTACK; 
 
    if(actions[DA_USE].on) 
	{ 
		cmd->buttons |= BT_USE;
		// clear double clicks if hit use button 
		dclicks = 0;                   
    } 

	if(actions[DA_JUMP].on)
	{
		cmd->buttons |= BT_JUMP;
	}

    // chainsaw overrides 
    for(i=0; i<NUMWEAPONS-1; i++)        
		if(actions[DA_WEAPON1+i].on) 
		{ 
			cmd->buttons &= ~BT_JUMP;
			cmd->buttons |= BT_CHANGE; 
			cmd->buttons |= i<<BT_WEAPONSHIFT; 
			break; 
		}

	if(actions[DA_NEXTWEAPON].on || actions[DA_PREVIOUSWEAPON].on)
	{
		cmd->buttons = BT_SPECIAL 
			| (actions[DA_NEXTWEAPON].on? BTS_NEXTWEAPON : BTS_PREVWEAPON);
	}

    // mouse
    /*if(mousebuttons[mousebforward]) 
		forward += forwardmove[speed];*/
    
    // forward double click
    if(actions[DA_FORWARD].on != dclickstate && dclicktime > 1 && cfg.dclickuse)  
    { 
		dclickstate = actions[DA_FORWARD].on; 
		if (dclickstate) 
			dclicks++; 
		if (dclicks == 2) 
		{ 
			cmd->buttons |= BT_USE; 
			dclicks = 0; 
		} 
		else 
			dclicktime = 0; 
	} 
	else 
	{ 
		dclicktime += ticdup; 
		if (dclicktime > 20) 
		{ 
			dclicks = 0; 
			dclickstate = 0; 
		} 
	}
    
    // strafe double click
    bstrafe = strafe; //mousebuttons[mousebstrafe] || joybuttons[joybstrafe]; 
    if(bstrafe != dclickstate2 && dclicktime2 > 1 && cfg.dclickuse) 
    { 
		dclickstate2 = bstrafe; 
		if (dclickstate2) 
			dclicks2++; 
		if (dclicks2 == 2) 
		{ 
			cmd->buttons |= BT_USE; 
			dclicks2 = 0; 
		} 
		else 
			dclicktime2 = 0; 
    } 
    else 
    { 
		dclicktime2 += ticdup; 
		if (dclicktime2 > 20) 
		{ 
			dclicks2 = 0; 
			dclickstate2 = 0; 
		} 
    } 
	if(!pausestate)
	{
		if((!cfg.usemlook && !actions[DA_MLOOK].on) || 
			demorecording || demoplayback || players[consoleplayer].playerstate == PST_DEAD) 
		{
			forward += mousey;
		}
		else
		{
			float adj = (((mousey*0x8)<<16)/(float)ANGLE_180 * 180 * 110.0/85.0);
			if(cfg.mlookInverseY) adj = -adj;
			//cmd->lookdirdelta = G_MakeLookDelta(adj);
			cplr->plr->clLookDir += adj;
		}
	}

    if(strafe) 
		side += mousex*2; 
    else 
		turn -= mousex*0x8; 

    mousex = mousey = 0; 
	 
    if(forward > MAXPLMOVE) 
		forward = MAXPLMOVE; 
    else if (forward < -MAXPLMOVE) 
		forward = -MAXPLMOVE; 
    if(side > MAXPLMOVE) 
		side = MAXPLMOVE; 
    else if (side < -MAXPLMOVE) 
		side = -MAXPLMOVE; 
 
    cmd->forwardmove += forward; 
    cmd->sidemove += side;

	if(cmd->forwardmove > MAXPLMOVE/3 && cfg.lookSpring && !actions[DA_MLOOK].on)
		cmd->buttons |= BT_LOOKCENTER;

	// Key look.
	if(cplr->playerstate == PST_LIVE && !pausestate)
	{
		i = 0;
		if(actions[DA_LOOKCENTER].on)
		{
			cmd->buttons |= BT_LOOKCENTER;
			i = TOCENTER;
		}
		else 
		{
			if(actions[DA_LOOKUP].on)
			{
				i += cfg.lookSpeed;
			}
			if(actions[DA_LOOKDOWN].on)
			{
				i -= cfg.lookSpeed;
			}	
		}
		/*if(i)
		{
		//cmd->lookdirdelta = G_MakeLookDelta(i);
		//G_AdjustLookDir(cplr, i);
		}*/
		G_AdjustLookDir(cplr, i);
	}

	if(!pausestate) G_AdjustAngle(cplr, turn);
	G_SetCmdViewAngles(cmd, cplr);

	// special buttons
    if(sendpause) 
    { 
		sendpause = false; 
		cmd->buttons = BT_SPECIAL | BTS_PAUSE; 
    } 
	if(sendsave) 
    { 
		sendsave = false; 
		cmd->buttons = BT_SPECIAL | BTS_SAVEGAME | (savegameslot<<BTS_SAVESHIFT); 
    } 

	if(IS_CLIENT)
	{
		memcpy(&players[consoleplayer].cmd, cmd, sizeof(*cmd));
	}
} 
 

//
// G_DoLoadLevel 
//
extern  gamestate_t     wipegamestate; 
 
void G_DoLoadLevel (void) 
{ 
    int             i; 

	NetSv_SendGameState(GSF_CHANGE_MAP, DDSP_ALL_PLAYERS);

    // set the sky map for the episode
    if(gamemode == commercial)
    {
		skytexture = R_TextureNumForName ("SKY3");
		if(gamemap < 12)
			skytexture = R_TextureNumForName ("SKY1");
		else if(gamemap < 21)
			skytexture = R_TextureNumForName ("SKY2");
    }
    else switch (gameepisode) 
	{
	case 1:
		skytexture = R_TextureNumForName ("SKY1"); 
		break; 
	case 2:
		skytexture = R_TextureNumForName ("SKY2"); 
		break; 
	case 3: 
		skytexture = R_TextureNumForName ("SKY3"); 
		break; 
	case 4:	// Special Edition sky
		skytexture = R_TextureNumForName ("SKY4");
		break;
	} 
	// Set Doomsday sky parameters.
	gi.SkyParams(DD_SKY, DD_HORIZON, -0.105f);	// -9.45 degrees.
	gi.SkyParams(DD_SKY, DD_HEIGHT, .49f);		// Really .48
	gi.SkyParams(0, DD_ENABLE, 0);	
	gi.SkyParams(0, DD_COLOR_LIMIT, 0);
	gi.SkyParams(0, DD_TEXTURE, skytexture);
	
    levelstarttic = gametic;        // for time calculation
    
    if (wipegamestate == GS_LEVEL) 
		wipegamestate = -1;             // force a wipe 

    gamestate = GS_LEVEL; 

	if(!gi.Get(DD_NOVIDEO))
	{
		gl.Clear(DGL_COLOR_BUFFER_BIT);
	}
	
    for (i=0 ; i<MAXPLAYERS ; i++) 
    { 
		if (players[i].plr->ingame && players[i].playerstate == PST_DEAD) 
			players[i].playerstate = PST_REBORN; 
		memset(players[i].frags, 0, sizeof(players[i].frags)); 
    } 
		 
    P_SetupLevel (gameepisode, gamemap, 0, gameskill);    
    gi.Set(DD_DISPLAYPLAYER, consoleplayer);	// view the guy you are playing    
    starttime = I_GetTime (); 
    gameaction = ga_nothing; 
    Z_CheckHeap ();
    
    // clear cmd building stuff
    memset (gamekeydown, 0, sizeof(gamekeydown)); 
    joyxmove = joyymove = 0; 
    mousex = mousey = 0; 
    sendpause = sendsave = paused = false; 
    memset (mousebuttons, 0, sizeof(mousebuttons)); 
    memset (joybuttons, 0, sizeof(joybuttons)); 
} 
 

int CCmdCycleSpy(int argc, char **argv)
{
	// FIXME: The engine should do this.
#if 0
	if(gamestate == GS_LEVEL && !deathmatch)
	{ // Cycle the display player
		do
		{
			gi.Set(DD_DISPLAYPLAYER, displayplayer+1);
			if(displayplayer == MAXPLAYERS)
			{
				gi.Set(DD_DISPLAYPLAYER, 0);
			}
		} while(!players[displayplayer].plr->ingame
			&& displayplayer != consoleplayer);
	}
#endif
	return true;
}

//
// G_Responder  
// Get info needed to make ticcmd_ts for the players.
// Return false if the event should be checked for bindings.
// 
boolean G_Responder (event_t* ev) 
{ 
	int		i;
    
    // any other key pops up menu if in demos
    if(gameaction == ga_nothing && !singledemo && 
		(demoplayback || gamestate == GS_DEMOSCREEN)) 
    { 
		if(ev->type == ev_keydown || (ev->type == ev_mouse && ev->data1) || 
			(ev->type == ev_joystick && ev->data1)) 
		{ 
			M_StartControlPanel (); 
			return false; 
		} 
		return false; 
    } 
 
    if(gamestate == GS_LEVEL) 
    { 
		if(HU_Responder(ev)) 
			return true;	// chat ate the event 
		if(ST_Responder(ev)) 
			return true;	// status window ate it 
		if(AM_Responder(ev)) 
			return true;	// automap ate it 
    } 
	 
    if (gamestate == GS_FINALE) 
    { 
		if(F_Responder(ev)) 
			return true;	// finale ate the event 
    } 

    switch (ev->type) 
    { 
    case ev_keydown:
		/*if(ev->data1 == KEY_PAUSE) 
		{ 
			sendpause = true; 
			return true; 
		}*/
		if(ev->data1 < NUMKEYS) gamekeydown[ev->data1] = true; 
		return false;  
 
	case ev_keyup: 
		if(ev->data1 < NUMKEYS) gamekeydown[ev->data1] = false; 
		return false;   // always let key up events filter down 

	case ev_keyrepeat:
		return false; 
		 
	case ev_mouse:
		mousex = (ev->data1 * (cfg.mouseSensiX*2+5)) / 6;
		mousey = (ev->data2 * (cfg.mouseSensiY*2+5)) / 6;
		return(true); // eat events

	case ev_mousebdown:
		for(i=0; i<3; i++)
			if(ev->data1 & (1<<i))
				mousebuttons[i] = true;
		return(false);

	case ev_mousebup:
		for(i=0; i<3; i++)
			if(ev->data1 & (1<<i))
				mousebuttons[i] = false;
		return(false);

	case ev_joystick: // Joystick movement
		joyxmove = ev->data1;
		joyymove = ev->data2;
		return(true); // eat events

	case ev_joybdown: 
		for(i=0; i<32; i++) 
			if(ev->data1 & (1<<i)) 
				joybuttons[i] = true;
		return(false); // eat events

	case ev_joybup: 
		for(i=0; i<32; i++) 
			if(ev->data1 & (1<<i)) 
				joybuttons[i] = false;
		return(false); // eat events
 
	case ev_povup:
		povangle = -1;
		// If looking around with PoV, don't allow bindings.
		if(cfg.povLookAround) return true;
		break;

	case ev_povdown:
		povangle = ev->data1;
		if(cfg.povLookAround) return true;
		break;

	default: 
		break; 
    } 
	// The event wasn't used. 
    return false; 
} 

void G_SpecialButton(player_t *pl)
{
	if (pl->plr->ingame) 
	{ 
		if (pl->cmd.buttons & BT_SPECIAL) 
		{ 
			switch (pl->cmd.buttons & BT_SPECIALMASK) 
			{ 
			case BTS_PAUSE: 
				paused ^= 1; 
				if (paused) 
					S_PauseSound (); 
				else 
					S_ResumeSound (); 
				pl->cmd.buttons = 0;
				break; 
				
			case BTS_SAVEGAME: 
				if (!savedescription[0]) 
					strcpy (savedescription, "NET GAME"); 
				savegameslot =  
					(pl->cmd.buttons & BTS_SAVEMASK)>>BTS_SAVESHIFT; 
				gameaction = ga_savegame; 
				pl->cmd.buttons = 0;
				break; 
			} 
		} 
	}
}
 
// This is called at all times, no matter gamestate.
void Plr_Ticker(void)
{
	boolean pausestate = paused || (!netgame && menuactive && !demoplayback);
	int	i, count;
	ticcmd_t temp;

	// Players get to think away all their buffered cmds.
	for(i=0; i<MAXPLAYERS; i++)
		if(players[i].plr->ingame)
		{
			int ticfound;
			count = 0;
			// Only run cmds when they are available.
			while((ticfound = gi.GetTicCmd(&temp, i)) || demoplayback)
			{
				memcpy(&players[i].cmd, &temp, sizeof(ticcmd_t));
				count++;
				
				/*gi.Message( "P_Tick: cl:%i time:%i found:%i count:%i\n",
				i, leveltime, ticfound, count);*/
				
				// Demos override normal cmds.
				if(demoplayback)
					G_ReadDemoTiccmd(&players[i]);//.cmd);
				
				if(!pausestate)
				{
					// Since demo playback is synchronous (one cmd per tic)
					// demo recording should be, too.
					if(demorecording)
						G_WriteDemoTiccmd(&players[i].cmd);
				}
				
				// Check for special buttons (pause and netsave).
				G_SpecialButton(&players[i]);
				
				// The player thinks.
				if(gamestate == GS_LEVEL)
				{
					if(!pausestate) P_PlayerThink(&players[i]);
				}
				
				// When playing back demos, only one tic is run.
				if(demoplayback) break;
			}
		}
}

 
 
//
// G_Ticker
// Make ticcmd_ts for the players.
//
void G_Ticker (void) 
{ 
    int		i;
//    int		buf; 
    //ticcmd_t*	cmd;
    
	if(IS_CLIENT && !gi.Get(DD_GAME_READY)) return;

    // do player reborns if needed
    for (i=0 ; i<MAXPLAYERS ; i++) 
	{
		if(players[i].plr->ingame && players[i].playerstate == PST_REBORN) 
			G_DoReborn (i);
	}
		
	// do things to change the game state
	while (gameaction != ga_nothing) 
	{ 
		switch (gameaction) 
		{ 
		case ga_loadlevel: 
			G_DoLoadLevel (); 
			break; 
		case ga_newgame: 
			G_DoNewGame (); 
			break; 
		case ga_loadgame: 
			G_DoLoadGame (); 
			break; 
		case ga_savegame: 
			G_DoSaveGame (); 
			break; 
		case ga_playdemo: 
			G_DoPlayDemo (); 
			break; 
		case ga_completed: 
			G_DoCompleted (); 
			break; 
		case ga_victory: 
			F_StartFinale (); 
			break; 
		case ga_worlddone: 
			G_DoWorldDone (); 
			break; 
		case ga_screenshot: 
			M_ScreenShot (); 
			gameaction = ga_nothing; 
			break; 
		case ga_nothing: 
			break; 
		} 
	}
		
	// Look around.
	if(povangle != -1)
	{
		targetLookOffset = povangle/8.0f;
		if(targetLookOffset == .5f)
		{
			if(lookOffset < 0) targetLookOffset = -.5f;
		}
		else if(targetLookOffset > .5)
			targetLookOffset -= 1;
	}
	else targetLookOffset = 0;
	
	if(targetLookOffset != lookOffset && cfg.povLookAround)
	{
		float diff = (targetLookOffset - lookOffset)/2;
		if(diff > .075f) diff = .075f;
		if(diff < -.075f) diff = -.075f; 
		lookOffset += diff;
	}
	
	// get commands, check consistancy,
	// and build new consistancy check
/*	buf = (gametic/ticdup)%BACKUPTICS; 
	
	for (i=0 ; i<MAXPLAYERS ; i++)
	{
		if (players[i].plr->ingame) 
		{ 
			cmd = &players[i].cmd; 
			
			//memcpy (cmd, &netcmds[i][buf], sizeof(ticcmd_t)); 
			
			gi.GetTicCmd(cmd, i);
			
			if(demoplayback) G_ReadDemoTiccmd(cmd); 
			if(demorecording) G_WriteDemoTiccmd(cmd);
			
			// check for turbo cheats
			if(cmd->forwardmove > TURBOTHRESHOLD 
				&& !(gametic&31) && ((gametic>>5)&3) == i)
			{
				static char turbomessage[80];
				extern char *player_names[4];
				sprintf(turbomessage, "%s is turbo!", player_names[i]);
				players[consoleplayer].message = turbomessage;
			}
		}
	}*/

	// Players get to think first.
	Plr_Ticker();
	
	// check for special buttons
/*	for (i=0 ; i<MAXPLAYERS ; i++)
	{
		if (players[i].plr->ingame) 
		{ 
			if (players[i].cmd.buttons & BT_SPECIAL) 
			{ 
				switch (players[i].cmd.buttons & BT_SPECIALMASK) 
				{ 
				case BTS_PAUSE: 
					paused ^= 1; 
					if (paused) 
						S_PauseSound (); 
					else 
						S_ResumeSound (); 
					break; 
					
				case BTS_SAVEGAME: 
					if (!savedescription[0]) 
						strcpy (savedescription, "NET GAME"); 
					savegameslot =  
						(players[i].cmd.buttons & BTS_SAVEMASK)>>BTS_SAVESHIFT; 
					gameaction = ga_savegame; 
					break; 
				} 
			} 
		}
	}*/
	
	// do main actions
	switch (gamestate) 
	{ 
	case GS_LEVEL: 
		P_Ticker (); 
		if(IS_DEDICATED) break;
		ST_Ticker (); 
		AM_Ticker (); 
		HU_Ticker ();            
		break; 
		
	case GS_INTERMISSION: 
		WI_Ticker (); 
		break; 
		
	case GS_FINALE: 
		F_Ticker (); 
		break; 
		
	case GS_DEMOSCREEN: 
		if(IS_DEDICATED) break;
		D_PageTicker (); 
		break; 
	}        

	if(netgame && server) NetSv_Ticker();
} 
 
 
//
// PLAYER STRUCTURE FUNCTIONS
// also see P_SpawnPlayer in P_Things
//

//
// G_InitPlayer 
// Called at the start.
// Called by the game initialization functions.
//
void G_InitPlayer (int player) 
{ 
    player_t*	p; 
 
    // set up the saved info         
    p = &players[player]; 
	 
    // clear everything else to defaults 
    G_PlayerReborn (player); 
	 
} 
 
 

//
// G_PlayerFinishLevel
// Can when a player completes a level.
//
void G_PlayerFinishLevel (int player) 
{ 
    player_t*	p; 
	 
    p = &players[player]; 
	 
    memset (p->powers, 0, sizeof (p->powers)); 
    memset (p->cards, 0, sizeof (p->cards)); 
    p->plr->mo->flags &= ~MF_SHADOW;		// cancel invisibility 
    p->plr->extralight = 0;			// cancel gun flashes 
    p->plr->fixedcolormap = 0;		// cancel ir gogles 
    p->damagecount = 0;			// no palette changes 
    p->bonuscount = 0; 
} 
 

void ClearPlayer(player_t *p)
{
	ddplayer_t	*ddplayer = p->plr;
	int			playeringame = ddplayer->ingame;

	memset(p, 0, sizeof(*p));
	// Restore the pointer to ddplayer.
	p->plr = ddplayer;
	// Also clear ddplayer.
	memset(ddplayer, 0, sizeof(*ddplayer));
	// Restore the pointer to this player.
	ddplayer->extradata = p;
	// Restore the playeringame data.
	ddplayer->ingame = playeringame;
}


//
// G_PlayerReborn
// Called after a player dies 
// almost everything is cleared and initialized 
//
void G_PlayerReborn (int player) 
{ 
    player_t*	p; 
    int			i; 
    int			frags[MAXPLAYERS]; 
    int			killcount;
    int			itemcount;
    int			secretcount; 
	 
    memcpy(frags, players[player].frags, sizeof(frags)); 
    killcount = players[player].killcount; 
    itemcount = players[player].itemcount; 
    secretcount = players[player].secretcount; 
	 
    p = &players[player]; 
    ClearPlayer(p);
 
    memcpy(players[player].frags, frags, sizeof(players[player].frags)); 
    players[player].killcount = killcount; 
    players[player].itemcount = itemcount; 
    players[player].secretcount = secretcount; 
 
    p->usedown = p->attackdown = true;	// don't do anything immediately 
    p->playerstate = PST_LIVE;       
    p->health = MAXHEALTH; 
    p->readyweapon = p->pendingweapon = wp_pistol; 
    p->weaponowned[wp_fist] = true; 
    p->weaponowned[wp_pistol] = true; 
    p->ammo[am_clip] = 50; 
	 
    for (i=0 ; i<NUMAMMO ; i++) 
		p->maxammo[i] = maxammo[i]; 

	// We'll need to update almost everything.
	p->update |= PSF_STATE | PSF_HEALTH | PSF_ARMOR_TYPE 
		| PSF_ARMOR_POINTS | PSF_POWERS | PSF_KEYS
		| PSF_OWNED_WEAPONS	| PSF_AMMO | PSF_MAX_AMMO 
		| PSF_PENDING_WEAPON | PSF_READY_WEAPON;
}

//
// G_CheckSpot  
// Returns false if the player cannot be respawned
// at the given mapthing_t spot  
// because something is occupying it 
//

boolean
G_CheckSpot
( int		playernum,
 mapthing_t*	mthing ) 
{ 
    fixed_t		x;
    fixed_t		y; 
    subsector_t*	ss; 
    unsigned		an; 
    mobj_t*		mo; 
    int			i;
	
    if(!players[playernum].plr->mo)
    {
		// first spawn of level, before corpses
		for (i=0 ; i<playernum ; i++)
			if (players[i].plr->mo->x == mthing->x << FRACBITS
				&& players[i].plr->mo->y == mthing->y << FRACBITS)
				return false;	
			return true;
    }
	
    x = mthing->x << FRACBITS; 
    y = mthing->y << FRACBITS; 
	
    if(!P_CheckPosition(players[playernum].plr->mo, x, y) ) 
		return false; 
	
    // flush an old corpse if needed 
    if (bodyqueslot >= BODYQUESIZE) 
		P_RemoveMobj (bodyque[bodyqueslot%BODYQUESIZE]); 
    bodyque[bodyqueslot%BODYQUESIZE] = players[playernum].plr->mo; 
    bodyqueslot++; 
	
    // spawn a teleport fog 
    ss = R_PointInSubsector (x,y); 
    an = ( ANG45 * (mthing->angle/45) ) >> ANGLETOFINESHIFT; 
	
    mo = P_SpawnMobj (x+20*finecosine[an], y+20*finesine[an] 
		      , ss->sector->floorheight 
			  , MT_TFOG); 
	
    if(players[consoleplayer].plr->viewz != 1) 
		S_NetStartSound (mo, sfx_telept);	// don't start sound on first frame 
	
    return true; 
} 


//
// G_DeathMatchSpawnPlayer 
// Spawns a player at one of the random death match spots 
// called at level load and each death 
//
void G_DeathMatchSpawnPlayer (int playernum) 
{ 
    int             i,j; 
    int				selections; 
	
    selections = deathmatch_p - deathmatchstarts; 
    if (selections < 4) 
		I_Error ("Only %i deathmatch spots, 4 required", selections); 
	
    for (j=0 ; j<20 ; j++) 
    { 
		i = P_Random() % selections; 
		if (G_CheckSpot (playernum, &deathmatchstarts[i]) ) 
		{ 
			//deathmatchstarts[i].type = playernum+1; 
			P_SpawnPlayer(&deathmatchstarts[i], playernum); 
			return; 
		} 
    } 
	
    // no good spot, so the player will probably get stuck 
    P_SpawnPlayer(&playerstarts[playernum], playernum); 
} 

//
// G_DoReborn 
// 
void G_DoReborn (int playernum) 
{ 
    int                             i; 
	 
	// I can't understand why this wasn't called earlier...
	if(G_CheckDemoStatus()) return;

    if (!netgame)
    {
		// reload the level from scratch
		gameaction = ga_loadlevel;  
    }
    else 
    {
		if(IS_CLIENT)// && playernum != consoleplayer) 
		{
			// Quick rebirth. The position doesn't matter,
			// the server will correct it very soon.
			P_SpawnPlayer(&playerstarts[0], playernum);			
			return;
		}
		gi.conprintf( "G_DoReborn for %i.\n", playernum);

		// respawn at the start

		// first dissasociate the corpse 
		players[playernum].plr->mo->player = NULL;   
		 
		// spawn at random spot if in death match 
		if (deathmatch) 
		{ 
			G_DeathMatchSpawnPlayer (playernum); 
			return; 
		} 
		 
		if (G_CheckSpot (playernum, &playerstarts[playernum]) ) 
		{ 
			P_SpawnPlayer(&playerstarts[playernum], playernum); 
			return; 
		}
		
		// try to spawn at one of the other players spots 
		for (i=0 ; i<MAXPLAYERS ; i++)
		{
			if (G_CheckSpot (playernum, &playerstarts[i]) ) 
			{ 
				//playerstarts[i].type = playernum+1;	// fake as other player 
				P_SpawnPlayer(&playerstarts[i], playernum); 
				//playerstarts[i].type = i+1;		// restore 
				return; 
			}	    
			// he's going to be inside something.  Too bad.
		}
		P_SpawnPlayer (&playerstarts[playernum], playernum); 
    } 
} 
 
 
void G_ScreenShot (void) 
{ 
    gameaction = ga_screenshot; 
} 
 


// DOOM Par Times
int pars[4][10] = 
{ 
    {0}, 
    {0,30,75,120,90,165,180,180,30,165}, 
    {0,90,90,90,120,90,360,240,30,170}, 
    {0,90,45,90,150,90,90,165,30,135} 
}; 

// DOOM II Par Times
int cpars[32] =
{
    30,90,120,120,90,150,120,120,270,90,	//  1-10
    210,150,150,150,210,150,420,150,210,150,	// 11-20
    240,150,180,150,150,300,330,420,300,180,	// 21-30
    120,30					// 31-32
};
 

//
// G_DoCompleted 
//
boolean		secretexit; 
extern char*	pagename; 
 
void G_ExitLevel (void) 
{ 
    secretexit = false; 
    gameaction = ga_completed; 
} 

// Here's for the german edition.
void G_SecretExitLevel (void) 
{ 
    // IF NO WOLF3D LEVELS, NO SECRET EXIT!
    if ( (gamemode == commercial)
		&& (W_CheckNumForName("map31")<0))
		secretexit = false;
    else
		secretexit = true; 
    gameaction = ga_completed; 
} 
 
void G_DoCompleted (void) 
{ 
    int             i; 

	// Allow multilevel demos.
	//if(G_CheckDemoStatus()) return;	// Demo loop continues.

    gameaction = ga_nothing; 
	
    for (i=0 ; i<MAXPLAYERS ; i++) 
	{
		if (players[i].plr->ingame) 
		{
			G_PlayerFinishLevel (i);        // take away cards and stuff 
			// Update this client's stats.
			NetSv_SendPlayerState(i, DDSP_ALL_PLAYERS, 
				PSF_ALL_FRAGS | PSF_COUNTERS, true);
		}
	}
	
	if (automapactive) 
		AM_Stop (); 
	
	if ( gamemode != commercial)
		switch(gamemap)
	{
	  case 8:
		  gameaction = ga_victory;
		  return;
	  case 9: 
		  for (i=0 ; i<MAXPLAYERS ; i++) 
			  players[i].didsecret = true; 
		  break;
	}
	
	//#if 0  Hmmm - why?
	if ( (gamemap == 8)
		&& (gamemode != commercial) ) 
	{
		// victory 
		gameaction = ga_victory; 
		return; 
	} 
	
	if ( (gamemap == 9)
		&& (gamemode != commercial) ) 
	{
		// exit secret level 
		for (i=0; i<MAXPLAYERS; i++) 
			players[i].didsecret = true; 
	} 
	//#endif
		
	wminfo.didsecret = players[consoleplayer].didsecret; 
	wminfo.last = gamemap - 1;
	
	// wminfo.next is 0 biased, unlike gamemap
	if ( gamemode == commercial)
	{
		if (secretexit)
			switch(gamemap)
		{
		  case 15: wminfo.next = 30; break;
		  case 31: wminfo.next = 31; break;
		}
		else
			switch(gamemap)
		{
		  case 31:
		  case 32: wminfo.next = 15; break;
		  default: wminfo.next = gamemap;
		}
	}
	else
	{
		if (secretexit) 
			wminfo.next = 8; 	// go to secret level 
		else if (gamemap == 9) 
		{
			// returning from secret level 
			switch (gameepisode) 
			{ 
			case 1: 
				wminfo.next = 3; 
				break; 
			case 2: 
				wminfo.next = 5; 
				break; 
			case 3: 
				wminfo.next = 6; 
				break; 
			case 4:
				wminfo.next = 2;
				break;
			}                
		} 
		else 
			wminfo.next = gamemap;          // go to next level 
	}
	
	wminfo.maxkills = totalkills; 
	wminfo.maxitems = totalitems; 
	wminfo.maxsecret = totalsecret; 

	G_PrepareWIData();

	// Tell the clients what's going on.
	NetSv_Intermission(IMF_BEGIN, 0, 0);

	gamestate = GS_INTERMISSION; 
	viewactive = false; 
	automapactive = false; 

	if (statcopy)
		memcpy (statcopy, &wminfo, sizeof(wminfo));
	
	WI_Start (&wminfo); 
} 

void G_PrepareWIData(void)
{
	int		i;

	wminfo.epsd = gameepisode-1; 
	wminfo.maxfrags = 0; 
	if(gamemode == commercial)
		wminfo.partime = 35*cpars[gamemap-1]; 
	else
		wminfo.partime = 35*pars[gameepisode][gamemap]; 
	wminfo.pnum = consoleplayer; 
	for (i=0 ; i<MAXPLAYERS ; i++) 
	{ 
		wminfo.plyr[i].in = players[i].plr->ingame; 
		wminfo.plyr[i].skills = players[i].killcount; 
		wminfo.plyr[i].sitems = players[i].itemcount; 
		wminfo.plyr[i].ssecret = players[i].secretcount; 
		wminfo.plyr[i].stime = leveltime; 
		memcpy(wminfo.plyr[i].frags, players[i].frags, 
			sizeof(wminfo.plyr[i].frags)); 
	} 
}

//
// G_WorldDone 
//
void G_WorldDone (void) 
{ 
    gameaction = ga_worlddone; 
	
    if (secretexit) 
		players[consoleplayer].didsecret = true; 
	
    if ( gamemode == commercial )
    {
		switch (gamemap)
		{
		case 15:
		case 31:
			if (!secretexit)
				break;
		case 6:
		case 11:
		case 20:
		case 30:
			F_StartFinale ();
			break;
		}
    }
} 

void G_DoWorldDone (void) 
{        
    gamestate = GS_LEVEL; 
    gamemap = wminfo.next+1; 
    G_DoLoadLevel (); 
    gameaction = ga_nothing; 
    viewactive = true; 
} 



//
// G_InitFromSavegame
// Can be called by the startup code or the menu task. 
//
extern boolean setsizeneeded;
void R_ExecuteSetViewSize (void);

char	savename[256];

void G_LoadGame (char* name) 
{ 
    strcpy (savename, name); 
    gameaction = ga_loadgame; 
} 
 
#define VERSIONSIZE		16 


void G_DoLoadGame (void) 
{ 
    int		length; 
    int		i; 
    int		a,b,c; 
    char	vcheck[VERSIONSIZE]; 
	
    gameaction = ga_nothing; 
	
    length = M_ReadFile (savename, &savebuffer); 
    save_p = savebuffer + SAVESTRINGSIZE;
    
    // skip the description field 
    memset (vcheck,0,sizeof(vcheck)); 
    sprintf (vcheck,"version %i", SAVE_VERSION); 
    if (strcmp (save_p, vcheck)) 
	{
		int saveVer;
		sscanf(save_p, "version %i", &saveVer);
		if(saveVer >= SAVE_VERSION_BASE) 
		{
			// Must be from the wrong game.
			gi.Message( "Bad savegame version.\n");
			return;
		}
		// Just give a warning.
		gi.Message( "Savegame ID '%s': incompatible?\n", save_p);
	}
    save_p += VERSIONSIZE; 
			 
    gameskill = *save_p++; 
    gameepisode = *save_p++; 
    gamemap = *save_p++; 
    for (i=0 ; i<4 /*MAXPLAYERS*/ ; i++) 
		players[i].plr->ingame = *save_p++; 
	
    // load a base level 
    G_InitNew (gameskill, gameepisode, gamemap); 
	
    // get the times 
    a = *save_p++; 
    b = *save_p++; 
    c = *save_p++; 
    leveltime = (a<<16) + (b<<8) + c; 
	
    // dearchive all the modifications
    P_UnArchivePlayers (); 
    P_UnArchiveWorld (); 
    P_UnArchiveThinkers (); 
    P_UnArchiveSpecials (); 
	
    if (*save_p != 0x1d) 
		I_Error ("Bad savegame");
    
    // done 
    Z_Free (savebuffer); 
	savebuffer = NULL;
	
    if (setsizeneeded)
		R_ExecuteSetViewSize ();
    
    // draw the pattern into the back screen
	//    R_FillBackScreen ();   
} 
 

//
// G_SaveGame
// Called by the menu task.
// Description is a 24 byte text string 
//
void
G_SaveGame
( int	slot,
  char*	description ) 
{ 
    savegameslot = slot; 
    strcpy (savedescription, description); 
    sendsave = true; 
} 
 
void G_DoSaveGame (void) 
{ 
    char	name[100]; 
    char	name2[VERSIONSIZE]; 
    char*	description; 
    int		length; 
    int		i; 
	
    if (M_CheckParm("-cdrom"))
		sprintf(name,"c:\\doomdata\\"SAVEGAMENAME"%d.dsg",savegameslot);
    else
		sprintf (name,SAVEGAMENAME"%d.dsg",savegameslot); 
    description = savedescription; 
	
    //save_p = savebuffer = screens[1]+0x4000; 
	save_p = savebuffer = (byte*) Z_Malloc(SAVEGAMESIZE, PU_STATIC, 0);
	
    memcpy (save_p, description, SAVESTRINGSIZE); 
    save_p += SAVESTRINGSIZE; 
    memset (name2,0,sizeof(name2)); 
    sprintf (name2,"version %i", SAVE_VERSION); 
    memcpy (save_p, name2, VERSIONSIZE); 
    save_p += VERSIONSIZE; 
	
    *save_p++ = gameskill; 
    *save_p++ = gameepisode; 
    *save_p++ = gamemap; 
    for (i=0 ; i<4/*MAXPLAYERS*/ ; i++) 
		*save_p++ = players[i].plr->ingame; 
    *save_p++ = leveltime>>16; 
    *save_p++ = leveltime>>8; 
    *save_p++ = leveltime; 
	
    P_ArchivePlayers (); 
    P_ArchiveWorld (); 
    P_ArchiveThinkers (); 
    P_ArchiveSpecials (); 
	
    *save_p++ = 0x1d;		// consistancy marker 
	
    length = save_p - savebuffer; 
    if (length > SAVEGAMESIZE) 
		I_Error ("Savegame buffer overrun"); 
    M_WriteFile (name, savebuffer, length); 
    gameaction = ga_nothing; 
    savedescription[0] = 0;		 
	
    //players[consoleplayer].message = GGSAVED; 
	P_SetMessage(players + consoleplayer, GGSAVED);
	
	Z_Free(savebuffer);
	savebuffer = NULL;

    // draw the pattern into the back screen
	//    R_FillBackScreen ();	
} 
 

//
// G_InitNew
// Can be called by the startup code or the menu task,
// consoleplayer, displayplayer, playeringame[] should be set. 
//
skill_t	d_skill; 
int     d_episode; 
int     d_map; 
 
void
G_DeferedInitNew
( skill_t	skill,
  int		episode,
  int		map) 
{ 
    d_skill = skill; 
    d_episode = episode; 
    d_map = map; 
    gameaction = ga_newgame; 
} 


void G_DoNewGame (void) 
{
//	int		i;

    demoplayback = false; 
    netdemo = false;
    //netgame = false;
	if(!netgame)
	{
	    deathmatch = false;
	/*for(i=1; i<4; i++)
		players[i].plr->ingame = 0;*/
		respawnparm = false;
		fastparm = false;
		nomonsters = false;
	}
/*    gi.Set(DD_CONSOLEPLAYER, 0);
	gi.Set(DD_DISPLAYPLAYER, 0);*/
    G_InitNew(d_skill, d_episode, d_map); 
    gameaction = ga_nothing; 
} 

// The sky texture to be used instead of the F_SKY1 dummy.
extern  int	skytexture; 


void
G_InitNew
( skill_t	skill,
 int		episode,
 int		map ) 
{ 
    int             i; 
	
    if (paused) 
    { 
		paused = false; 
		S_ResumeSound (); 
    } 
	
	
    if (skill > sk_nightmare) 
		skill = sk_nightmare;
	
	
    // This was quite messy with SPECIAL and commented parts.
    // Supposedly hacks to make the latest edition work.
    // It might not work properly.
    if (episode < 1)
		episode = 1; 
	
    if ( gamemode == retail )
    {
		if (episode > 4)
			episode = 4;
    }
    else if ( gamemode == shareware )
    {
		if (episode > 1) 
			episode = 1;	// only start episode 1 on shareware
    }  
    else
    {
		if (episode > 3)
			episode = 3;
    }
  	
    if (map < 1) 
		map = 1;
    
    if ( (map > 9)
		&& ( gamemode != commercial) )
		map = 9; 
	
    M_ClearRandom (); 
	
    if (skill == sk_nightmare || respawnparm )
		respawnmonsters = true;
    else
		respawnmonsters = false;
	
    if (fastparm || (skill == sk_nightmare && gameskill != sk_nightmare) )
    { 
		for (i=S_SARG_RUN1 ; i<=S_SARG_PAIN2 ; i++) 
			states[i].tics >>= 1; 
		mobjinfo[MT_BRUISERSHOT].speed = 20*FRACUNIT; 
		mobjinfo[MT_HEADSHOT].speed = 20*FRACUNIT; 
		mobjinfo[MT_TROOPSHOT].speed = 20*FRACUNIT; 
    } 
    else if (skill != sk_nightmare && gameskill == sk_nightmare) 
    { 
		for (i=S_SARG_RUN1 ; i<=S_SARG_PAIN2 ; i++) 
			states[i].tics <<= 1; 
		mobjinfo[MT_BRUISERSHOT].speed = 15*FRACUNIT; 
		mobjinfo[MT_HEADSHOT].speed = 10*FRACUNIT; 
		mobjinfo[MT_TROOPSHOT].speed = 10*FRACUNIT; 
    } 
	
			 
    // force players to be initialized upon first level load         
    for (i=0 ; i<MAXPLAYERS ; i++) 
		players[i].playerstate = PST_REBORN; 
	
    usergame = true;                // will be set false if a demo 
    paused = false; 
    demoplayback = false; 
    automapactive = false; 
    viewactive = true; 
    gameepisode = episode; 
    gamemap = map; 
    gameskill = skill; 
	
    viewactive = true;
	
    G_DoLoadLevel (); 
} 
 

//
// DEMO RECORDING 
// 
#define DEMOMARKER		0x80


void G_ReadDemoTiccmd (player_t *pl) 
{ 
	ticcmd_t *cmd = &pl->cmd;

    if (*demo_p == DEMOMARKER || actions[DA_STOPDEMO].on) 
    {
		// end of demo data stream 
		G_CheckDemoStatus (); 
		return; 
    } 
    cmd->forwardmove = ((signed char)*demo_p++); 
    cmd->sidemove = ((signed char)*demo_p++); 

    //cmd->angleturn = ((unsigned char)*demo_p++)<<8; 
	G_AdjustAngle(pl, ((unsigned char)*demo_p++)<<8);
	G_SetCmdViewAngles(cmd, pl);
    
	cmd->buttons = (unsigned char)*demo_p++; 
} 

/*void G_ReadDemoTiccmd (player_t *pl)
{
	ticcmd_t *cmd = &pl->cmd;
	int look;

	if (*demo_p == DEMOMARKER || actions[HA_STOPDEMO].on)
	{       // end of demo data stream
		G_CheckDemoStatus ();
		return;
	}
	cmd->forwardmove = ((signed char)*demo_p++);
	cmd->sidemove = ((signed char)*demo_p++);
	G_AdjustAngle(pl, ((unsigned char)*demo_p++)<<8);
	cmd->buttons = (unsigned char)*demo_p++;
	look = (cmd->lookfly = ((unsigned char)*demo_p++)) & 15;
	cmd->arti = (unsigned char)*demo_p++;
	if(look > 7) look -= 16;
	G_AdjustLookDir(pl, look);
	G_SetCmdViewAngles(cmd, pl);
}
*/


void G_WriteDemoTiccmd (ticcmd_t* cmd) 
{ 
#if 0
	int	offset = demo_p - demobuffer;

	// Check that we have enough memory in the buffer.
	if(offset+8 > maxDemoSize)
	{
		// We need to allocate a new, larger buffer! Allocate 32 kb more.
		byte *newbuffer = Z_Malloc(maxDemoSize+=0x8000, PU_STATIC, NULL);
		// Copy the old buffer data.
		memcpy(newbuffer, demobuffer, offset);
		// Free the old buffer.
		Z_Free(demobuffer);
		demobuffer = newbuffer;
		// Update the pointers.
		demo_p = demobuffer + offset;
		
		gi.Message("Demo buffer resized to %d bytes.\n", maxDemoSize);
	}

    if (actions[DA_STOPDEMO].on)           // press q to end demo recording 
		G_CheckDemoStatus (); 
    *demo_p++ = cmd->forwardmove; 
    *demo_p++ = cmd->sidemove; 
    *demo_p++ = (cmd->angleturn+128)>>8; 
    *demo_p++ = cmd->buttons; 
    demo_p -= 4; 
	
    G_ReadDemoTiccmd (cmd);         // make SURE it is exactly the same 
#endif
} 
 
 
 
//
// G_RecordDemo 
// 
void G_RecordDemo (char* name) 
{ 
    int             i; 
	
    usergame = false; 
    strcpy (demoname, name); 
    strcat (demoname, ".lmp"); 
    maxDemoSize = 0x20000;
    i = M_CheckParm ("-maxdemo");
    if(i && i<myargc-1)
		maxDemoSize = atoi(gi.Argv(i+1))*1024;
    demobuffer = Z_Malloc (maxDemoSize,PU_STATIC,NULL); 
	
    demorecording = true; 
} 
 
 
void G_BeginRecording (void) 
{ 
    int             i; 
		
    demo_p = demobuffer;
	
    *demo_p++ = VERSION;
    *demo_p++ = gameskill; 
    *demo_p++ = gameepisode; 
    *demo_p++ = gamemap; 
    *demo_p++ = deathmatch; 
    *demo_p++ = respawnparm;
    *demo_p++ = fastparm;
    *demo_p++ = nomonsters;
    *demo_p++ = consoleplayer;
	 
    for (i=0 ; i<4/*MAXPLAYERS*/ ; i++) 
		*demo_p++ = players[i].plr->ingame; 		 
} 
 

//
// G_PlayDemo 
//

char*	defdemoname; 
 
void G_DeferedPlayDemo (char* name) 
{ 
    defdemoname = name; 
    gameaction = ga_playdemo; 
} 
 
void G_DoPlayDemo (void) 
{ 
    skill_t	skill; 
    int		i, episode, map; 
	byte	demoversion;
	
    gameaction = ga_nothing; 
    demobuffer = demo_p = W_CacheLumpName (defdemoname, PU_STATIC); 
	demoversion = *demo_p++;
	
	gi.Message("Demo '%s' from version %d.\n", defdemoname, demoversion);

    /*if(*demo_p++ != VERSION)
    {
		gi.Message( "Demo is from a different game version!\n");
		gameaction = ga_nothing;
		return;
    }*/
    
    skill = *demo_p++; 
    episode = *demo_p++; 
    map = *demo_p++; 
    deathmatch = *demo_p++;
    respawnparm = *demo_p++;
    fastparm = *demo_p++;
    nomonsters = *demo_p++;
    gi.Set(DD_CONSOLEPLAYER, *demo_p++);
	
    for (i=0 ; i<4/*MAXPLAYERS*/ ; i++) 
		players[i].plr->ingame = *demo_p++; 
    if (players[1].plr->ingame) 
    { 
		//netgame = true; 
		netdemo = true; 
    }
	
    // don't spend a lot of time in loadlevel 
    precache = false;
    G_InitNew (skill, episode, map); 
    precache = true; 
	
    usergame = false; 
    demoplayback = true; 

	players[consoleplayer].plr->flags &= ~DDPF_FIXANGLES;
} 

//
// G_TimeDemo 
//
void G_TimeDemo (char* name) 
{ 	 
    nodrawers = M_CheckParm ("-nodraw"); 
    noblit = M_CheckParm ("-noblit"); 
    timingdemo = true; 
    singletics = true; 
	
    defdemoname = name; 
    gameaction = ga_playdemo; 
} 
 
 
/* 
=================== 
= 
= G_CheckDemoStatus 
= 
= Called after a death or level completion to allow demos to be cleaned up 
= Returns true if a new demo loop action will take place 
=================== 
*/ 
 
boolean G_CheckDemoStatus (void) 
{ 
    int             i, endtime; 
	
	actions[DA_STOPDEMO].on = false;

    if (timingdemo) 
    { 
		endtime = I_GetTime (); 
		I_Error ("timed %i gametics in %i realtics",gametic 
			, endtime-starttime); 
    } 
	
    if (demoplayback) 
    { 
		if (singledemo) 
			I_Quit (); 
		
		Z_ChangeTag (demobuffer, PU_CACHE); 
		demoplayback = false; 
		netdemo = false;
		//	netgame = false;
		deathmatch = false;
		for(i=1; i<4; i++) players[i].plr->ingame = 0;
		respawnparm = false;
		fastparm = false;
		nomonsters = false;
		gi.Set(DD_CONSOLEPLAYER, 0);
		D_AdvanceDemo (); 
		return true; 
    } 
	
    if(demorecording) 
    { 
		*demo_p++ = DEMOMARKER; 
		M_WriteFile (demoname, demobuffer, demo_p - demobuffer); 
		Z_Free (demobuffer); 
		demorecording = false; 
		gi.Message("Demo %s recorded.\n",demoname); 
    } 
    return false; 
} 
 
 
 
