// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	Cheat sequence checking.
//
//-----------------------------------------------------------------------------


static const char
rcsid[] = "$Id: m_cheat.c,v 1.1 1997/02/03 21:24:34 b1 Exp $";

#include "m_cheat.h"
#include "s_sound.h"
#include "sounds.h"
#include "dstrings.h"
#include "g_game.h"
#include "p_inter.h"
#include "p_local.h"

// Dimensions given in characters.
#define ST_MSGWIDTH			52

//
// CHEAT SEQUENCE PACKAGE
//

static int		firsttime = 1;
static unsigned char	cheat_xlate_table[256];


//
// Called in st_stuff module, which handles the input.
// Returns a 1 if the cheat was successful, 0 if failed.
//
int
cht_CheckCheat
( cheatseq_t*	cht,
 char		key )
{
    int i;
    int rc = 0;
	
    if (firsttime)
    {
		firsttime = 0;
		for (i=0;i<256;i++) cheat_xlate_table[i] = SCRAMBLE(i);
    }
	
    if (!cht->p)
		cht->p = cht->sequence; // initialize if first time
	
    if (*cht->p == 0)
		*(cht->p++) = key;
    else if
		(cheat_xlate_table[(unsigned char)key] == *cht->p) cht->p++;
    else
		cht->p = cht->sequence;
	
    if (*cht->p == 1)
		cht->p++;
    else if (*cht->p == 0xff) // end of sequence character
    {
		cht->p = cht->sequence;
		rc = 1;
    }
	
    return rc;
}

void
cht_GetParam
( cheatseq_t*	cht,
 char*		buffer )
{
	
    unsigned char *p, c;
	
    p = cht->sequence;
    while (*(p++) != 1);
    
    do
    {
		c = *p;
		*(buffer++) = c;
		*(p++) = 0;
    }
    while (c && *p!=0xff );
	
    if (*p==0xff)
		*buffer = 0;
	
}


void cht_GodFunc(player_t *plyr)
{
	plyr->cheats ^= CF_GODMODE;
	if (plyr->cheats & CF_GODMODE)
	{
		if (plyr->plr->mo) plyr->plr->mo->health = 100;
		plyr->health = 100;
	}
	P_SetMessage(plyr, (plyr->cheats & CF_GODMODE)? 
		STSTR_DQDON : STSTR_DQDOFF);
}

void cht_GiveFunc(player_t *plyr, boolean weapons, boolean ammo,
				  boolean armor, boolean cards)
{
	int		i;

	if(armor)
	{
		plyr->armorpoints = 200;
		plyr->armortype = 2;
	}
	if(weapons)
	{
		for (i=0;i<NUMWEAPONS;i++)
			plyr->weaponowned[i] = true;
	}
	if(ammo)
	{
		for (i=0;i<NUMAMMO;i++)
			plyr->ammo[i] = plyr->maxammo[i];
	}
	if(cards)
	{
		for (i=0;i<NUMCARDS;i++)
			plyr->cards[i] = true;
	}
}

void cht_MusicFunc(player_t *plyr, char *buf)
{
	int	off, musnum;

	if (gamemode == commercial)
	{
		off = (buf[0]-'0')*10 + buf[1]-'0';
		musnum = mus_runnin + off - 1;
		if (off < 1 || off > 35)
			P_SetMessage(plyr, STSTR_NOMUS);
		else
			S_ChangeMusic(musnum, 1);
	}
	else
	{
		off = (buf[0]-'1')*9 + (buf[1]-'1');
		musnum = mus_e1m1 + off;
		if (off > 31)
			P_SetMessage(plyr, STSTR_NOMUS);
		else
			S_ChangeMusic(musnum, 1);
	}
}

void cht_NoClipFunc(player_t *plyr)
{
	plyr->cheats ^= CF_NOCLIP;
	P_SetMessage(plyr, (plyr->cheats & CF_NOCLIP)? STSTR_NCON : STSTR_NCOFF);
}

boolean cht_WarpFunc(player_t *plyr, char *buf)
{
	int		epsd, map;

	if (gamemode == commercial)
	{
		epsd = 1;
		map = (buf[0] - '0')*10 + buf[1] - '0';
	}
	else
	{
		epsd = buf[0] - '0';
		map = buf[1] - '0';
	}
	
	// Catch invalid maps.
	if (epsd < 1)
		return false;
	
	if (map < 1)
		return false;
	
	// Ohmygod - this is not going to work.
	if ((gamemode == retail)
		&& ((epsd > 4) || (map > 9)))
		return false;
	
	if ((gamemode == registered)
		&& ((epsd > 3) || (map > 9)))
		return false;
	
	if ((gamemode == shareware)
		&& ((epsd > 1) || (map > 9)))
		return false;
	
	if ((gamemode == commercial)
		&& (( epsd > 1) || (map > 32)))
		return false;
	
	// So be it.
	P_SetMessage(plyr, STSTR_CLEV);
	G_DeferedInitNew(gameskill, epsd, map);
	return true;
}

void cht_PowerUpFunc(player_t *plyr, int i)
{
	if (!plyr->powers[i])
		P_GivePower( plyr, i);
	else if (i!=pw_strength)
		plyr->powers[i] = 1;
	else
		plyr->powers[i] = 0;
}

void cht_ChoppersFunc(player_t *plyr)
{
	plyr->weaponowned[wp_chainsaw] = true;
	plyr->powers[pw_invulnerability] = true;
}

void cht_PosFunc(player_t *plyr)
{
	static char	buf[ST_MSGWIDTH];

	sprintf(buf, "ang=0x%x;x,y=(0x%x,0x%x)",
		players[consoleplayer].plr->mo->angle,
		players[consoleplayer].plr->mo->x,
		players[consoleplayer].plr->mo->y);
	P_SetMessage(plyr, buf);
}


