#include "DoomDef.h"
#include "R_local.h"
#include "P_local.h"
#include "i_sound.h"
#include "soundst.h"
#include "settings.h"

boolean S_StopSoundID(int sound_id, int priority);

/*
===============================================================================

		MUSIC & SFX API

===============================================================================
*/

#define DDVOL(x)				((1000*x)/127)
#define DDPAN(x)				((500*x)/128)
#define DDPITCH(x)				((1000*x)/127)
#define SOUNDCURVE(dist,vol)	((soundCurve[dist]*vol)/255)

enum { VX, VY, VZ };

//static channel_t channel[MAX_CHANNELS];

static	channel_t	*Channel = 0;
static	int			numChannels = 0;


static int rs; //the current registered song.
int mus_song = -1;
int mus_lumpnum;
void *mus_sndptr;
byte *soundCurve;
int s_CDTrack = 0;

extern sfxinfo_t S_sfx[];
extern musicinfo_t S_music[];

//int snd_Channels = 8;

/*boolean		reverbDebug = false;
int			snd_3D = 0;*/
int			snd_MaxChannels = 20;
//float		snd_ReverbFactor = 1;

extern int startepisode;
extern int startmap;

int AmbChan;
sector_t *listenerSector = NULL;


void S_StopChannel(channel_t *chan)
{
	if(chan->handle)
	{
		gi.StopSound(chan->handle);
		if(S_sfx[chan->sound_id].usefulness > 0)
			S_sfx[chan->sound_id].usefulness--;
		memset(chan, 0, sizeof(*chan));
	}
}

void S_Start(void)
{
	int i;

	S_StartSong((gameepisode-1)*9 + gamemap-1, true);

	//stop all sounds
	for(i=0; i < numChannels; i++)
	{
		if(Channel[i].handle)
		{
/*			gi.StopSound(Channel[i].handle);
			if(S_sfx[Channel[i].sound_id].usefulness > 0)
			{
				S_sfx[Channel[i].sound_id].usefulness--;
			}
			Channel[i].handle = 0;
			Channel[i].mo = NULL;*/
			S_StopChannel(Channel+i);
			if(AmbChan == i)
			{
				AmbChan = -1;
			}
		}
	}
	// Reset all channels.
	memset(Channel, 0, numChannels*sizeof(channel_t));
	listenerSector = NULL;
}

void S_StartSong(int song, boolean loop)
{
	if(song == mus_song)
	{ // don't replay an old song
		return;
	}
	if(rs)
	{
		gi.StopSong();
		gi.W_ChangeCacheTag(mus_lumpnum, PU_CACHE);		
	}
	if(song < mus_e1m1 || song > NUMMUSIC)
	{
		return;
	}
	mus_lumpnum = W_GetNumForName(S_music[song].name);
	mus_sndptr = W_CacheLumpNum(mus_lumpnum, PU_MUSIC);
	gi.PlaySong(mus_sndptr, W_LumpLength(mus_lumpnum), true);
	mus_song = song;
}

int S_GetSfxLumpNum(sfxinfo_t *sound)
{
	if(sound->name == 0)
		return 0;
	if (sound->link) sound = sound->link;
	return W_GetNumForName(sound->name);
}

// Get a free channel, but also get rid of stopped channels.
// If for_mobj is NULL, any free channel will do. Otherwise, an earlier
// channel owned by for_mobj is used.
channel_t *S_GetFreeChannel(mobj_t *for_mobj)
{
	int			i;
	channel_t	*chan = NULL;

	if(for_mobj && !for_mobj->thinker.function) for_mobj = NULL;
	// Check through all the channels.
	for(i=0; i<numChannels; i++)
	{
		// Is this a free channel?
		if(!chan && !gi.SoundIsPlaying(Channel[i].handle))
		{
			// This will be used if a clear channel is needed.
			chan = Channel + i;
			if(!for_mobj) break; // This'll do.
		}
		if(for_mobj && Channel[i].mo == for_mobj)
		{
			// Mobjs are allowed only one channel.
			chan = Channel + i;
			break;
		}
	}
	// No suitable channel found, and there already are as much channels as there can be?
	if(!chan && numChannels >= snd_MaxChannels)
	{
		// Pick one at random, then.
		chan = Channel + M_Random() % numChannels;
	}
	// Was a suitable channel found?
	if(chan)
	{
		S_StopChannel(chan);
		return chan;
	}	
	// We need to allocate a new channel.
	Channel = realloc(Channel, sizeof(channel_t) * (++numChannels));
	chan = Channel + numChannels-1;
	memset(chan, 0, sizeof(channel_t));
	return chan;
}

int testcvar = 0;

// Fills in the position and velocity.
void S_FillSound3D(mobj_t *mo, sound3d_t *desc)
{
	desc->flags |= DDSOUNDF_POS;
	desc->pos[VX] = mo->x;
	desc->pos[VY] = mo->z;
	desc->pos[VZ] = mo->y;

	if(mo->thinker.function) desc->pos[VY] += mo->height/2;
		
	if(mo == players[displayplayer].plr->mo)
	{
		int tableAngle = mo->angle >> ANGLETOFINESHIFT;
		int xoff = finecosine[tableAngle] * 64, yoff = finesine[tableAngle] * 64;
		desc->pos[VX] += xoff;
		desc->pos[VZ] += yoff;
		//gi.Message( "x:%i y:%i\n", desc->pos[VX]>>16, desc->pos[VZ]>>16);
	}
	
	if(mo->thinker.function)
	{
		desc->flags |= DDSOUNDF_MOV;
		desc->mov[VX] = mo->momx * 35;
		desc->mov[VY] = mo->momz * 35;
		desc->mov[VZ] = mo->momy * 35;
	}
}

// Return value is in range 0-255. 0 means the sound is coming from the left.
// 128 is the center.
static int CalcSep(mobj_t *listener, mobj_t *sound, unsigned int dirangle)
{
	unsigned int angle = R_PointToAngle2(listener->x, listener->y, sound->x, sound->y);
	int sep;
	
	sep = (int)(angle>>24) - (int)(dirangle>>24);
	if(sep > 128) 
		sep -= 256;
	else if(sep < -128)
		sep += 256;
	sep = 128 - sep*2;
	if(sep > 256) sep = 512-sep;
	else if(sep < 0) sep = -sep;
	return sep;
}

#if 0
void S_StartSound(mobj_t *origin, int sound_id)
{
	int dist = 0, vol;
	int i;
	int priority;
	int sep;
	int angle;
	int absx;
	int absy;
	mobj_t *plrmo = players[displayplayer].plr->mo;

	static int sndcount = 0;
	int chan;

	if(sound_id==0 || snd_MaxVolume == 0)
		return;
	if(origin == NULL)
	{
		origin = plrmo;
	}

// calculate the distance before other stuff so that we can throw out
// sounds that are beyond the hearing range.
	if(origin && plrmo)
	{
		absx = abs(origin->x - plrmo->x);
		absy = abs(origin->y - plrmo->y);
		dist = absx+absy-(absx > absy ? absy>>1 : absx>>1);
		dist >>= FRACBITS;
	
		if(dist >= MAX_SND_DIST)
		{
			return; //sound is beyond the hearing range...
		}
		if(dist < 0)
		{
			dist = 0;
		}
	}
	priority = S_sfx[sound_id].priority;
	priority *= (10 - (dist/160));
	if(!S_StopSoundID(sound_id, priority))
	{
		return; // other sounds have greater priority
	}
	for(i=0; origin && i<snd_Channels; i++)
	{
		if(origin->player)
		{
			i = snd_Channels;
			break; // let the player have more than one sound.
		}
		if(origin == channel[i].mo)
		{ // only allow other mobjs one sound
			S_StopSound(channel[i].mo);
			break;
		}
	}
	if(i >= snd_Channels)
	{
		if(sound_id >= sfx_wind)
		{
			if(AmbChan != -1 && S_sfx[sound_id].priority <=
				S_sfx[channel[AmbChan].sound_id].priority)
			{
				return; //ambient channel already in use
			}
			else
			{
				AmbChan = -1;
			}
		}
		for(i=0; i<snd_Channels; i++)
		{
			if(channel[i].mo == NULL)
			{
				break;
			}
		}
		if(i >= snd_Channels)
		{
			//look for a lower priority sound to replace.
			sndcount++;
			if(sndcount >= snd_Channels)
			{
				sndcount = 0;
			}
			for(chan=0; chan < snd_Channels; chan++)
			{
				i = (sndcount+chan)%snd_Channels;
				if(priority >= channel[i].priority)
				{
					chan = -1; //denote that sound should be replaced.
					break;
				}
			}
			if(chan != -1)
			{
				return; //no free channels.
			}
			else //replace the lower priority sound.
			{
				if(channel[i].handle)
				{
					if(gi.SoundIsPlaying(channel[i].handle))
					{
						gi.StopSound(channel[i].handle);
					}
					if(S_sfx[channel[i].sound_id].usefulness > 0)
					{
						S_sfx[channel[i].sound_id].usefulness--;
					}

					if(AmbChan == i)
					{
						AmbChan = -1;
					}
				}
			}
		}
	}
	if(S_sfx[sound_id].lumpnum == 0)
	{
		S_sfx[sound_id].lumpnum = I_GetSfxLumpNum(&S_sfx[sound_id]);
	}
	if(S_sfx[sound_id].snd_ptr == NULL)
	{
		S_sfx[sound_id].snd_ptr = W_CacheLumpNum(S_sfx[sound_id].lumpnum,
			PU_SOUND);
	}

	// calculate the volume based upon the distance from the sound origin.
//      vol = (snd_MaxVolume*16 + dist*(-snd_MaxVolume*16)/MAX_SND_DIST)>>9;
	vol = soundCurve[dist];

	if(origin == plrmo || !origin || !plrmo)
	{
		sep = 128;
	}
	else
	{
		angle = R_PointToAngle2(plrmo->x, plrmo->y, origin->x, origin->y);
		angle = (angle-viewangle)>>24;
		sep = angle*2-128;
		if(sep < 64)
			sep = -sep;
		if(sep > 192)
			sep = 512-sep;
	}

	channel[i].pitch = (byte)(127+(M_Random()&7)-(M_Random()&7));
	channel[i].handle = gi.PlaySound(S_sfx[sound_id].snd_ptr, DDVOL(vol), DDPAN(sep), 
		DDPITCH(channel[i].pitch));
	channel[i].mo = origin;
	channel[i].sound_id = sound_id;
	channel[i].priority = priority;
	if(sound_id >= sfx_wind)
	{
		AmbChan = i;
	}
	if(S_sfx[sound_id].usefulness == -1)
	{
		S_sfx[sound_id].usefulness = 1;
	}
	else
	{
		S_sfx[sound_id].usefulness++;
	}
}

void S_StartSoundAtVolume(mobj_t *origin, int sound_id, int volume)
{
//	int dist;
	int i;
//	int sep;
	mobj_t *plrmo = players[displayplayer].plr->mo;

	static int sndcount;
//	int chan;

	if(sound_id == 0 || snd_MaxVolume == 0)
		return;
	if(origin == NULL)
	{
		origin = plrmo;
	}

	//gi.Message( "S_StartSoundAtVolume: origin:%p\n", origin);

	if(volume == 0)
	{
		return;
	}
	volume = (volume*(snd_MaxVolume+1)*8)>>7;

// no priority checking, as ambient sounds would be the LOWEST.
	for(i=0; i<snd_Channels; i++)
	{
		if(channel[i].mo == NULL)
		{
			break;
		}
	}
	if(i >= snd_Channels)
	{
		return;
	}
	if(S_sfx[sound_id].lumpnum == 0)
	{
		S_sfx[sound_id].lumpnum = I_GetSfxLumpNum(&S_sfx[sound_id]);
	}
	if(S_sfx[sound_id].snd_ptr == NULL)
	{
		S_sfx[sound_id].snd_ptr = W_CacheLumpNum(S_sfx[sound_id].lumpnum,
			PU_SOUND);
		/*#ifdef __WATCOMC__
		_dpmi_lockregion(S_sfx[sound_id].snd_ptr,
			lumpinfo[S_sfx[sound_id].lumpnum].size);
		#endif*/
	}
	channel[i].pitch = (byte)(127-(M_Random()&3)+(M_Random()&3));
	channel[i].handle = gi.PlaySound(S_sfx[sound_id].snd_ptr, 
		DDVOL(volume), DDPAN(128), DDPITCH(channel[i].pitch));
	channel[i].mo = origin;
	channel[i].sound_id = sound_id;
	channel[i].priority = 1; //super low priority.
	if(S_sfx[sound_id].usefulness == -1)
	{
		S_sfx[sound_id].usefulness = 1;
	}
	else
	{
		S_sfx[sound_id].usefulness++;
	}
	//gi.Message( "S_StartSoundAtVolume: ends!\n", origin);
}
#endif

// Broadcasts the sound to everybody.
void S_NetStartSound(mobj_t *origin, int sound_id)
{
	S_StartSound(origin, sound_id);
	NetSv_Sound(origin, sound_id, 0);
}

// Broadcasts the sound to everybody except 'not_to_player'.
void S_NetAvoidStartSound(mobj_t *origin, int sound_id, player_t *not_to_player)
{
	S_StartSound(origin, sound_id);
	NetSv_Sound(origin, sound_id, NSSF_NOT | (not_to_player-players));
}

void S_StartSectorSound(sector_t *sector, int sound_id)
{
	S_StartSound( (mobj_t*) &sector->soundorg, sound_id);
	NetSv_SectorSound(sector, sound_id);
}

void S_StartSound(mobj_t *origin, int sound_id)
{
	S_StartSoundAtVolume(origin, sound_id, 127);
}

void S_StartSoundAtVolume(mobj_t *origin, int sound_id, int volume)
{
	int dist;
	int absx;
	int absy;
	int vol;
	int sep;
	mobj_t *plrmo = players[displayplayer].plr->mo;

	static int sndcount = 0;
	channel_t *channel;

	if(sound_id == 0 || gi.Get(DD_SFX_VOLUME) == 0)
		return;

	if(volume == 0)	return;

	//if(origin == plrmo) origin = NULL;

	/*gi.Message( "Play %s, origin %p, vol %i, use %i.\n", S_sfx[sound_id].name, origin, volume,
		S_sfx[sound_id].usefulness);*/

	// calculate the distance before other stuff so that we can throw out
	// sounds that are beyond the hearing range.
	if(volume >= 255) // A very loud sound?
	{
		dist = 0;	// Might as well be; the sound can be heard from a great distance.
	}
	else if(plrmo && origin)
	{
		absx = abs(origin->x - plrmo->x);
		absy = abs(origin->y - plrmo->y);
		dist = absx+absy-(absx > absy ? absy>>1 : absx>>1);
		dist >>= FRACBITS;
		if(dist >= MAX_SND_DIST)
		{
			return; // sound is beyond the hearing range...
		}
		if(dist < 0) dist = 0;
	}
	else
	{
		// The player has no mo! We're probably in the start screen.
		origin = NULL;	
		dist = 0;		// The sound is very close.
	}

	// We won't play the same sound too many times.
	if(S_sfx[sound_id].usefulness >= 3 && origin && origin != plrmo)
	{
		// We'll replace the most distant of the existing sounds.
		int	i, maxdist = 0;
		for(i=0; i<numChannels; i++)
		{
			if(Channel[i].sound_id == sound_id && Channel[i].priority >= maxdist)
			{
				maxdist = Channel[i].priority;
				channel = Channel + i;
			}
		}
		if(dist > maxdist) return; // Don't replace a close sound with a far one.
		gi.StopSound(channel->handle);
	}
	else
	{
		if(S_sfx[sound_id].lumpnum == 0)
		{
			S_sfx[sound_id].lumpnum = S_GetSfxLumpNum(&S_sfx[sound_id]);
		}
		if(S_sfx[sound_id].snd_ptr == NULL)
		{
			S_sfx[sound_id].snd_ptr = W_CacheLumpNum(S_sfx[sound_id].lumpnum,PU_SOUND);
		}
		if(S_sfx[sound_id].usefulness < 0)
		{
			S_sfx[sound_id].usefulness = 1;
		}
		else
		{
			S_sfx[sound_id].usefulness++;
		}		
		// The player can have any number of sounds, but other mobjs only one.
		channel = S_GetFreeChannel(origin == plrmo? NULL : origin);
	}
	channel->mo = origin;
	channel->volume = volume;
	if(channel->volume > 127) channel->volume = 127;
	channel->pitch = (byte) (127+(M_Random()&7)-(M_Random()&7));
	channel->sound_id = sound_id;
	channel->priority = dist;

	if(cfg.snd_3D && origin) // Play the sound in 3D?
	{
		sound3d_t desc;
		desc.flags = DDSOUNDF_VOLUME | DDSOUNDF_PITCH;
		// Very loud sounds have virtually no rolloff.
		if(volume >= 255) desc.flags |= DDSOUNDF_VERY_LOUD;
		desc.volume = (channel->volume*1000)/127;
		desc.pitch = (channel->pitch*1000)/128;
		S_FillSound3D(origin, &desc);

		if(!channel->handle)
			channel->handle = gi.Play3DSound(S_sfx[sound_id].snd_ptr, &desc);
		else
			gi.Update3DSound(channel->handle, &desc);
	}
	else // Play the sound in 2D.
	{
		if(!origin || !plrmo || origin == plrmo)
		{
			sep = 128;
			vol = channel->volume;
		}
		else
		{
			vol = SOUNDCURVE(dist, channel->volume); //soundCurve[dist];
			sep = CalcSep(plrmo, origin, gi.Get(DD_VIEWANGLE));
		}
		if(!channel->handle)
		{
			channel->handle = gi.PlaySound(S_sfx[sound_id].snd_ptr, (vol*1000)/127,
				(sep*500)/128, (channel->pitch*1000)/128);
		}
		else
		{
			gi.UpdateSound(channel->handle, (vol*1000)/127, (sep*500)/128, 
				(channel->pitch*1000)/128);
		}
	}

#ifdef PRINT_DEBUG_INFO
	printf( "PLAY:   i:%d handle:%d volume:%3d vol:%3d lump:%8s mo:%p\n", i, Channel[i].handle, 
		Channel[i].volume, vol, S_sfx[Channel[i].sound_id].lumpname, Channel[i].mo);
#endif
}

#if 0
boolean S_StopSoundID(int sound_id, int priority)
{
	int i;
	int lp; //least priority
	int found;

	if(S_sfx[sound_id].numchannels == -1)
	{
		return(true);
	}
	lp = -1; //denote the argument sound_id
	found = 0;
	for(i=0; i<numChannels; i++)
	{
		if(Channel[i].sound_id == sound_id && Channel[i].mo)
		{
			found++; //found one.  Now, should we replace it??
			if(priority >= channel[i].priority)
			{ // if we're gonna kill one, then this'll be it
				lp = i;
				priority = channel[i].priority;
			}
		}
	}
	if(found < S_sfx[sound_id].numchannels)
	{
		return(true);
	}
	else if(lp == -1)
	{
		return(false); // don't replace any sounds
	}
	if(channel[lp].handle)
	{
		if(gi.SoundIsPlaying(channel[lp].handle))
		{
			gi.StopSound(channel[lp].handle);
		}
		if(S_sfx[channel[i].sound_id].usefulness > 0)
		{
			S_sfx[channel[i].sound_id].usefulness--;
		}
		channel[lp].mo = NULL;
	}
	return(true);
}
#endif

void S_StopSound(mobj_t *origin)
{
	int i;

	for(i=0;i<numChannels;i++)
	{
		if(Channel[i].mo == origin)
		{
/*			gi.StopSound(channel[i].handle);
			if(S_sfx[channel[i].sound_id].usefulness > 0)
			{
				S_sfx[channel[i].sound_id].usefulness--;
			}
			channel[i].handle = 0;
			channel[i].mo = NULL;*/
			S_StopChannel(&Channel[i]);
			if(AmbChan == i)
			{
				AmbChan = -1;
			}
		}
	}
}

void S_SoundLink(mobj_t *oldactor, mobj_t *newactor)
{
	int i;

	for(i=0;i<numChannels;i++)
	{
		if(Channel[i].mo == oldactor)
			Channel[i].mo = newactor;
	}
}

void S_PauseSound(void)
{
	gi.PauseSong();
}

void S_ResumeSound(void)
{
	gi.ResumeSong();
}

static int nextcleanup;

void S_UpdateSounds(mobj_t *listener)
{
	int i;
	int dist, vol;
	int sep;
	int absx;
	int absy;
	
	if(!listener || gi.Get(DD_SFX_VOLUME) == 0) return;

	testcvar += ANGLE_MAX / 140;

	// Update the listener first.
	if(cfg.snd_3D)
	{
		listener3d_t lis;
		lis.flags = DDLISTENERF_POS | DDLISTENERF_MOV | DDLISTENERF_YAW | DDLISTENERF_PITCH;
		lis.pos[VX] = listener->x;
		lis.pos[VY] = listener->z + listener->height - (5<<FRACBITS);
		lis.pos[VZ] = listener->y;
		lis.mov[VX] = listener->momx * 35;
		lis.mov[VY] = listener->momz * 35;
		lis.mov[VZ] = listener->momy * 35;
		lis.yaw = -(listener->angle / (float) ANGLE_MAX * 360 - 90);
		lis.pitch = listener->player? LOOKDIR2DEG(listener->player->plr->lookdir) : 0;
		// If the sector changes, so does the reverb.
		if(listener->subsector->sector != listenerSector && cfg.snd_ReverbFactor > 0)
		{
			listenerSector = listener->subsector->sector;
			lis.flags |= DDLISTENERF_SET_REVERB;
			lis.reverb.space = listenerSector->reverb[SRD_SPACE];
			lis.reverb.decay = listenerSector->reverb[SRD_DECAY];
			lis.reverb.volume = listenerSector->reverb[SRD_VOLUME] * cfg.snd_ReverbFactor;
			lis.reverb.damping = listenerSector->reverb[SRD_DAMPING];
			if(cfg.reverbDebug)
			{
				gi.Message( "Sec %i: s:%.2f dc:%.2f v:%.2f dm:%.2f\n", 
					listenerSector-sectors, lis.reverb.space, lis.reverb.decay, lis.reverb.volume, lis.reverb.damping);
			}
		}
		if(cfg.snd_ReverbFactor == 0 && listenerSector)
		{
			listenerSector = NULL;
			lis.flags |= DDLISTENERF_DISABLE_REVERB;
		}
		gi.UpdateListener(&lis);
	}

	if(nextcleanup < gametic)
	{
		for(i = 0; i < NUMSFX; i++)
		{
			if(S_sfx[i].usefulness == 0 && S_sfx[i].snd_ptr)
			{
				gi.W_ChangeCacheTag(S_sfx[i].lumpnum, PU_CACHE);
				S_sfx[i].usefulness = -1;
				S_sfx[i].snd_ptr = NULL;
			}
		}
		nextcleanup = gametic + 35*30; // every 30 seconds
	}
	for(i=0; i<numChannels; i++)
	{
		if(!Channel[i].handle || !Channel[i].sound_id) continue; // Empty channel.

		// Take care of stopped sounds.
		if(!gi.SoundIsPlaying(Channel[i].handle))
		{
			S_StopChannel(Channel + i);
			continue;
		}

		// Does this sound need updating?
		if(Channel[i].mo == NULL)
		{
			// No, apparently not.
			continue;
		}
		else
		{
			// The sound has a source.
			if(cfg.snd_3D)
			{
				sound3d_t desc;
				desc.flags = 0;
				// Fill in position and velocity.
				S_FillSound3D(Channel[i].mo, &desc);
				gi.Update3DSound(Channel[i].handle, &desc);
			}
			else // 2D mode.
			{
				absx = abs(Channel[i].mo->x - listener->x);
				absy = abs(Channel[i].mo->y - listener->y);
				dist = absx+absy-(absx > absy ? absy>>1 : absx>>1);
				dist >>= FRACBITS;
				if(dist >= MAX_SND_DIST)
				{
					S_StopSound(Channel[i].mo);
					continue;
				}
				if(dist < 0) dist = 0;
				//vol = (SoundCurve[dist]*(15*8)*Channel[i].volume)>>14;
				//vol = soundCurve[dist];
				vol = SOUNDCURVE(dist, Channel[i].volume);
				if(Channel[i].mo == listener)
				{
					sep = 128;
				}
				else
				{
					sep = CalcSep(listener, Channel[i].mo, gi.Get(DD_VIEWANGLE));
				}
				gi.UpdateSound(Channel[i].handle, DDVOL(vol), DDPAN(sep), DDPITCH(Channel[i].pitch));
			}
		}
	}
}

#if 0
void S_UpdateSounds(mobj_t *listener)
{
	int i, dist, vol;
	int angle;
	int sep;
	int priority;
	int absx;
	int absy;

	listener = players[displayplayer].plr->mo;
	if(!listener) return;
	if(snd_MaxVolume == 0)
	{
		return;
	}
	if(nextcleanup < gametic)
	{
		for(i=0; i < NUMSFX; i++)
		{
			if(S_sfx[i].usefulness == 0 && S_sfx[i].snd_ptr)
			{
				gi.W_ChangeCacheTag(S_sfx[i].lumpnum, PU_CACHE);
				S_sfx[i].usefulness = -1;
				S_sfx[i].snd_ptr = NULL;
			}
		}
		nextcleanup = gametic+35; //CLEANUP DEBUG cleans every second
	}
	for(i=0;i<snd_Channels;i++)
	{
		if(!channel[i].handle || S_sfx[channel[i].sound_id].usefulness == -1)
		{
			continue;
		}
		if(!gi.SoundIsPlaying(channel[i].handle))
		{
			if(S_sfx[channel[i].sound_id].usefulness > 0)
			{
				S_sfx[channel[i].sound_id].usefulness--;
			}
			channel[i].handle = 0;
			channel[i].mo = NULL;
			channel[i].sound_id = 0;
			if(AmbChan == i)
			{
				AmbChan = -1;
			}
		}
		if(channel[i].mo == NULL || channel[i].sound_id == 0
			|| channel[i].mo == listener)
		{
			continue;
		}
		else
		{
			absx = abs(channel[i].mo->x - listener->x);
			absy = abs(channel[i].mo->y - listener->y);
			dist = absx+absy-(absx > absy ? absy>>1 : absx>>1);
			dist >>= FRACBITS;
//          dist = P_AproxDistance(channel[i].mo->x-listener->x, channel[i].mo->y-listener->y)>>FRACBITS;

			if(dist >= MAX_SND_DIST)
			{
				S_StopSound(channel[i].mo);
				continue;
			}
			if(dist < 0)
				dist = 0;

// calculate the volume based upon the distance from the sound origin.
//          vol = (*((byte *)W_CacheLumpName("SNDCURVE", PU_CACHE)+dist)*(snd_MaxVolume*8))>>7;
			vol = soundCurve[dist];

			angle = R_PointToAngle2(listener->x, listener->y, channel[i].mo->x, channel[i].mo->y);
			angle = (angle-viewangle)>>24;
			sep = angle*2-128;
			if(sep < 64)
				sep = -sep;
			if(sep > 192)
				sep = 512-sep;
			gi.UpdateSound(channel[i].handle, DDVOL(vol), DDPAN(sep), DDPITCH(channel[i].pitch));
			priority = S_sfx[channel[i].sound_id].priority;
			priority *= (10 - (dist>>8));
			channel[i].priority = priority;
		}
	}
}
#endif

void S_Init(void)
{
	soundCurve = Z_Malloc(MAX_SND_DIST, PU_STATIC, NULL);
/*	if(snd_Channels > 8)
	{
		snd_Channels = 8;
	}*/
	S_SetMaxVolume(true);
}

void S_GetChannelInfo(SoundInfo_t *s)
{
	int i;
	ChanInfo_t *c;

	s->channelCount = numChannels > 16? 16 : numChannels;
	s->musicVolume = snd_MusicVolume;
	s->soundVolume = snd_MaxVolume;
	for(i = 0; i < s->channelCount; i++)
	{
		c = &s->chan[i];
		c->id = Channel[i].sound_id;
		c->priority = S_sfx[c->id].usefulness; //Channel[i].priority;
		c->name = S_sfx[c->id].name;
		c->mo = Channel[i].mo;
		if(c->mo)
			c->distance = Channel[i].volume;/*P_AproxDistance(c->mo->x-gi.Get(DD_VIEWX), c->mo->y-gi.Get(DD_VIEWY))
				>>FRACBITS*/
		else
			c->distance = 0;
	}
}

void S_SetMaxVolume(boolean fullprocess)
{
	int i;

	if(!fullprocess)
	{
		soundCurve[0] = (*((byte *)W_CacheLumpName("SNDCURVE", PU_CACHE))*(snd_MaxVolume*8))>>7;
	}
	else
	{
		for(i = 0; i < MAX_SND_DIST; i++)
		{
			soundCurve[i] = (*((byte *)W_CacheLumpName("SNDCURVE", PU_CACHE)+i)) * 255/127 /*(snd_MaxVolume*8))>>7*/;
		}
	}
}

static boolean musicPaused;

void S_SetMusicVolume(void)
{
	gi.SetMIDIVolume(snd_MusicVolume);
	if(snd_MusicVolume == 0)
	{
		gi.PauseSong();
		musicPaused = true;
	}
	else if(musicPaused)
	{
		musicPaused = false;
		gi.ResumeSong();
	}
}

void S_ShutDown(void)
{
}


//==========================================================================
//
// CONSOLE COMMANDS
//
//==========================================================================

int CCmdCD(int argc, char **argv)
{
	if(argc > 1)
	{
		if(!strcmpi(argv[1], "init"))
		{
			if(!gi.CD(DD_INIT,0))
				gi.conprintf( "CD init successful.\n");
			else
				gi.conprintf( "CD init failed.\n");
		}
		else if(!strcmpi(argv[1], "info") && argc == 2)
		{
			int secs = gi.CD(DD_GET_TIME_LEFT,0);
			gi.conprintf( "CD available: %s\n", gi.CD(DD_AVAILABLE,0)? "yes" : "no");
			gi.conprintf( "First track: %d\n", gi.CD(DD_GET_FIRST_TRACK,0));
			gi.conprintf( "Last track: %d\n", gi.CD(DD_GET_LAST_TRACK,0));
			gi.conprintf( "Current track: %d\n", gi.CD(DD_GET_CURRENT_TRACK,0));
			gi.conprintf( "Time left: %d:%02d\n", secs/60, secs%60);
			gi.conprintf( "Play mode: ");
			if(musicPaused)
				gi.conprintf( "paused\n");
			else if(s_CDTrack)
				gi.conprintf( "looping track %d\n", s_CDTrack);
			else
				gi.conprintf( "map track\n");
			return true;
		}
		else if(!strcmpi(argv[1], "play") && argc == 3)
		{
			s_CDTrack = atoi(argv[2]);
			if(!gi.CD(DD_PLAY_LOOP, s_CDTrack)) 
			{
				gi.conprintf( "Playing track %d.\n", s_CDTrack);
			}
			else
			{
				gi.conprintf( "Error playing track %d.\n", s_CDTrack);
				return false;
			}
		}
		else if(!strcmpi(argv[1], "stop") && argc == 2)
		{
			gi.CD(DD_STOP,0);
			gi.conprintf( "CD stopped.\n");
		}
		else if(!strcmpi(argv[1], "resume") && argc == 2)
		{
			gi.CD(DD_RESUME,0);
			gi.conprintf( "CD resumed.\n");
		}
		else
			gi.conprintf( "Bad command. Try 'cd'.\n");
	}
	else
	{
		gi.conprintf( "CD player control. Usage: CD (cmd)\n");
		gi.conprintf( "Commands are: init, info, play (track#), map, map (#), stop, resume.\n");
	}
	return true;
}

int CCmdMidi(int argc, char **argv)
{
	if(argc == 1)
	{
		gi.conprintf( "Usage: midi (cmd)\n");
		gi.conprintf( "Commands are: reset, play (name), map, map (num).\n");
		return true;
	}
	if(argc == 2)
	{
		if(!strcmpi(argv[1], "reset"))
		{
			if(rs)
			{
				gi.StopSong();
				gi.W_ChangeCacheTag(mus_lumpnum, PU_CACHE);
				rs = 0;
			}
			mus_song = -1;
			gi.conprintf( "MIDI has been reset.\n");
		}
		else if(!strcmpi(argv[1], "map")) 
		{
			gi.conprintf( "Playing the song of the current map (%d).\n", gamemap);
			S_StartSong(gamemap, true);
		}
		else
			return false;
	}
	else if(argc == 3)
	{
		if(!strcmpi(argv[1], "map"))
		{
			gi.conprintf( "Playing song for map %d.\n", atoi(argv[2]));
			S_StartSong(atoi(argv[2]), true);
		}
		else
			return false;
	}
	else 
		return false;
	// Oh, we're done.
	return true;
}
