#include "DoomDef.h"
#include "P_local.h"
#include "H_Net.h"
#include "soundst.h"
#include "settings.h"

// TYPES --------------------------------------------------------------------

// PUBLIC DATA --------------------------------------------------------------

boolean		handshake_received = false;

// CODE ---------------------------------------------------------------------

int CCmdSetColor(int argc, char **argv)
{
	if(argc != 2)
	{
		gi.conprintf( "Usage: %s (color)\n", argv[0]);
		gi.conprintf( "Color 4 uses the player number as color.\n");
		return true;
	}
	cfg.netColor = atoi(argv[1]);
	if(server)
	{
		cfg.PlayerColor[0] = PLR_COLOR(0, cfg.netColor);

		// Change the mobj.
		players[0].plr->mo->flags &= ~MF_TRANSLATION;
		players[0].plr->mo->flags |= cfg.PlayerColor[0] << MF_TRANSSHIFT;

		// Tell the clients about the change.
		NetSv_SendPlayerInfo(0, DDSP_ALL_PLAYERS);
	}
	else
	{
		// Tell the server about the change.
		NetCl_SendPlayerInfo();
	}
	return true;
}

int CCmdSetMap(int argc, char **argv)
{
	if(!server) return false;
	if(argc != 3)
	{
		gi.conprintf( "Usage: %s (episode) (map)\n", argv[0]);
		return true;
	}
	// Update game mode.	
	deathmatch = cfg.netDeathmatch;
	nomonsters = cfg.netNomonsters;
	respawnparm = cfg.netRespawn;
	cfg.jumpEnabled = cfg.netJumping;

	G_DeferedInitNew(gameskill, atoi(argv[1]), atoi(argv[2]));	
	return true;
}

ccmd_t netCCmds[] =
{
	"setcolor",		CCmdSetColor,	"Set player color.",
	"setmap",		CCmdSetMap,		"Set map.",
	NULL
};

int H_NetServerClose(int before)
{
	if(!before)
	{
		// Restore normal game state.
		deathmatch = false;
		nomonsters = false;

		P_SetMessage(&players[consoleplayer], "NETGAME ENDS", true);
		S_StartSound(NULL, sfx_dorcls);
	}
	return true;
}

int H_NetServerStarted(int before)
{
	if(before) return true;

	// Set the game parameters.
	cfg.PlayerColor[0] = PLR_COLOR(0, cfg.netColor);
	deathmatch = cfg.netDeathmatch;
	nomonsters = cfg.netNomonsters;
	respawnparm = cfg.netRespawn;
	cfg.jumpEnabled = cfg.netJumping;

	G_CheckDemoStatus();
	advancedemo = false;

	// Start the game.	
	if(!cfg.netSlot)
	{
		G_InitNew(cfg.netSkill, cfg.netEpisode, cfg.netMap);
	}
	else
	{
		char filename[16];
		sprintf(filename, SAVEGAMENAME"%c.hsg", cfg.netSlot-1);
		// Load the game from the slot.
		G_LoadGame(filename);
	}
	// Close the menu, the game begins!!
	MN_DeactivateMenu();
	return true;
}

int H_NetConnect(int before)
{
	// We do nothing before the actual connection is made.
	if(before) return true;
	
	// After connecting we tell the server a bit about ourselves.
	NetCl_SendPlayerInfo();

	// Request a handshake.
	gi.SendPacket(DDSP_RELIABLE, DDPT_HELLO, 0, 0);
	return true;
}

int H_NetDisconnect(int before)
{
	if(before) return true;

	// Restore normal game state.
	deathmatch = false;
	nomonsters = false;
	return true;
}

int H_NetPlayerEvent(int plrNumber, int peType, void *data)
{
	char		msgbuff[256];

	// If this isn't a netgame, we won't react.
	if(!netgame) return true;

	if(peType == DDPE_ARRIVAL)
	{
		if(server)
		{
			NetSv_NewPlayerEnters(plrNumber);
		}
		else if(plrNumber == consoleplayer)
		{
			// We have arrived, the game should be begun.
			gi.Message("PE: arrived in netgame.\n");
			//H_NetServerStarted(false);			
			gamestate = GS_WAITING;
		}
		else 
		{
			// Client responds to new player?			
			gi.Message( "PE: (client) player %i has arrived.\n", plrNumber);
		}
	}
	else if(peType == DDPE_EXIT)
	{
		gi.Message( "PE: removing player %i's mobj.\n", plrNumber);
		// Spawn a teleport fog.
		P_SpawnTeleFog(players[plrNumber].plr->mo->x,
			players[plrNumber].plr->mo->y);
		// Let's get rid of the mobj.
		P_RemoveMobj(players[plrNumber].plr->mo);
		players[plrNumber].plr->mo = NULL;
		players[plrNumber].playerstate = PST_REBORN;

		// Print a notification.
		sprintf(msgbuff, "%s left the game", gi.GetPlayerName(plrNumber));
		MN_TextFilter(msgbuff);
		P_SetMessage(&players[consoleplayer], msgbuff, true);
		S_StartSound(NULL, sfx_chat);

		if(server) P_DealPlayerStarts();
	}
	// Don't show local chat messages.
	else if(peType == DDPE_CHAT_MESSAGE && plrNumber != consoleplayer)
	{
		int i, num, oldecho = cfg.echoMsg;
		// Count the number of players.
		for(i=num=0; i<MAXPLAYERS; i++)
			if(players[i].plr->ingame) num++;
		// If there are more than two players, include the name of
		// the player who sent this.
		if(num > 2)
			sprintf(msgbuff, "%s: %s", gi.GetPlayerName(plrNumber), data);
		else
			strcpy(msgbuff, data);
		MN_TextFilter(msgbuff);
		// The chat message is already echoed by the console.
		cfg.echoMsg = false;
		P_SetMessage(&players[consoleplayer], msgbuff, true);
		cfg.echoMsg = oldecho;
		S_StartSound(NULL, sfx_chat);
	}
	return true;
}

int H_NetWorldEvent(int type, int parm, void *data)
{
	short *ptr;
	int i;
	int mtype, flags;
	fixed_t x, y, z, momx, momy, momz;

	switch(type)
	{
	//
	// Server events:
	//
	case DDWE_HANDSHAKE: 
		// A new player is entering the game. We as a server should send him
		// the handshake packet(s) to update his world.
		gi.Message( "H_NetWorldEvent: Sending a handshake to player %i.\n", parm);
				
		// First, the game state.
		NetSv_SendGameState(GSF_CHANGE_MAP | GSF_CAMERA_INIT, parm);
		
		// Player info.
		for(i=0; i<MAXPLAYERS; i++)
			if(players[i].plr->ingame && i != parm)
				NetSv_SendPlayerInfo(i, parm);
		break;

	//
	// Client events:
	//
	case DDWE_PROJECTILE:
		// Projectile data consists of shorts.
		for(ptr=data, i=0; i<parm; i++)
		{
			flags = *(unsigned short*) ptr & DDMS_FLAG_MASK;
			mtype = *(unsigned short*) ptr & ~DDMS_FLAG_MASK;
			ptr++;
			x = *ptr++ << 16;
			y = *ptr++ << 16;
			z = *ptr++ << 16;
			momx = momy = momz = 0;
			if(flags & DDMS_MOVEMENT_XY)
			{
				momx = *ptr++ << 8;
				momy = *ptr++ << 8;
			}
			if(flags & DDMS_MOVEMENT_Z)
			{
				momz = *ptr++ << 8;
			}
			NetCl_SpawnMissile(mtype, x, y, z, momx, momy, momz);
		}
		break;

	case DDWE_SECTOR_SOUND:
		// High word: sector number, low word: sound id.
		if(parm & 0xffff)
			S_StartSound( (mobj_t*) &sectors[parm >> 16].soundorg, parm & 0xffff);
		else
			S_StopSound( (mobj_t*) &sectors[parm >> 16].soundorg);
		break;

	default:
		return false;
	}
	return true;		
}

void H_HandlePacket(int fromplayer, int type, void *data, int length)
{
	byte *bData = data;

	//
	// Server events.
	//
	if(server)
	{
		switch(type)
		{
		case GPT_PLAYER_INFO:
			// A player has changed color or other settings.
			NetSv_ChangePlayerInfo(fromplayer, data);
			break;
		}
		return;
	}
	//
	// Client events.
	//
	switch(type) 
	{
	case GPT_GAME_STATE:
		NetCl_UpdateGameState(data);

		// Tell the engine we're ready to proceed. It'll start handling
		// the world updates after this variable is set.
		gi.Set(DD_GAME_READY, true);

		// Tell the server that we're ready. 
		// Note: this *must be* done, or the server won't know if we're 
		// ready to begin.
		gi.SendPacket(DDSP_RELIABLE, DDPT_OK, NULL, 0);
		break;

	case GPT_MESSAGE:
		P_SetMessage(&players[consoleplayer], data, true);
		break;

	case GPT_CONSOLEPLAYER_STATE:
		NetCl_UpdatePlayerState(data, consoleplayer);
		break;

	case GPT_PLAYER_STATE:
		NetCl_UpdatePlayerState(bData+1, bData[0]);
		break;

	case GPT_PSPRITE_STATE:
		NetCl_UpdatePSpriteState(data);
		break;

	case GPT_SOUND:
		NetCl_PlaySound(data);
		break;

	case GPT_SECTOR_SOUND:
		NetCl_PlaySectorSound(data);
		break;

	case GPT_INTERMISSION:
		NetCl_Intermission(data);
		break;

	case GPT_FINALE:
		NetCl_Finale(data);
		break;

	case GPT_PLAYER_INFO:
		NetCl_UpdatePlayerInfo(data);
		break;

	default:
		gi.Message( "H_HandlePacket: received unknown packet, type=%i.\n", type);
	}
}

