#include "DoomDef.h"
#include "P_local.h"
#include "H_Net.h"
#include "settings.h"

// MACROS -----------------------------------------------------------------

// Plane tracker flags.
#define PTF_MOVING		0x01		// Height currently moving.
#define PTF_TEXTURE		0x02		// Texture has been changed.
#define PTF_SOUND_SENT	0x04		// The move sound has been sent.

// Sector tracker flags.
#define STF_LIGHT_CHANGED	0x01	// Light level changed in the sector.

#define UPD_BUFFER_LEN	500

// How long is the largest possible sector update?
#define MAX_SECTORUPD	20
#define MAX_SIDEUPD		9

#define WRITE_SHORT(byteptr, val)	{(*(short*)(byteptr) = (val)); byteptr += 2;}

// TYPES ------------------------------------------------------------------

// Tracker planes.
enum
{
	PT_FLOOR,
	PT_CEILING
};

typedef struct
{
	byte flags;
	fixed_t initial;
	fixed_t dest;
	fixed_t speed;
} plane_tracker_t;

// There's one of these for each sector.
typedef struct
{
	plane_tracker_t	plane[2];	// Floor and ceiling.
	byte flags;
} sector_tracker_t;

typedef struct
{
	unsigned short initialTop, initialMid, initialBottom;
} side_tracker_t;

// EXTERNAL FUNCTIONS -----------------------------------------------------

void R_SetAllDoomsdayFlags();

// PRIVATE DATA -----------------------------------------------------------

static sector_tracker_t	*trackers;
static side_tracker_t *sideTrackers;

// CODE -------------------------------------------------------------------

#if 0
int NetSv_PackSector(byte *buffer, int sectornum, int force)
{
	sector_t *sec = &sectors[sectornum];
	sector_tracker_t *track = &trackers[sectornum];
	plane_tracker_t *pt;
	byte *ptr = buffer;
	packet_sectorupdate_t *upd = (packet_sectorupdate_t*) ptr;

	upd->flags = 0;
	upd->sector = sectornum;

	ptr += sizeof(*upd);
	// Check floor first.
	pt = &track->plane[PT_FLOOR];
	if(!force && pt->initial != sec->floorheight || force & DDSU_FLOOR_HEIGHT)
	{
		upd->flags |= DDSU_FLOOR_HEIGHT;
		WRITE_SHORT(ptr, sec->floorheight >> FRACBITS);
	}
	if(!force && pt->flags & PTF_MOVING || force & DDSU_FLOOR_MOVING)
	{
		upd->flags |= DDSU_FLOOR_MOVING;
		// Destination and speed.
		WRITE_SHORT(ptr, pt->dest >> FRACBITS);
		WRITE_SHORT(ptr, (pt->speed&0xffff00) >> 8);
	}
	if(pt->flags & PTF_TEXTURE)
	{
		upd->flags |= DDSU_FLOORPIC;
		WRITE_SHORT(ptr, sec->floorpic);
	}
	// Then ceiling.
	pt = &track->plane[PT_CEILING];
	if(!force && pt->initial != sec->ceilingheight || force & DDSU_CEILING_HEIGHT)
	{
		upd->flags |= DDSU_CEILING_HEIGHT;
		WRITE_SHORT(ptr, sec->ceilingheight >> FRACBITS);
	}
	if(!force && pt->flags & PTF_MOVING || force & DDSU_CEILING_MOVING)
	{
		upd->flags |= DDSU_CEILING_MOVING;
		// Destination and speed.
		WRITE_SHORT(ptr, pt->dest >> FRACBITS);
		WRITE_SHORT(ptr, (pt->speed&0xffff00) >> 8);
	}
	if(pt->flags & PTF_TEXTURE)
	{
		upd->flags |= DDSU_CEILINGPIC;
		WRITE_SHORT(ptr, sec->ceilingpic);
	}
	if(track->flags & STF_LIGHT_CHANGED)
	{	
		upd->flags |= DDSU_LIGHT_LEVEL;
		*ptr++ = sec->lightlevel;
	}
	// If nothing was written, this sector can be skipped.
	if(!upd->flags) return 0;
	return ptr-buffer;
}

int NetSv_PackSide(byte *buffer, int sidenum, int forceflags)
{
	side_t *side = sides + sidenum;
	side_tracker_t *st = sideTrackers + sidenum;	
	byte *ptr = buffer;
	packet_wallupdate_t *upd = (packet_wallupdate_t*) ptr;

	upd->flags = 0;
	upd->side = sidenum;

	ptr += sizeof(*upd);		
	if((!forceflags && st->initialTop != side->toptexture) 
		|| forceflags & DDWU_TOP)
	{
		upd->flags |= DDWU_TOP;
		WRITE_SHORT(ptr, side->toptexture);
	}
	if((!forceflags && st->initialMid != side->midtexture) 
		|| forceflags & DDWU_MID)
	{
		upd->flags |= DDWU_MID;
		WRITE_SHORT(ptr, side->midtexture);
	}
	if((!forceflags && st->initialBottom != side->bottomtexture) 
		|| forceflags & DDWU_BOTTOM)
	{
		upd->flags |= DDWU_BOTTOM;
		WRITE_SHORT(ptr, side->bottomtexture);
	}
	if(!upd->flags) return 0;
	return ptr-buffer;
}

// If speed is zero, the move will be immediate.
// Always give a positive speed (really doesn't matter).
void NetSv_PlaneMoves(int sector, boolean ceiling, fixed_t speed)
{
	byte	buffer[MAX_SECTORUPD];
	int		len;

	if(!server) return;

	/*gi.Message( "NetSv_PlaneMoves: sec%i, ceil?%i, dest:%f, speed:%f, tex:%i.\n",
		sector, ceiling, FIX2FLT(dest_height), FIX2FLT(speed), texture);*/

	len = NetSv_PackSector(buffer, sector, 
		ceiling? (speed? DDSU_CEILING_MOVING : DDSU_CEILING_HEIGHT) 
		: (speed? DDSU_FLOOR_MOVING : DDSU_FLOOR_HEIGHT));
	gi.SendPacket(DDSP_ALL_PLAYERS, DDPT_SECTOR_UPDATE, buffer, len);
}

// Texture -1 means 'no change'.
void NetSv_TextureChanges(int sidenum, int flags)
{
	byte buffer[MAX_SIDEUPD];
	int len;

	if(!netgame || !server) return;

	/*gi.Message( "NetSv_TextureChanges: side%i, top%i, mid%i, bot%i\n",
		sidenum, newTop, newMid, newBottom);*/

	len = NetSv_PackSide(buffer, sidenum, flags);
	gi.SendPacket(DDSP_ALL_PLAYERS, DDPT_WALL_UPDATE, buffer, len);
}

void NetSv_InitTrackers(void)
{
	int	i;
	sector_t *sec;
	side_t *side;

	if(!netgame || !server) return; // Only if needed.

	trackers = Z_Malloc(sizeof(sector_tracker_t) * numsectors, PU_LEVEL, 0);
	memset(trackers, 0, sizeof(sector_tracker_t) * numsectors);
	// Store the initial heights. No need to send data for nothing.
	for(i=0, sec=sectors; i<numsectors; i++, sec++)
	{
		trackers[i].plane[PT_FLOOR].initial = sec->floorheight;
		trackers[i].plane[PT_CEILING].initial = sec->ceilingheight;
	}

	// Side trackers keep track of changed wall textures.
	sideTrackers = Z_Malloc(sizeof(side_tracker_t) * numsides, PU_LEVEL, 0);
	for(i=0, side=sides; i<numsides; i++, side++)
	{
		sideTrackers[i].initialMid = side->midtexture;
		sideTrackers[i].initialTop = side->toptexture;
		sideTrackers[i].initialBottom = side->bottomtexture;
	}
}

void NetSv_PlaneSound(sector_t *sector, boolean isCeiling, int sound_id, int pace)
{
	byte buffer[20], *ptr = buffer;
	int secnum;
	sector_tracker_t *secTrack;
	plane_tracker_t *pTrack;

	if(!netgame || !server) return; // Only if needed.	

	secnum = sector - sectors;
	secTrack = &trackers[secnum];
	pTrack = &secTrack->plane[!isCeiling? PT_FLOOR : PT_CEILING];

	if(!(pTrack->flags & PTF_SOUND_SENT))
	{
		pTrack->flags |= PTF_SOUND_SENT; 
		WRITE_SHORT(ptr, secnum);
		*ptr++ = (isCeiling? 0x80 : 0) | pace;
		WRITE_SHORT(ptr, sound_id);
		gi.SendPacket(DDSP_ALL_PLAYERS, DDPT_PLANE_SOUND, buffer, ptr-buffer);
	}
}

// To report a stop in movement, set speed to zero.
void NetSv_SectorReport(sector_t *sector, boolean isCeiling, fixed_t dest,
						fixed_t speed, int floorpic)
{
	int secnum;
	sector_tracker_t *secTrack;
	plane_tracker_t *pTrack;
	fixed_t current;

	if(!netgame || !server) return; // Only if needed.	

	secnum = sector - sectors;
	secTrack = &trackers[secnum];
	pTrack = &secTrack->plane[!isCeiling? PT_FLOOR : PT_CEILING];
	current = isCeiling? sector->ceilingheight : sector->floorheight;

	// Check any changes. 
	if(floorpic != -1)
	{
		// Mark the texture change.
		secTrack->plane[PT_FLOOR].flags |= PTF_TEXTURE;
	}
	if(pTrack->flags & PTF_MOVING)
	{
		// The plane was reported moving the last time.
		if(!speed)
		{
			// The plane has stopped.
			pTrack->flags &= ~(PTF_MOVING | PTF_SOUND_SENT);
			// Send a notification about this.
			NetSv_PlaneMoves(secnum, isCeiling, 0);
		}
		else
		{
			if(pTrack->speed != speed || pTrack->dest != dest)
			{
				pTrack->speed = speed;
				pTrack->dest = dest;
				pTrack->flags &= ~PTF_SOUND_SENT;
				// Send a notification about this.
				NetSv_PlaneMoves(secnum, isCeiling, speed);
			}
		}
	}
	else
	{
		// The plane was not moving the last time.
		pTrack->speed = speed;
		pTrack->dest = dest;
		if(speed)
		{
			pTrack->flags &= ~PTF_SOUND_SENT;
			pTrack->flags |= PTF_MOVING;
		}
		// Send a notification about this.
		NetSv_PlaneMoves(secnum, isCeiling, speed);
	}
}

void NetSv_LightReport(sector_t *sector)
{
	byte buffer[MAX_SECTORUPD];
	int len, secNum = sector-sectors;
	sector_tracker_t *sTrack = trackers + secNum;

	if(!netgame || !server) return;

	sTrack->flags |= STF_LIGHT_CHANGED;
	// Send a notification about this. The Light Level flag
	// isn't noticed in forcing, but it prevents the sending
	// of the floor/ceiling data.
	len = NetSv_PackSector(buffer, secNum, DDSU_LIGHT_LEVEL);
	gi.SendPacket(DDSP_ALL_PLAYERS, DDPT_SECTOR_UPDATE, buffer, len);
}

// Initial world updates are sent reliably.
void NetSv_SendWorldUpdate(int playernum)
{
	int		i;
	byte	buffer[UPD_BUFFER_LEN], *ptr = buffer;

	// Update sectors.
	for(i=0; i<numsectors; i++)
	{
		ptr += NetSv_PackSector(ptr, i, 0);
		// Buffer getting full?
		if(ptr-buffer >= UPD_BUFFER_LEN-MAX_SECTORUPD
			|| i==numsectors-1) // Always send after the last one.
		{
			// Send out a packet.
			gi.SendPacket(playernum | DDSP_RELIABLE, DDPT_SECTOR_UPDATE, 
				buffer, ptr-buffer);
			// Back to the beginning of the buffer.
			ptr = buffer;
		}
	}
	// Update wall textures.
	for(ptr=buffer, i=0; i<numsides; i++)
	{
		ptr += NetSv_PackSide(ptr, i, 0);
		// Buffer getting full?
		if(ptr-buffer >= UPD_BUFFER_LEN-MAX_SIDEUPD 
			|| i==numsides-1) // Always send after the last one.
		{
			gi.SendPacket(playernum | DDSP_RELIABLE, DDPT_WALL_UPDATE, 
				buffer, ptr-buffer);
			ptr = buffer;
		}
	}
}
#endif

// Server calls this when new players enter the game.
void NetSv_NewPlayerEnters(int plrnumber)
{
	player_t	*plr = &players[plrnumber];
	
	gi.Message( "NetSv_NewPlayerEnters: spawning player %i.\n", plrnumber);

	plr->playerstate = PST_REBORN;	// Force an init.

	// Re-deal player starts.
	P_DealPlayerStarts();

	// Spawn the player into the world.
	// FIXME: Spawn a telefog in front of the player.
	P_SpawnPlayer(&playerstarts[plr->startspot], plrnumber);
	// Get rid of anybody at the starting spot.
	P_Telefrag(plr->plr->mo);
}

void NetSv_SendMessage(int plrNum, char *msg)
{
	if(!netgame || !server || plrNum == consoleplayer) return;
	gi.SendPacket(plrNum | DDSP_RELIABLE, GPT_MESSAGE,
		msg, strlen(msg)+1);
}

void NetSv_SendPlayerState(int srcPlrNum, int destPlrNum, int flags, boolean reliable)
{
	int pType = srcPlrNum==destPlrNum? GPT_CONSOLEPLAYER_STATE : GPT_PLAYER_STATE;
	player_t *pl = &players[srcPlrNum];
	byte buffer[UPD_BUFFER_LEN], *ptr = buffer, fl;
	int i, k;

	if(!netgame || !server || srcPlrNum == consoleplayer 
		|| destPlrNum == consoleplayer || !pl->plr->ingame
		|| !players[destPlrNum].plr->ingame) return;

	// Include the player number if necessary.
	if(pType == GPT_PLAYER_STATE) *ptr++ = srcPlrNum;

	// The first byte contains the flags.
	WRITE_SHORT(ptr, flags);
	if(flags & PSF_STATE) 
		*ptr++ = pl->playerstate | (pl->armortype << 4);
	if(flags & PSF_HEALTH) *ptr++ = pl->health;
	if(flags & PSF_ARMOR_POINTS) *ptr++ = pl->armorpoints;
	if(flags & PSF_INVENTORY)
	{
		*ptr++ = pl->inventorySlotNum;
		for(i=0; i<pl->inventorySlotNum; i++)
			WRITE_SHORT(ptr, (pl->inventory[i].type & 0xff) 
				| ((pl->inventory[i].count & 0xff) << 8));
	}
	if(flags & PSF_POWERS) // Austin!
	{
		// First see which powers should be sent.
		for(i=1, *ptr=0; i<NUMPOWERS; i++)
			if(pl->powers[i]) *ptr |= 1 << (i-1);
		ptr++;
		for(i=1; i<NUMPOWERS; i++)
			if(pl->powers[i])
				*ptr++ = pl->powers[i]/35 + 1;	// Send as seconds.
	}
	if(flags & PSF_KEYS) 
	{
		for(i=0, *ptr=0; i<NUMKEYS; i++) if(pl->keys[i]) *ptr |= 1<<i;
		ptr++;
	}
	if(flags & PSF_FRAGS)
	{
		for(k=0, i=0; i<MAXPLAYERS; i++) k += pl->frags[i];
		WRITE_SHORT(ptr, k);
	}
	if(flags & PSF_ALL_FRAGS)
	{
		for(i=0; i<MAXPLAYERS; i++) WRITE_SHORT(ptr, pl->frags[i]);
	}
	if(flags & PSF_OWNED_WEAPONS)
	{
		for(fl=0, i=0; i<8; i++) if(pl->weaponowned[i]) fl |= 1<<i;
		*ptr++ = fl;
	}
	if(flags & PSF_AMMO)
	{
		for(i=0; i<NUMAMMO; i++) WRITE_SHORT(ptr, pl->ammo[i]);
	}
	if(flags & PSF_MAX_AMMO)
	{
		for(i=0; i<NUMAMMO; i++) WRITE_SHORT(ptr, pl->maxammo[i]);
	}
	if(flags & PSF_COUNTERS)
	{
		WRITE_SHORT(ptr, pl->killcount);
		*ptr++ = pl->itemcount;
		*ptr++ = pl->secretcount;
	}
	if(flags & PSF_PENDING_WEAPON || flags & PSF_READY_WEAPON)
	{
		// These two will be in the same byte.
		fl = 0;
		if(flags & PSF_PENDING_WEAPON) fl |= pl->pendingweapon & 0xf;
		if(flags & PSF_READY_WEAPON) fl |= (pl->readyweapon & 0xf) << 4;
		*ptr++ = fl;
	}
	if(flags & PSF_CHICKEN_TIME)
	{
		*ptr++ = pl->chickenTics/35 + 1;
	}

	/*gi.conprintf( "PlrUpd (%i bytes, flags=%x):\n", ptr-buffer, flags);
	for(i=0; i<ptr-buffer; i++)
	{
		gi.conprintf( "%02x ", buffer[i]);
		if(i && !(i%10)) gi.conprintf( "\n");
	}
	gi.conprintf( "\n");*/

	// Finally, send the packet.
	gi.SendPacket(destPlrNum | (reliable? DDSP_RELIABLE : 0),
		pType, buffer, ptr-buffer);	
}

void NetSv_PSpriteChange(int plrNum, int state)
{
	byte buffer[20], *ptr = buffer;

	if(!netgame || !server) return;

	WRITE_SHORT(ptr, state);
	gi.SendPacket(plrNum, GPT_PSPRITE_STATE, buffer, ptr-buffer);
}

void NetSv_Sound(mobj_t *origin, int sound_id, int toPlr)
{
	int i;
	byte buffer[20], *ptr;
	boolean not;
	int dest;

	if(!netgame || !server) return;

	ptr = buffer;
	not = (toPlr & NSSF_NOT) != 0;
	dest = toPlr & NSSF_MASK;

	*ptr++ = origin? SNDF_ORIGIN : 0;
	*ptr++ = sound_id;
	if(origin)
	{
		WRITE_SHORT(ptr, origin->x >> 16);
		WRITE_SHORT(ptr, origin->y >> 16);
		WRITE_SHORT(ptr, origin->z >> 16);
	}
	// Send to the appropriate clients.
	for(i=1; i<MAXPLAYERS; i++)
	{
		if(!players[i].plr->ingame) continue;
		if(!toPlr || (i == dest) != not)
		{
			mobj_t *plrMo = players[i].plr->mo;
			// Check the distance.
			if(origin && P_AproxDistance(origin->x - plrMo->x, 
				origin->y - plrMo->y) > FRACUNIT*MAX_SND_DIST) continue;
			// We can send to this client.
			gi.SendPacket(i, GPT_SOUND, buffer, ptr-buffer);
		} 
	}
}

void NetSv_SectorSound(sector_t *sector, int sound_id)
{
	byte buffer[20], *ptr = buffer;

	if(!netgame || !server) return;

	*ptr++ = sound_id;
	WRITE_SHORT(ptr, sector-sectors);
	gi.SendPacket(DDSP_ALL_PLAYERS, GPT_SECTOR_SOUND, buffer, ptr-buffer);
}

void NetSv_SendGameState(int flags, int to)
{
	byte buffer[30], *ptr = buffer;
	packet_gamestate_t *gs = (packet_gamestate_t*) ptr;

	if(!server) return;

	// We can't init the camera if we don't know which 
	// player this packet is going to.
	if(to < 1 || to >= MAXPLAYERS) flags &= ~GSF_CAMERA_INIT;

	gs->flags = flags;
	gs->episode = gameepisode;
	gs->map = gamemap;
	gs->skill = gameskill;
	gs->deathmatch = deathmatch;
	gs->monsters = !nomonsters;
	gs->respawn = respawnparm;
	gs->jumping = cfg.jumpEnabled;
	ptr += sizeof(packet_gamestate_t);

	if(flags & GSF_CAMERA_INIT)
	{
		mobj_t *mo = players[to].plr->mo;
		WRITE_SHORT(ptr, mo->x >> 16);
		WRITE_SHORT(ptr, mo->y >> 16);
		WRITE_SHORT(ptr, mo->z >> 16);
		WRITE_SHORT(ptr, mo->angle >> 16);
	}

	// Send the packet.
	gi.SendPacket(to | DDSP_RELIABLE, GPT_GAME_STATE, buffer, ptr-buffer);
}

void NetSv_Intermission(int flags, int state, int time)
{
	byte buffer[8], *ptr = buffer;

	if(!server) return;

	*ptr++ = flags;
	if(flags & IMF_STATE) *ptr++ = state;
	if(flags & IMF_TIME) WRITE_SHORT(ptr, time);
	gi.SendPacket(DDSP_ALL_PLAYERS | DDSP_RELIABLE, GPT_INTERMISSION,
		buffer, ptr-buffer);
}

void NetSv_Finale(int flags)
{
	byte buffer[8], *ptr = buffer;

	if(!server) return;

	*ptr++ = flags;
	gi.SendPacket(DDSP_ALL_PLAYERS | DDSP_RELIABLE, GPT_FINALE,
		buffer, ptr-buffer);
}

void NetSv_SendPlayerInfo(int whose, int to_whom)
{
	byte buffer[10], *ptr = buffer;

	if(!server) return;

	*ptr++ = whose;
	*ptr++ = cfg.PlayerColor[whose];
	gi.SendPacket(to_whom | DDSP_RELIABLE, GPT_PLAYER_INFO,
		buffer, ptr-buffer);
}

void NetSv_ChangePlayerInfo(int from, byte *data)
{
	player_t *pl = players + from;
	int col;

	// Color is first.
	col = *data++;
	cfg.PlayerColor[from] = PLR_COLOR(from, col);

	gi.conprintf( "NetSv_ChangePlayerInfo: pl%i, col=%i\n", from, 
		cfg.PlayerColor[from]);

	if(pl->plr->mo)
	{
		// Change the player's mobj.
		pl->plr->mo->flags &= ~MF_TRANSLATION;
		pl->plr->mo->flags |= col << MF_TRANSSHIFT;
	}

	// Re-deal start spots.
	P_DealPlayerStarts();

	// Tell the other clients about the change.
	NetSv_SendPlayerInfo(from, DDSP_ALL_PLAYERS);
}

void NetSv_Ticker(void)
{
	player_t *plr;
	int	i, palette;

	// This is done here for servers.
	R_SetAllDoomsdayFlags();
	
	// Set the camera filters for players.
	// This is from SB_PaletteFlash().
	for(i=1; i<MAXPLAYERS; i++)
	{
		if(!players[i].plr->ingame) continue;

		plr = &players[i];
		if(plr->damagecount)
		{
			palette = (plr->damagecount+7)>>3;
			if(palette >= NUMREDPALS)
			{
				palette = NUMREDPALS-1;
			}
			palette += STARTREDPALS;
		}
		else if(plr->bonuscount)
		{
			palette = (plr->bonuscount+7)>>3;
			if(palette >= NUMBONUSPALS)
			{
				palette = NUMBONUSPALS-1;
			}
			palette += STARTBONUSPALS;
		}
		else
		{
			palette = 0;
		}
		plr->plr->filter = H_GetFilterColor(palette);
	}

	// Send the player state updates.
	for(i=1; i<MAXPLAYERS; i++)
	{
		if((gametic+i) % 10) continue; // Don't send on every tic.
		if(!players[i].plr->ingame || !players[i].update) continue;
		// Always send reliably (unwise?).
		NetSv_SendPlayerState(i, i, players[i].update, true);
		players[i].update = 0;
	}
}

