
//**************************************************************************
//**
//** CL_MAIN.C
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "dd_def.h"
#include "dd_net.h"
#include "i_net.h"
#include "i_win32.h"
#include "i_timer.h"
#include "console.h"
#include "cl_def.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// Client mobj.
typedef struct
{
	int		tic;		// Zero if not in use.
	mobj_t	*thing;
} clmobj_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

boolean		handshake_received = false;
int			game_ready = false;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

int Cl_GameReady()
{
	return (handshake_received && game_ready);
}

void Cl_CleanUp()
{
	CON_Printf( "Cl_CleanUp.\n");

	Cl_CleanClientMobjs();
	Cl_RemoveMovers();
}

void Cl_AnswerHandshake(handshake_packet_t *shake)
{
	int		i;

	// Check the version number.
	if(shake->version != SV_VERSION)
	{
		ST_Message( "Cl_AnswerHandshake: version conflict! (you:%i server:%i)\n",
			SV_VERSION, shake->version);
		CON_Execute("net disconnect", false);
		return;
	}

	// Update time and player ingame status.
	gametic = maketic = 0; 
	resend_start = 0;
	for(i=0; i<MAXPLAYERS; i++)
		players[i].ingame = (shake->playerMask & (1<<i)) != 0;	
	consoleplayer = displayplayer = shake->yourConsole;
	clients[consoleplayer].numtics = 0;
	clients[consoleplayer].firsttic = 0;

	server = false;

	if(handshake_received) return;

	// This prevents redundant re-initialization.
	handshake_received = true;

	// Soon after this packet will follow the game's handshake.
	game_ready = false;

	CON_Printf( "Cl_AnswerHandshake: myConsole:%i.\n",
		shake->yourConsole);

	// Tell the game that we have arrived. The level will
	// be changed when the game's handshake arrives (handled
	// in the dll).
	gx.NetPlayerEvent(consoleplayer, DDPE_ARRIVAL, 0);

	// Prepare the client-side data.
	Cl_InitClientMobjs();
	Cl_InitMovers();

	// Get ready for ticking.
	firstNetUpdate = firstRunTic = true;
}

void Cl_HandlePlayerInfo(playerinfo_packet_t* info)
{
	boolean present;

	CON_Printf( "Cl_HandlePlayerInfo: console:%i name:%s\n",
		info->console, info->name);

	// Is the console number valid?
	if(info->console < 0 || info->console >= MAXPLAYERS) return;

	present = players[info->console].ingame;
	players[info->console].ingame = true;
	strcpy(clients[info->console].name, info->name);

	if(!present)
	{
		// This is a new player! Let the game know about this.
		gx.NetPlayerEvent(info->console, DDPE_ARRIVAL, 0);
	}
}

void Cl_PlayerLeaves(int number)
{
	CON_Printf( "Cl_PlayerLeaves: player %i has left.\n", number);
	players[number].ingame = false;
	gx.NetPlayerEvent(number, DDPE_EXIT, 0);
}


//========================================================================
//
// Cl_GetPackets
//
// Client's packet handler.
// Handles all the events the server sends.
//
//========================================================================

void Cl_GetPackets(void)
{
	int		i;

	// All messages come from the server.
	while(Net_GetPacket())
	{
		// First check for packets that are only valid when 
		// a game is in progress.
		if(Cl_GameReady())
		{
			switch(netbuffer.msg.type)
			{
			case psv_frame:
				Cl_HandleFrame();
				break;
				
			case psv_sector_update:
				Cl_HandleSectorUpdate();
				break;
				
			case psv_wall_update:
				Cl_HandleWallUpdate();
				break;

			case psv_plane_sound:
				Cl_HandlePlaneSound();
				break;
			}
		}
		// How about the rest?
		switch(netbuffer.msg.type)
		{
		case pkt_ping:
			Net_PingResponse();
			break;

		case psv_handshake:
			Cl_AnswerHandshake( (handshake_packet_t*) netbuffer.msg.data);
			break;
			
		case pkt_player_info:
			Cl_HandlePlayerInfo( (playerinfo_packet_t*) netbuffer.msg.data);		
			break;
			
		case psv_player_exit:
			Cl_PlayerLeaves(Msg_ReadByte());
			break;
			
		case pkt_chat:
			D_ShowChatMessage();
			gx.NetPlayerEvent(netbuffer.msg.data[0], DDPE_CHAT_MESSAGE,
				netbuffer.msg.data+3);
			break;
			
		case psv_server_close:	// We should quit?
			CON_Execute("net disconnect", true);		
			break;

		case psv_console_text:
			i = Msg_ReadLong();
			CON_FPrintf(i, netbuffer.cursor);
			break;
			
		default:
			if(netbuffer.msg.type >= pkt_game_marker)
			{
				gx.HandlePacket(netbuffer.player,
					netbuffer.msg.type, netbuffer.msg.data,
					netbuffer.length);
			}
		}
	}
}

void Cl_Ticker(void)
{
	if(!Cl_GameReady()) return;

	Cl_PredictMovement();	
}