
//**************************************************************************
//**
//** CL_MOBJ.C
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "dd_def.h"
#include "dd_net.h"
#include "i_net.h"
#include "i_win32.h"
#include "console.h"
#include "cl_def.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

typedef struct
{
	int		num;		// How many used? (MAX_CLMOBJS at most)
	int		time;		
	int		age;		// Buffer age in client time.
	mobj_t	*list;		
} clmobj_buffer_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

clmobj_buffer_t	cmbInput, cmbVisible;

int	predict_tics = 2;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static boolean	buffers_ok = false;
static int		camera_age;

// CODE --------------------------------------------------------------------

void Cl_UnsetThingPosition (mobj_t *thing)
{
	P_UnlinkThing(thing, DDLINK_SECTOR | 
		(thing->ddflags&DDMF_SOLID? DDLINK_BLOCKMAP : 0));
}

void Cl_SetThingPosition (mobj_t *thing)
{
	P_LinkThing(thing, DDLINK_SECTOR |
		(thing->ddflags&DDMF_SOLID? DDLINK_BLOCKMAP : 0));
}

void Cl_InitBuffer(clmobj_buffer_t *buf)
{
	int	i;
	mobj_t *mo;

	buf->num = 0;
	buf->time = 0;
	buf->age = 0;
	buf->list = Z_Malloc(i = sizeof(mobj_t) * MAX_CLMOBJS, PU_STATIC, 0);
	memset(buf->list, 0, i);
	for(i=0, mo=buf->list; i<MAX_CLMOBJS; i++, mo++)
	{
		mo->ddflags |= DDMF_REMOTE;
	}
}

// Allocate the client mobj buffers.
void Cl_InitClientMobjs()
{
	// Only need to initialize once.
	if(buffers_ok) return;
	buffers_ok = true;

	Cl_InitBuffer(&cmbInput);
	Cl_InitBuffer(&cmbVisible);
}

void Cl_ZeroBuffer(clmobj_buffer_t *buf)
{
	buf->num = 0;
	buf->time = 0;
	buf->age = 0;
}

// Unlink the visible buffer.
void Cl_CleanClientMobjs()
{
	int		i;
	mobj_t	*mo;

	// Unlink all the mobjs in the visible buffer.
	for(i=0, mo=cmbVisible.list; i<cmbVisible.num; i++, mo++)
		Cl_UnsetThingPosition(mo);

	// Clear both buffers.
	Cl_ZeroBuffer(&cmbInput);
	Cl_ZeroBuffer(&cmbVisible);
}

void Cl_UnpackMobj(mobj_t *mobj)
{
	unsigned short i = Msg_ReadShort();
	byte flags = (i & MDP_FLAG_MASK) >> 8;

	mobj->sprite = i & MDP_SPRITE_MASK;
	if(flags & MDP_ACCURATE)
	{
		mobj->x = (Msg_ReadShort() << 16) + (Msg_ReadByte() << 8); 
		mobj->y = (Msg_ReadShort() << 16) + (Msg_ReadByte() << 8); 
		if(flags & MDP_POSITION_Z)
		{
			mobj->z = (Msg_ReadShort() << 16) + (Msg_ReadByte() << 8); 
		}
		else
		{
			mobj->z = DDMAXINT;	// Will be corrected after linking in.
		}
	}
	else
	{
		mobj->x = Msg_ReadShort() << 16; 
		mobj->y = Msg_ReadShort() << 16;
		if(flags & MDP_POSITION_Z)
		{
			mobj->z = Msg_ReadShort() << 16;
		}
		else
		{
			mobj->z = DDMAXINT; // Will be corrected after linking in.
		}
	}
	if(flags & MDP_MOVEMENT_XY)
	{
		mobj->momx = Msg_ReadShort() << 8; 
		mobj->momy = Msg_ReadShort() << 8;
	}
	else
	{
		mobj->momx = mobj->momy = 0;
	}
	if(flags & MDP_MOVEMENT_Z)
	{
		mobj->momz = Msg_ReadShort() << 8;
	}
	else
	{
		mobj->momz = 0;
	}
	if(flags & MDP_DIRECTION)
	{
		if(flags & MDP_ACCURATE)
		{
			mobj->angle = Msg_ReadShort() << 16;
		}
		else
		{
			mobj->angle = Msg_ReadByte() << 24;
		}
	}
	mobj->ddflags = DDMF_REMOTE;	// Client mobjs always have this.
	mobj->floorclip = 0;
	if(flags & MDP_EXTRA_FLAGS)
	{
		i = Msg_ReadShort();
		if((i & MDPE_BRIGHTSHADOW) == MDPE_BRIGHTSHADOW) 
		{
			mobj->ddflags |= DDMF_BRIGHTSHADOW;
		}
		else
		{
			if(i & MDPE_SHADOW) mobj->ddflags |= DDMF_SHADOW;
			if(i & MDPE_ALTSHADOW) mobj->ddflags |= DDMF_ALTSHADOW;
		}
		//if(i & MDPE_FITTOP) mobj->ddflags |= DDMF_FITTOP;
		if(i & MDPE_NOFITBOTTOM) mobj->ddflags |= DDMF_NOFITBOTTOM;
		if(i & MDPE_VIEWALIGN) mobj->ddflags |= DDMF_VIEWALIGN;
		if(i & MDPE_SOLID) mobj->ddflags |= DDMF_SOLID;
		mobj->ddflags |= (i & MDPE_LIGHTOFFSET) >> MDPE_LIGHTOFFSETSHIFT
			<< DDMF_LIGHTOFFSETSHIFT;
		mobj->ddflags |= (i & MDPE_TRANSCLASS) >> MDPE_CLASSSHIFT
			<< DDMF_CLASSTRSHIFT;
		mobj->ddflags |= (i & MDPE_TRANSLATION) >> MDPE_TRANSSHIFT
			<< DDMF_TRANSSHIFT;
		// Extended flags?
		if(i & MDPE_EXTENDED) 
		{
			i = Msg_ReadByte();
			mobj->ddflags |= (i & MDPE2_LIGHTSCALE) >> MDPE_LIGHTSCALESHIFT 
				<< DDMF_LIGHTSCALESHIFT;
		}
		if(i & MDPE_FLOORCLIP)
		{
			mobj->floorclip = ((char) Msg_ReadByte()) << 14;
		}
	}
	if(flags & MDP_FRAME)
	{
		i = Msg_ReadByte();
		mobj->frame = i & MDP_FRAME_MASK;
		if(i & MDPF_FULLBRIGHT) mobj->frame |= FF_FULLBRIGHT;
		// Radius included?
		if(mobj->frame & FF_FULLBRIGHT) 
		{
			mobj->radius = Msg_ReadByte() << 16;
		}
	}
	else
	{
		mobj->frame = 0;
	}
}

// Moves everything from the input buffer to the visible buffer.
// Leaves the input buffer empty.
void Cl_SwapBuffers()
{
	int		i;
	mobj_t	*mo;

	// First clear the visible buffer.
	for(i=0, mo=cmbVisible.list; i<cmbVisible.num; i++, mo++)
		Cl_UnsetThingPosition(mo);

	// Update the visible buffer.
	cmbVisible.num = cmbInput.num;
	cmbVisible.age = 0;
	cmbVisible.time = cmbInput.time;

	// Copy all the mobjs from the input buffer.
	memcpy(cmbVisible.list, cmbInput.list, sizeof(mobj_t) * cmbInput.num);
	
	// Clear the input buffer.
	cmbInput.num = 0;

	// Link in all the visible mobjs.
	for(i=0, mo=cmbVisible.list; i<cmbVisible.num; i++, mo++)
	{
		Cl_SetThingPosition(mo);
		// Fix the Z coordinate, if necessary.
		if(mo->z == DDMAXINT)
			mo->z = mo->subsector->sector->floorheight;
	}
}

void Cl_HandleFrameMobjs(int time)
{
	mobj_t	*mo, temp;
	int		count;

	// Place the new mobjs into the input buffer.
	if(time != cmbInput.time && cmbInput.num)
	{
		CON_Printf( "time:%i - mobjs left in input buffer!\n", cmbInput.time);
		// The input buffer contains mobjs and the frame has evidently
		// changed. Move everything to visible buffer before adding the
		// new mobjs. Notice that this only happens if the server's End 
		// Frame event was not received.
		Cl_SwapBuffers();
	}
	// Update the input buffer's time.
	cmbInput.time = time;

	// Read all the mobjs from the packet.
	count = Msg_ReadByte();
	while(count-- > 0)
	{
		// If there is no room in the input buffer, we'll have to 
		// discard the rest of the mobjs.
		if(cmbInput.num == MAX_CLMOBJS) 
		{
			// Oh well.
			mo = &temp;
		}
		else
		{
			// Insert into the input buffer.
			mo = cmbInput.list + cmbInput.num++;
		}
		// Unpack the mobj data.
		Cl_UnpackMobj(mo);
	}
}

void Cl_HandleFramePlayer()
{
	int		plrNum;
	mobj_t	*mo;

	plrNum = Msg_ReadByte() & 0xf;
	mo = players[plrNum].mo;
	//if(!mo) break;	// Hmm?
	// Update received, reset age.
	clients[plrNum].age = 0;
	Cl_UnsetThingPosition(mo);
	Cl_UnpackMobj(mo);
	Cl_SetThingPosition(mo);
}

// A camera packet tells us the position, direction and movement 
// of the viewpoint.
void Cl_HandleFrameCamera()
{
	ddplayer_t *plr = &players[consoleplayer];
	byte flags = Msg_ReadByte();
	boolean fixangles = flags & CDP_FIX_ANGLES;

	Cl_UnsetThingPosition(plr->mo);
	Cl_UnpackMobj(plr->mo);
	Cl_SetThingPosition(plr->mo);
	// Fix the Z coordinate, if necessary.
	if(plr->mo->z == DDMAXINT)
		plr->mo->z = plr->mo->subsector->sector->floorheight;
	if(fixangles) 
	{
		plr->clAngle = plr->mo->angle;
		plr->clLookDir = Msg_ReadShort() / (float) DDMAXSHORT * 110;	

		/*CON_Printf( "Cl_HCamPkt: fixangles to %f / %f\n",
			plr->clAngle / (float) ANGLE_MAX * 360, 
			plr->clLookDir);*/
	}
	// Also unpack viewheight.
	plr->viewheight = Msg_ReadShort() << 8;
	// Color filter. Set to zero if not in the packet.
	plr->filter = flags & CDP_FILTER? Msg_ReadLong() : 0;
	// Extra light.
	plr->extralight = flags & CDP_EXTRA_LIGHT? Msg_ReadByte() : 0;
	plr->mo->ddflags |= DDMF_DONTDRAW; 
	camera_age = 0;
}

void Cl_HandleFrame()
{
	int	time = Msg_ReadByte();	// Frame time (ID of sorts).
	int i, count;

	Cl_HandleFrameCamera();
	// The number of players.
	count = Msg_ReadByte();	
	for(i=0; i<count; i++) Cl_HandleFramePlayer();
	Cl_HandleFrameMobjs(time);
	// The number of missile spawn requests.
	count = Msg_ReadByte();
	gx.NetWorldEvent(DDWE_PROJECTILE, count, netbuffer.cursor);
	Cl_SwapBuffers();
}

void Cl_MoveThing(mobj_t *mobj)
{
	Cl_UnsetThingPosition(mobj);
	mobj->x += mobj->momx;
	mobj->y += mobj->momy;
	mobj->z += mobj->momz;
	Cl_SetThingPosition(mobj);
}

void Cl_PredictMovement()
{
	int			i;
	mobj_t		*mo;
	ddplayer_t	*plr;
	client_t	*cl;

	// Predict camera (consoleplayer) movement.
	plr = &players[consoleplayer];
	if(plr->mo)
	{
		camera_age++;
		if(camera_age <= predict_tics)
		{
			Cl_MoveThing(plr->mo);
		}
	}
	// Predict the movement of players.
	for(i=0, cl=clients, plr=players; i<MAXPLAYERS; i++, cl++, plr++)
		if(plr->ingame && cl->age < predict_tics && i != consoleplayer)
		{
			cl->age++;
			if(plr->mo) Cl_MoveThing(plr->mo);			
		}

	// Make the visible buffer a tic older.
	cmbVisible.age++;
	
	// We can't predict too far ahead.
	if(cmbVisible.age > predict_tics) return;

	// Predict the movement of mobjs in the visible buffer.
	for(i=0, mo=cmbVisible.list; i<cmbVisible.num; i++, mo++)
		Cl_MoveThing(mo);
}