#ifndef __GAME_NETWORK_DEF_H__
#define __GAME_NETWORK_DEF_H__

#include "../dd_share.h"

#ifdef __JHERETIC__
#include "../JHeretic/DoomDef.h"
#endif

#define PLR_COLOR(pl, x)	((x)<0 || (x)>3? (pl)%4 : (x))

// Game packet types.
enum
{
	GPT_GAME_STATE = DDPT_FIRST_GAME_EVENT,
	GPT_WEAPON_FIRE,
	GPT_PLANE_MOVE,
	GPT_MESSAGE,			// Non-chat messages.
	GPT_CONSOLEPLAYER_STATE,
	GPT_PLAYER_STATE,
	GPT_PSPRITE_STATE,
	GPT_SOUND,
	GPT_SECTOR_SOUND,
	GPT_FLOOR_MOVE_SOUND,
	GPT_CEILING_MOVE_SOUND,
	GPT_INTERMISSION,
	GPT_FINALE,
	GPT_PLAYER_INFO
};

// This packet is sent by servers to clients when the game state
// changes.
typedef struct
{
	byte	flags;
	byte	episode, map;
	byte	deathmatch : 2;
	byte	monsters : 1;
	byte	respawn : 1;
	byte	jumping : 1;
	byte	skill : 3;
} packet_gamestate_t;

// Game state flags.
#define GSF_CHANGE_MAP		0x01		// Level has changed.
#define GSF_CAMERA_INIT		0x02		// After gamestate follows camera init.

// Player state update flags.
#define PSF_STATE			0x0001		// Dead or alive / armor type.
#define PSF_ARMOR_TYPE		0x0001		// Upper four bits of the 1st byte.
#define PSF_HEALTH			0x0002		
#define PSF_ARMOR_POINTS	0x0004
#define PSF_POWERS			0x0010
#define PSF_KEYS			0x0020
#define PSF_FRAGS			0x0040		// A sum of all the frags.
#define PSF_ALL_FRAGS		0x0080		// Full frags breakdown.
#define PSF_OWNED_WEAPONS	0x0100
#define PSF_AMMO			0x0200
#define PSF_MAX_AMMO		0x0400
#define PSF_COUNTERS		0x0800		// Kill, item and secret counts.
#define PSF_PENDING_WEAPON	0x1000
#define PSF_READY_WEAPON	0x2000

#ifdef __JHERETIC__
#define PSF_INVENTORY		0x0008		// ArtifactCount and invSlotNum, too.
#define PSF_CHICKEN_TIME	0x4000
#endif

// Sound packet flags.
#define SNDF_ORIGIN			0x01		// Sound has an origin.
#define SNDF_SECTOR			0x02		// Originates from a degenmobj.

// Intermission flags.
#define IMF_BEGIN			0x01
#define IMF_END				0x02
#define IMF_STATE			0x04
#define IMF_TIME			0x08

// Finale flags.
#define FINF_BEGIN			0x01

// Networking.
int D_NetServerOpen(int before);
int D_NetServerClose(int before);
int D_NetServerStarted(int before);
int D_NetConnect(int before);
int D_NetDisconnect(int before);
int D_NetPlayerEvent(int plrNumber, int peType, void *data);
int D_NetWorldEvent(int type, int tic, void *data);
void D_HandlePacket(int fromplayer, int type, void *data, int length);

// Console commands.
extern ccmd_t netCCmds[];

#include "d_NetSv.h"
#include "d_NetCl.h"

#endif