#ifdef __JDOOM__
#include "../JDoom/doomdef.h"
#include "../JDoom/doomstat.h"
#include "../JDoom/p_local.h"
#include "../JDoom/s_sound.h"
#include "../JDoom/d_config.h"
#include "../JDoom/g_game.h"
#include "../JDoom/f_finale.h"
#include "../JDoom/wi_stuff.h"
#include "../JDoom/am_map.h"
#endif

#ifdef __JHERETIC__
#include "../JHeretic/DoomDef.h"
#include "../JHeretic/P_local.h"
#include "../JHeretic/soundst.h"
#include "../JHeretic/settings.h"
#endif

#include "d_Net.h"

// Private Data ----------------------------------------------------------

static byte *readbuffer;

// Code ------------------------------------------------------------------

// Mini-Msg routines.
void NetCl_SetReadBuffer(byte *data)
{
	readbuffer = data;
}

byte NetCl_ReadByte()
{
	return *readbuffer++;
}

short NetCl_ReadShort()
{
	readbuffer += 2;
	return *(short*) (readbuffer-2);
}
//-------

void NetCl_UpdateGameState(byte *data)
{
	packet_gamestate_t *gs = (packet_gamestate_t*) data;

	deathmatch = gs->deathmatch;
	nomonsters = !gs->monsters;
	respawnparm = gs->respawn;
	cfg.jumpEnabled = gs->jumping;

	// Some diagnostics.
	gi.Message("Server: map%i, episode%i, skill%i, dm%i, rs%i, mnst%i, jump%i\n",
		gs->map, gs->episode, gs->skill, deathmatch, respawnparm,
		!nomonsters, cfg.jumpEnabled);

#ifdef __JHERETIC__
	prevmap = gamemap;
#endif

	// Do we need to change the map?
	if(gs->flags & GSF_CHANGE_MAP)
	{
		G_InitNew(gs->skill, gs->episode, gs->map);
	}
	else
	{
		gameskill = gs->skill;
		gameepisode = gs->episode;
		gamemap = gs->map;
	}

	// Camera init included?
	if(gs->flags & GSF_CAMERA_INIT)
	{
		player_t *pl = players + consoleplayer;
		mobj_t *mo = pl->plr->mo;
		NetCl_SetReadBuffer(data + sizeof(*gs));
		mo->x = NetCl_ReadShort() << 16;
		mo->y = NetCl_ReadShort() << 16;
		mo->z = NetCl_ReadShort() << 16;
		pl->plr->clAngle = mo->angle = NetCl_ReadShort() << 16;
		pl->plr->viewz = mo->z + pl->plr->viewheight;
	}
}

void NetCl_SpawnMissile(int type, fixed_t x, fixed_t y, fixed_t z, 
						fixed_t momx, fixed_t momy, fixed_t momz)
{
	mobj_t *mis = P_SpawnMobj(x, y, z, type);
	mis->angle = R_PointToAngle2(x, y, x+momx, y+momy);
	mis->momx = momx;
	mis->momy = momy;
	mis->momz = momz;
	// Also play seesound.
	if(mis->info->seesound)
	{
		S_StartSound(mis, mis->info->seesound);
	}

#ifdef __JHERETIC__
	if(type == MT_MACEFX1)
	{
		mis->special1 = 16; // tics till dropoff
	}
#endif
}

void NetCl_UpdatePlayerState(byte *data, int plrNum)
{
	player_t *pl = &players[plrNum];
	byte b;
	unsigned short flags;
	int i;
	int oldstate = pl->playerstate;
#ifdef __JHERETIC__
	unsigned short s;
#endif

	if(!gi.Get(DD_GAME_READY)) return;

	NetCl_SetReadBuffer(data);
	flags = NetCl_ReadShort();

	//gi.conprintf("NetCl_UpdPlrState: fl=%x\n", flags);

	if(flags & PSF_STATE) // and armor type
	{
		b = NetCl_ReadByte();
		pl->playerstate = b & 0xf;
		pl->armortype = b >> 4;
	}
	if(flags & PSF_HEALTH) 
	{
		pl->health = NetCl_ReadByte();
		pl->plr->mo->health = pl->health;
	}
	if(flags & PSF_ARMOR_POINTS) pl->armorpoints = NetCl_ReadByte();

#ifdef __JHERETIC__
	if(flags & PSF_INVENTORY)
	{
		pl->inventorySlotNum = NetCl_ReadByte();
		pl->artifactCount = 0;
		for(i=0; i<NUMINVENTORYSLOTS; i++)
		{
			if(i >= pl->inventorySlotNum)
			{
				pl->inventory[i].type = arti_none;
				pl->inventory[i].count = 0;
				continue;
			}
			s = NetCl_ReadShort();
			pl->inventory[i].type = s & 0xff;
			pl->inventory[i].count = s >> 8;
			if(pl->inventory[i].type != arti_none)
				pl->artifactCount += pl->inventory[i].count;
		}
		if(plrNum == consoleplayer)	P_CheckReadyArtifact();
	}
#endif

	if(flags & PSF_POWERS)
	{
		b = NetCl_ReadByte();
		for(i=1; i<NUMPOWERS; i++)
			if(b & (1 << (i-1)))
			{
				pl->powers[i] = NetCl_ReadByte() * 35;										
			}
	}
	if(flags & PSF_KEYS)
	{
		b = NetCl_ReadByte();
#ifdef __JDOOM__
		for(i=0; i<NUMCARDS; i++) pl->cards[i] = (b & (1<<i)) != 0;
#endif

#ifdef __JHERETIC__
		for(i=0; i<NUMKEYS; i++) pl->keys[i] = (b & (1<<i)) != 0;
#endif
	}
	if(flags & PSF_FRAGS)
	{
		pl->frags[consoleplayer] = NetCl_ReadShort();
	}
	if(flags & PSF_ALL_FRAGS)
	{
		for(i=0; i<MAXPLAYERS; i++) pl->frags[i] = NetCl_ReadShort();
	}
	if(flags & PSF_OWNED_WEAPONS)
	{
		b = NetCl_ReadByte();
		for(i=0; i<8; i++) pl->weaponowned[i] = (b & (1<<i)) != 0;
	}
	if(flags & PSF_AMMO)
	{
		for(i=0; i<NUMAMMO; i++) pl->ammo[i] = NetCl_ReadShort();
	}
	if(flags & PSF_MAX_AMMO)
	{
		for(i=0; i<NUMAMMO; i++) pl->maxammo[i] = NetCl_ReadShort();
	}
	if(flags & PSF_COUNTERS)
	{
		pl->killcount = NetCl_ReadShort();
		pl->itemcount = NetCl_ReadByte();
		pl->secretcount = NetCl_ReadByte();

		/*gi.conprintf( "plr%i: kills=%i items=%i secret=%i\n", pl-players,
			pl->killcount, pl->itemcount, pl->secretcount);*/
	}
	if(flags & PSF_PENDING_WEAPON || flags & PSF_READY_WEAPON)
	{
		b = NetCl_ReadByte();
		if(flags & PSF_PENDING_WEAPON) pl->pendingweapon = b & 0xf;
		if(flags & PSF_READY_WEAPON) pl->readyweapon = b >> 4;
	}

#ifdef __JHERETIC__
	if(flags & PSF_CHICKEN_TIME)
	{
		pl->chickenTics = NetCl_ReadByte() * 35;
	}
#endif

	if(oldstate != pl->playerstate && oldstate == PST_DEAD)
	{
		P_SetupPsprites(pl);
	}
}

void NetCl_UpdatePSpriteState(byte *data)
{
	unsigned short s;

	NetCl_SetReadBuffer(data);
	s = NetCl_ReadShort();
	P_SetPsprite(&players[consoleplayer], ps_weapon, s);
}

void NetCl_PlaySound(byte *data)
{
	int sound;
	fixed_t x, y, z;
	byte b;
	mobj_t *mo = NULL;

	NetCl_SetReadBuffer(data);
	b = NetCl_ReadByte();
	sound = NetCl_ReadByte();
	//gi.conprintf( "NetCl_Sound: id=%i\n", sound);
	if(b & SNDF_ORIGIN)
	{
		x = NetCl_ReadShort() << 16;
		y = NetCl_ReadShort() << 16;
		z = NetCl_ReadShort() << 16;
		//gi.conprintf( "- origin: %i %i %i\n", x>>16, y>>16, z>>16);
		// We need to create a temporary mobj to act as the sound origin.
		// An invisible barrel is nice.
		mo = P_SpawnMobj(x, y, z, MT_BARREL);
		P_UnsetThingPosition(mo);	// It doesn't have to be in any lists.
		mo->height = 1;
		mo->flags &= ~MF_SOLID;
		mo->flags |= MF_NOGRAVITY | MF_NOSECTOR | MF_NOBLOCKMAP;

#ifdef __JHERETIC__
		mo->flags2 |= MF2_DONTDRAW;
#endif

		mo->tics = 35*4;		// Sounds are definitely over in 4 seconds.
	}
	S_StartSound(mo, sound);
}

void NetCl_PlaySectorSound(byte *data)
{
	int sound, sector;

	NetCl_SetReadBuffer(data);
	sound = NetCl_ReadByte();
	sector = NetCl_ReadShort();
	S_StartSound( (mobj_t*) &sectors[sector].soundorg, sound);
}

void NetCl_Intermission(byte *data)
{
	int	flags;
	extern int interstate;
	extern int intertime;

	NetCl_SetReadBuffer(data);
	flags = NetCl_ReadByte();

	//gi.conprintf( "NetCl_Intermission: flags=%x\n", flags);

#ifdef __JDOOM__
	if(flags & IMF_BEGIN)
	{
		wminfo.maxkills = NetCl_ReadShort();
		wminfo.maxitems = NetCl_ReadShort();
		wminfo.maxsecret = NetCl_ReadShort();
		wminfo.next = NetCl_ReadByte();
		wminfo.last = NetCl_ReadByte();
		wminfo.didsecret = NetCl_ReadByte();

		G_PrepareWIData();

		gamestate = GS_INTERMISSION;
		viewactive = false; 
		if(automapactive) AM_Stop(); 

		WI_Start(&wminfo);
	}
	if(flags & IMF_END)
	{
		WI_End();
	}
	if(flags & IMF_STATE)
	{
		WI_SetState(NetCl_ReadByte());
	}
#endif

#ifdef __JHERETIC__
	if(flags & IMF_STATE) interstate = (int) NetCl_ReadByte();
	if(flags & IMF_TIME) intertime = NetCl_ReadShort();
	if(flags & IMF_BEGIN) 
	{
		gamestate = GS_INTERMISSION;
		IN_Start();
	}
	if(flags & IMF_END) 
	{
		IN_Stop();
	}
#endif
}

void NetCl_Finale(byte *data)
{
	int flags;

	NetCl_SetReadBuffer(data);
	flags = NetCl_ReadByte();
	if(flags & FINF_BEGIN) F_StartFinale();
}

// Clients have other players' info, but it's only "FYI"; they don't
// really need it.
void NetCl_UpdatePlayerInfo(byte *data)
{
	int num;

	NetCl_SetReadBuffer(data);
	num = NetCl_ReadByte();
	cfg.PlayerColor[num] = NetCl_ReadByte();

	gi.conprintf( "NetCl_UpdatePlayerInfo: pl=%i color=%i\n", num, 
		cfg.PlayerColor[num]);
}

// Send consoleplayer's settings to the server.
void NetCl_SendPlayerInfo()
{
	byte buffer[10], *ptr = buffer;
	
	if(!IS_CLIENT) return;

	*ptr++ = cfg.netColor;
	gi.SendPacket(DDSP_RELIABLE, GPT_PLAYER_INFO, buffer, ptr-buffer);
}