#ifdef __JDOOM__
#include "../JDoom/doomdef.h"
#include "../JDoom/doomstat.h"
#include "../JDoom/p_local.h"
#include "../JDoom/d_config.h"
#include "../JDoom/s_sound.h"
#include "../JDoom/st_stuff.h"
#endif

#ifdef __JHERETIC__
#include "../JHeretic/DoomDef.h"
#include "../JHeretic/P_local.h"
#include "../JHeretic/settings.h"
#endif

#include "d_Net.h"

// MACROS -----------------------------------------------------------------

#define UPD_BUFFER_LEN	500

// How long is the largest possible sector update?
#define MAX_SECTORUPD	20
#define MAX_SIDEUPD		9

#define WRITE_SHORT(byteptr, val)	{(*(short*)(byteptr) = (val)); byteptr += 2;}

// TYPES ------------------------------------------------------------------

// EXTERNAL FUNCTIONS -----------------------------------------------------

void R_SetAllDoomsdayFlags();

// PRIVATE DATA -----------------------------------------------------------

// CODE -------------------------------------------------------------------

void P_Telefrag(mobj_t *thing)
{
	P_TeleportMove(thing, thing->x, thing->y);
}

// Server calls this when new players enter the game.
void NetSv_NewPlayerEnters(int plrnumber)
{
	player_t *plr = &players[plrnumber];
	
	gi.Message( "NetSv_NewPlayerEnters: spawning player %i.\n", plrnumber);

	plr->playerstate = PST_REBORN;	// Force an init.

	// Re-deal player starts.
	P_DealPlayerStarts();

	// Spawn the player into the world.
	// FIXME: Spawn a telefog in front of the player.
	P_SpawnPlayer(&playerstarts[plr->startspot], plrnumber);
	// Get rid of anybody at the starting spot.
	P_Telefrag(plr->plr->mo);
}

void NetSv_SendMessage(int plrNum, char *msg)
{
	if(!netgame || !server || plrNum == consoleplayer) return;
	if(!players[plrNum].plr->ingame) return;
	gi.SendPacket(plrNum | DDSP_RELIABLE, GPT_MESSAGE,
		msg, strlen(msg)+1);
}

void NetSv_SendPlayerState(int srcPlrNum, int destPlrNum, int flags, boolean reliable)
{
	int pType = srcPlrNum==destPlrNum? GPT_CONSOLEPLAYER_STATE : GPT_PLAYER_STATE;
	player_t *pl = &players[srcPlrNum];
	byte buffer[UPD_BUFFER_LEN], *ptr = buffer, fl;
	int i, k;

	if(!netgame || !server || srcPlrNum == consoleplayer 
		|| destPlrNum == consoleplayer || !pl->plr->ingame
		|| (destPlrNum >= 0 && destPlrNum < MAXPLAYERS 
		&& !players[destPlrNum].plr->ingame)) return;

	// Include the player number if necessary.
	if(pType == GPT_PLAYER_STATE) *ptr++ = srcPlrNum;

	// The first byte contains the flags.
	WRITE_SHORT(ptr, flags);
	if(flags & PSF_STATE) 
		*ptr++ = pl->playerstate | (pl->armortype << 4);
	if(flags & PSF_HEALTH) *ptr++ = pl->health;
	if(flags & PSF_ARMOR_POINTS) *ptr++ = pl->armorpoints;

#ifdef __JHERETIC__
	if(flags & PSF_INVENTORY)
	{
		*ptr++ = pl->inventorySlotNum;
		for(i=0; i<pl->inventorySlotNum; i++)
			WRITE_SHORT(ptr, (pl->inventory[i].type & 0xff) 
				| ((pl->inventory[i].count & 0xff) << 8));
	}
#endif

	if(flags & PSF_POWERS) // Austin!
	{
		// First see which powers should be sent.
		for(i=1, *ptr=0; i<NUMPOWERS; i++)
			if(pl->powers[i]) *ptr |= 1 << (i-1);
		ptr++;
		for(i=1; i<NUMPOWERS; i++)
			if(pl->powers[i])
				*ptr++ = pl->powers[i]/35 + 1;	// Send as seconds.
	}
	if(flags & PSF_KEYS) 
	{
		*ptr = 0;

#ifdef __JDOOM__
		for(i=0; i<NUMCARDS; i++) if(pl->cards[i]) *ptr |= 1<<i;
#endif

#ifdef __JHERETIC__
		for(i=0; i<NUMKEYS; i++) if(pl->keys[i]) *ptr |= 1<<i;
#endif

		ptr++;
	}
	if(flags & PSF_FRAGS)
	{
		for(k=0, i=0; i<MAXPLAYERS; i++) k += pl->frags[i];
		WRITE_SHORT(ptr, k);
	}
	if(flags & PSF_ALL_FRAGS)
	{
		for(i=0; i<MAXPLAYERS; i++) WRITE_SHORT(ptr, pl->frags[i]);
	}
	if(flags & PSF_OWNED_WEAPONS)
	{
		for(fl=0, i=0; i<8; i++) if(pl->weaponowned[i]) fl |= 1<<i;
		*ptr++ = fl;
	}
	if(flags & PSF_AMMO)
	{
		for(i=0; i<NUMAMMO; i++) WRITE_SHORT(ptr, pl->ammo[i]);
	}
	if(flags & PSF_MAX_AMMO)
	{
		for(i=0; i<NUMAMMO; i++) WRITE_SHORT(ptr, pl->maxammo[i]);
	}
	if(flags & PSF_COUNTERS)
	{
		WRITE_SHORT(ptr, pl->killcount);
		*ptr++ = pl->itemcount;
		*ptr++ = pl->secretcount;
	}
	if(flags & PSF_PENDING_WEAPON || flags & PSF_READY_WEAPON)
	{
		// These two will be in the same byte.
		fl = 0;
		if(flags & PSF_PENDING_WEAPON) fl |= pl->pendingweapon & 0xf;
		if(flags & PSF_READY_WEAPON) fl |= (pl->readyweapon & 0xf) << 4;
		*ptr++ = fl;
	}

#ifdef __JHERETIC__
	if(flags & PSF_CHICKEN_TIME)
	{
		*ptr++ = pl->chickenTics/35 + 1;
	}
#endif

	/*gi.conprintf( "PlrUpd (%i bytes, flags=%x):\n", ptr-buffer, flags);
	for(i=0; i<ptr-buffer; i++)
	{
		gi.conprintf( "%02x ", buffer[i]);
		if(i && !(i%10)) gi.conprintf( "\n");
	}
	gi.conprintf( "\n");*/

	// Finally, send the packet.
	gi.SendPacket(destPlrNum | (reliable? DDSP_RELIABLE : 0),
		pType, buffer, ptr-buffer);	
}

void NetSv_PSpriteChange(int plrNum, int state)
{
	byte buffer[20], *ptr = buffer;

	if(!netgame || !server) return;

	WRITE_SHORT(ptr, state);
	gi.SendPacket(plrNum, GPT_PSPRITE_STATE, buffer, ptr-buffer);
}

void NetSv_Sound(mobj_t *origin, int sound_id, int toPlr)
{
	int i;
	byte buffer[20], *ptr;
	boolean not;
	int dest;

	if(!netgame || !server) return;

	ptr = buffer;
	not = (toPlr & NSSF_NOT) != 0;
	dest = toPlr & NSSF_MASK;

	*ptr++ = origin? SNDF_ORIGIN : 0;
	*ptr++ = sound_id;
	if(origin)
	{
		WRITE_SHORT(ptr, origin->x >> 16);
		WRITE_SHORT(ptr, origin->y >> 16);
		WRITE_SHORT(ptr, origin->z >> 16);
	}
	// Send to the appropriate clients.
	for(i=1; i<MAXPLAYERS; i++)
	{
		if(!players[i].plr->ingame) continue;
		if(!toPlr || (i == dest) != not)
		{
			mobj_t *plrMo = players[i].plr->mo;
			// Check the distance.
			if(origin && P_AproxDistance(origin->x - plrMo->x, 
				origin->y - plrMo->y) > FRACUNIT*MAX_SND_DIST) continue;
			// We can send to this client.
			gi.SendPacket(i, GPT_SOUND, buffer, ptr-buffer);
		} 
	}
}

void NetSv_SectorSound(sector_t *sector, int sound_id)
{
	byte buffer[20], *ptr = buffer;

	if(!netgame || !server) return;

	*ptr++ = sound_id;
	WRITE_SHORT(ptr, sector-sectors);
	gi.SendPacket(DDSP_ALL_PLAYERS, GPT_SECTOR_SOUND, buffer, ptr-buffer);
}

void NetSv_SendGameState(int flags, int to)
{
	byte buffer[30], *ptr = buffer;
	packet_gamestate_t *gs = (packet_gamestate_t*) ptr;

	if(!server) return;

	// We can't init the camera if we don't know which 
	// player this packet is going to.
	if(to < 1 || to >= MAXPLAYERS) flags &= ~GSF_CAMERA_INIT;

	gs->flags = flags;
	gs->episode = gameepisode;
	gs->map = gamemap;
	gs->skill = gameskill;
	gs->deathmatch = deathmatch;
	gs->monsters = !nomonsters;
	gs->respawn = respawnparm;
	gs->jumping = cfg.jumpEnabled;
	ptr += sizeof(packet_gamestate_t);

	if(flags & GSF_CAMERA_INIT)
	{
		mobj_t *mo = players[to].plr->mo;
		WRITE_SHORT(ptr, mo->x >> 16);
		WRITE_SHORT(ptr, mo->y >> 16);
		WRITE_SHORT(ptr, mo->z >> 16);
		WRITE_SHORT(ptr, mo->angle >> 16);
	}

	// Send the packet.
	gi.SendPacket(to | DDSP_RELIABLE, GPT_GAME_STATE, buffer, ptr-buffer);
}

void NetSv_Intermission(int flags, int state, int time)
{
	byte buffer[32], *ptr = buffer;

	if(!server) return;

	*ptr++ = flags;

#ifdef __JDOOM__
	if(flags & IMF_BEGIN)
	{
		// Only include the necessary information.
		WRITE_SHORT(ptr, wminfo.maxkills);
		WRITE_SHORT(ptr, wminfo.maxitems);
		WRITE_SHORT(ptr, wminfo.maxsecret);
		*ptr++ = wminfo.next;
		*ptr++ = wminfo.last;
		*ptr++ = wminfo.didsecret;
	}
#endif

	if(flags & IMF_STATE) *ptr++ = state;
	if(flags & IMF_TIME) WRITE_SHORT(ptr, time);
	gi.SendPacket(DDSP_ALL_PLAYERS | DDSP_RELIABLE, GPT_INTERMISSION,
		buffer, ptr-buffer);
}

void NetSv_Finale(int flags)
{
	byte buffer[8], *ptr = buffer;

	if(!server) return;

	*ptr++ = flags;
	gi.SendPacket(DDSP_ALL_PLAYERS | DDSP_RELIABLE, GPT_FINALE,
		buffer, ptr-buffer);
}

void NetSv_SendPlayerInfo(int whose, int to_whom)
{
	byte buffer[10], *ptr = buffer;

	if(!server) return;

	*ptr++ = whose;
	*ptr++ = cfg.PlayerColor[whose];
	gi.SendPacket(to_whom | DDSP_RELIABLE, GPT_PLAYER_INFO,
		buffer, ptr-buffer);
}

void NetSv_ChangePlayerInfo(int from, byte *data)
{
	player_t *pl = players + from;
	int col;

	// Color is first.
	col = *data++;
	cfg.PlayerColor[from] = PLR_COLOR(from, col);

	gi.conprintf( "NetSv_ChangePlayerInfo: pl%i, col=%i\n", from, 
		cfg.PlayerColor[from]);

	if(pl->plr->mo)
	{
		// Change the player's mobj.
		pl->plr->mo->flags &= ~MF_TRANSLATION;
		pl->plr->mo->flags |= col << MF_TRANSSHIFT;
	}

	// Re-deal start spots.
	P_DealPlayerStarts();

	// Tell the other clients about the change.
	NetSv_SendPlayerInfo(from, DDSP_ALL_PLAYERS);
}

void NetSv_Ticker(void)
{
	player_t *plr;
	int	i, palette;

	// This is done here for servers.
	R_SetAllDoomsdayFlags();
	
	// Set the camera filters for players.
	// This is from SB_PaletteFlash().
	for(i=1; i<MAXPLAYERS; i++)
	{
		if(!players[i].plr->ingame) continue;

		plr = &players[i];
		if(plr->damagecount)
		{
			palette = (plr->damagecount+7)>>3;
			if(palette >= NUMREDPALS)
			{
				palette = NUMREDPALS-1;
			}
			palette += STARTREDPALS;
		}
		else if(plr->bonuscount)
		{
			palette = (plr->bonuscount+7)>>3;
			if(palette >= NUMBONUSPALS)
			{
				palette = NUMBONUSPALS-1;
			}
			palette += STARTBONUSPALS;
		}
		else
		{
			palette = 0;
		}
#ifdef __JDOOM__
		plr->plr->filter = D_GetFilterColor(palette);
#endif

#ifdef __JHERETIC__
		plr->plr->filter = H_GetFilterColor(palette);
#endif
	}

	// Send the player state updates.
	for(i=1; i<MAXPLAYERS; i++)
	{
		if((gametic+i) % 10) continue; // Don't send on every tic.
		if(!players[i].plr->ingame || !players[i].update) continue;
		// Always send reliably (unwise?).
		NetSv_SendPlayerState(i, i, players[i].update, true);
		players[i].update = 0;
	}
}

