//**************************************************************************
//**
//** net_main.c :	Doomsday Networking
//** Author:		Jaakko Kernen
//** 
//** Client/server networking. 
//** Player number zero is always the server.
//** In single-player games there is only the server present.
//** Some cleaning up wouldn't hurt.
//**
//**************************************************************************

#include "dd_def.h"
#include "dd_net.h"
#include "i_win32.h"
#include "i_timer.h"
#include "i_net.h"
#include "console.h"
#include "cl_def.h"
#include "sv_def.h"
#include <stdlib.h> // for atoi()


// Don't run too many tics; if the FPS goes below 4, start
// slowing down.
#define MAXRUNTICS	9


byte			*localticcmds;

int				resend_start = 0;			// set when server needs our tics
int				resend_count = 0;

boolean			allow_net_traffic = true;	// Should net traffic be allowed?

ddplayer_t		players[MAXPLAYERS];
client_t		clients[MAXPLAYERS];		// All network data for the players.

int				netgame;	// true if a netgame is in progress
int				server;		// true if this computer is an open server.
int				consoleplayer;
int				displayplayer;

int             gametic, maketic;

int             skiptics;

int				gametime;
int				oldentertics;

boolean			firstRunTic = true;	// First tic in the whole game?
boolean			firstNetUpdate = true;

boolean			monitorSendQueue = false;
boolean			net_timerefresh = false;


extern void ST_NetProgress(void);
extern void ST_NetDone(void);
	

// Local packets are stored into this buffer.
boolean		reboundpacket;
netbuffer_t	reboundstore;


cvar_t netCVars[] =
{
	"net_msq",			0,	CVT_BYTE, &monitorSendQueue,	0, 1,	"Monitor send queue.",
	"net_timerefresh",	0,	CVT_BYTE, &net_timerefresh,		0, 1,	"Time client refresh.",
	NULL
};

/*int NetbufferSize (void)
{
	//return (int)&(((doomdata_t *)0)->cmds[netbuffer->numtics]);
	return (int)&(((doomdata_t *)0)->ticcmds[TICCMD_IDX(netbuffer->numtics)]);
}*/

/*unsigned NetbufferChecksum (void)
{
	unsigned                c;
	int             i,l;

	c = 0x1234567;

#if defined(NeXT) || defined(NORMALUNIX)
	return 0;                       // byte order problems
#endif

	l = (NetbufferSize () - (int)&(((doomdata_t *)0)->packettype))/4;
	for (i=0 ; i<l ; i++)
		c += ((unsigned *)&netbuffer->packettype)[i] * (i+1);

	return c & NCMD_CHECKSUM;
}

int ExpandTics (int low)
{
	int     delta;

	delta = low - (maketic&0xff);

	if (delta >= -64 && delta <= 64)
		return (maketic&~0xff) + low;
	if (delta > 64)
		return (maketic&~0xff) - 256 + low;
	if (delta < -64)
		return (maketic&~0xff) + 256 + low;

	I_Error ("ExpandTics: strange value %i at maketic %i",low,maketic);
	return 0;
}*/


//============================================================================

void Net_SendPacket(int to_player, int sp_flags)
{
	// Update the length of the message.
	netbuffer.length = netbuffer.cursor - netbuffer.msg.data;
	netbuffer.player = to_player;

	// Local packets rebound. Clients can't send packets to themselves.
	if((!netgame || server) && to_player == consoleplayer) 
	{
		reboundstore = netbuffer;
		reboundpacket = true;
		return;
	}

	if (!allow_net_traffic)
		return;

	I_SendPacket(sp_flags);
}

//==========================================================================
//
// NET_SendFrags
//
//==========================================================================

/*
==============
=
= HGetPacket
=
= Returns false if no packet is waiting
=
==============
*/

boolean Net_GetPacket(void)
{
	if(reboundpacket) // Local packets rebound.
	{
		netbuffer = reboundstore;
		netbuffer.player = consoleplayer;
		netbuffer.cursor = netbuffer.msg.data;
		reboundpacket = false;
		return true;
	}
	if(!netgame || !allow_net_traffic)
	{
		// Packets cannot be received.
		return false;
	}
	if(!I_GetPacket()) return false;

	// Reset the cursor for Msg_* routines.
	netbuffer.cursor = netbuffer.msg.data;

/*if (debugfile)
{
	int             realretrans;
			int     i;

	if (netbuffer->checksum & NCMD_SETUP)
		fprintf (debugfile,"setup packet\n");
	else
	{
		if (netbuffer->checksum & NCMD_RETRANSMIT)
			realretrans = ExpandTics (netbuffer->packettype);
		else
			realretrans = -1;
		fprintf (debugfile,"get %i = (%i + %i, R %i)[%i] ",doomcom->remotenode,
		ExpandTics(netbuffer->starttic),netbuffer->numtics, realretrans, doomcom->datalength);
		for (i=0 ; i<doomcom->datalength ; i++)
			fprintf (debugfile,"%i ",((byte *)netbuffer)[i]);
		fprintf (debugfile,"\n");
	}
}*/
	return true;
}

// This is the public interface of the message sender.
void D_SendPacket(int to_player, int type, void *data, int length)
{
	boolean reliable = (to_player & DDSP_RELIABLE) != 0;

	Msg_Begin(type);
	if(data) Msg_Write(data, length);

	if(server)
	{
		// The server can send packets to any player.
		// Only allow sending to the sixteen possible players.
		Net_SendPacket(to_player & DDSP_ALL_PLAYERS? NSP_BROADCAST 
			: to_player & 0xf, reliable? SPF_RELIABLE : 0); 
	}
	else
	{
		// As a client we can only send messages to the server.
		Net_SendPacket(0, reliable? SPF_RELIABLE : 0);
	}
}

// Prints the message in the console.
void D_ShowChatMessage()
{
	// The current packet in the netbuffer is a chat message,
	// let's unwrap and show it.
	CON_FPrintf(CBLF_GREEN, "%s: %s\n", 
		clients[netbuffer.msg.data[0]].name, 
		netbuffer.msg.data+3);
}

// All arguments are sent out as a chat message.
int CCmdChat(int argc, char **argv)
{
	char buffer[100];
	int	i, mode = !stricmp(argv[0], "chat") || !stricmp(argv[0], "say")? 0 
		: !stricmp(argv[0], "chatNum") || !stricmp(argv[0], "sayNum")? 1 : 2;
	unsigned short mask = 0;

	if(argc == 1)
	{
		CON_Printf( "Usage: %s %s(text)\n", argv[0],
			!mode? "" : mode==1? "(plr#) " : "(name) ");
		CON_Printf( "Chat messages are max. 80 characters long.\n");
		CON_Printf( "Use quotes to get around arg processing.\n");
		return true;
	}

	// Chatting is only possible when connected.
	if(!netgame) return false;

	// Too few arguments?
	if(mode && argc < 3) return false;

	// Assemble the chat message.
	strcpy(buffer, argv[!mode? 1 : 2]);
	for(i = (!mode? 2 : 3); i<argc; i++)
	{
		strcat(buffer, " ");
		strncat(buffer, argv[i], 80 - (strlen(buffer)+strlen(argv[i])+1));
	}
	buffer[80] = 0;

	// Send the message.
	switch(mode)
	{
	case 0: // chat
		mask = ~0;
		break;

	case 1: // chatNum
		mask = 1 << atoi(argv[1]);
		break;

	case 2: // chatTo
		for(i=0; i<MAXPLAYERS; i++)
			if(!stricmp(clients[i].name, argv[1]))
			{
				mask = 1 << i;
				break;
			}
	}
	Msg_Begin(pkt_chat);
	Msg_WriteByte(consoleplayer);
	Msg_WriteShort(mask);
	Msg_Write(buffer, strlen(buffer)+1);

	if(server)
	{
		if(mask == ~0) 
			Net_SendPacket(NSP_BROADCAST, SPF_RELIABLE);
		else 
			for(i=1; i<MAXPLAYERS; i++) if(mask & (1 << i)) 
				Net_SendPacket(i, SPF_RELIABLE);
	}
	else
	{
		Net_SendPacket(0, SPF_RELIABLE);
	}
	
	// Show the message locally.
	D_ShowChatMessage();

	// Inform the game, too.
	gx.NetPlayerEvent(consoleplayer, DDPE_CHAT_MESSAGE, buffer);
	return true;
}

/*static void HandlePacket()
{
	int type = netbuffer.msg.type;

	// A chat message?
	switch(netbuffer.msg.type)
	{
	case pkt_chat:
		if(server)
		{
			// Servers are responsible for relaying the chat messages
			// to all the recipients.
			
		}

		ShowChatMessage();
		if(gx.NetPlayerEvent) 
			gx.NetPlayerEvent(netbuffer->player, DDPE_CHAT_MESSAGE,
				netbuffer->ticcmds+2);
		break;
	}
}
*/


/*
=============
=
= Net_Update
=
= Builds ticcmds for console player
= sends out a packet
=============
*/


void Net_Update(void)
{
	int nowtime;
	int newtics;
	int	i;
	int realstart, numtics;

//
// check time
//
	nowtime = I_GetTime();
	if(firstNetUpdate)
	{
		firstNetUpdate = false;
		gametime = nowtime;
	}
	newtics = nowtime - gametime;
	gametime = nowtime;

	if (newtics <= 0)                       // nothing new to update
		goto listen;

	if (skiptics <= newtics)
	{
		newtics -= skiptics;
		skiptics = 0;
	}
	else
	{
		skiptics -= newtics;
		newtics = 0;
	}

	// Build new ticcmds for console player.
	for(i=0; i<newtics; i++)
	{
		I_StartTic();
		DD_ProcessEvents();

		if(debugfile) 
			fprintf(debugfile, "mktic:%i gt:%i newtics:%i >> %i\n", 
				maketic, gametic, newtics, maketic-gametic);
		
		if(maketic - gametic >= BACKUPTICS/2-1)
		{
			if(debugfile)
				fprintf(debugfile, "--- NetUpdate: backuptics full\n");
			break;          // can't hold any more
		}

		//if(debugfile) fprintf(debugfile, "mk:%i\n",maketic);
		//gx.BuildTiccmd (&localcmds[maketic%BACKUPTICS]);
		
		// Place the new ticcmd in the local ticcmds buffer.
		if(!isDedicated)
			gx.BuildTicCmd(&localticcmds[TICCMD_IDX(maketic%BACKUPTICS)]);

		maketic++;
	}

	// This is as far as dedicated servers go.
	if(isDedicated) goto listen;

	if (singletics)
		return;         // singletic update is syncronous

	if(!netgame || server || I_CheckSendQueue(0))
	{
		// Send the proper number of tics to the server.
		// If we are the server, the tics will rebound back to us.
		realstart = resend_start;
		numtics = maketic - resend_start;	// Send this many of them.
		if(numtics > BACKUPTICS) 
		{
			// Sod that.
			numtics = BACKUPTICS; 
		}
		
		// Reset the send marker. 
		resend_start = maketic;			
		
		Msg_Begin(pkt_ticcmd);
		Msg_WriteByte(numtics);
		
		// Copy the tics to the outgoing packet.
		for(i=0; i<numtics; i++)
		{
			Msg_Write(localticcmds + TICCMD_IDX((realstart+i)%BACKUPTICS), 
				TICCMD_SIZE);
		}
		if(debugfile)
			fprintf(debugfile, "sending tics %i:%i to server.\n", realstart, numtics);
		
		// Send the packet to the server, player zero.
		Net_SendPacket(0, 0);
	}
//
// listen for other packets
//
listen:
	// Call the correct packet handler.
	if(!netgame || server) // Single-player or server.
		Sv_GetPackets();
	else
		Cl_GetPackets();
}


/*
=====================
=
= CheckAbort
=
=====================
*/

void CheckAbort (void)
{
	event_t *ev;
	int             stoptic;

	stoptic = I_GetTime () + 2;
	while (I_GetTime() < stoptic)
		I_StartTic ();

	I_StartTic ();
	for ( ; eventtail != eventhead
	; eventtail = (++eventtail)&(MAXEVENTS-1) )
	{
		ev = &events[eventtail];
		if (ev->type == ev_keydown && ev->data1 == DDKEY_ESCAPE)
			I_Error ("Network game synchronization aborted.");
	}
}

// Called from I_InitNetwork to initialize the ticcmd arrays.
void Net_AllocArrays(void)
{
	int		i;

	// Local ticcmds are stored into this array before they're copied
	// to netplayer[0]'s ticcmds buffer.
	localticcmds = calloc(BACKUPTICS, TICCMD_SIZE);

	for(i=0; i<MAXPLAYERS; i++)
	{
		memset(clients+i, 0, sizeof(clients[i]));
		// The server stores ticcmds sent by the clients to these
		// buffers.
		clients[i].ticcmds = calloc(BACKUPTICS, TICCMD_SIZE);
	}
}


//
// This is the network one-time initialization.
// (into single-player mode)
//
void Net_InitGame (void)
{
	resend_start = 0;

	// In single-player mode there is only player number zero.
	consoleplayer = displayplayer = 0; 

	players[0].ingame = true;
	clients[0].ready = true;
	clients[0].connected = true;
}

void Net_StopGame (void)
{
	int		i;

	if(server)
	{
		// We are a server. This means we should inform all the 
		// connected clients that the server is about to stop.
		Msg_Begin(psv_server_close);
		// Send a bunch to be sure.
		Net_SendPacket(NSP_BROADCAST, SPF_RELIABLE);
		server = false;
	}
	else
	{
		Cl_CleanUp();
	}

	// Netgame has ended.
	netgame = false;

	// All remote players are forgotten.
	for(i=0; i<MAXPLAYERS; i++) 
	{
		players[i].ingame = false;
		clients[i].ready = clients[i].connected = false;
	}
	consoleplayer = displayplayer = 0;
	players[0].ingame = true;
	clients[0].ready = true;	
	clients[0].connected = true;
}


int D_GetTicCmd(void *cmd, int player)
{
	client_t *cl = &clients[player];

/*	if(debugfile)
	{
		fprintf(debugfile, "** GetTicCmd **  plr:%i  gametic:%i  nettic:%i\n",
			player, gametic, clients[player].tics);
	}
*/	
	// Are there any commands left for this player?
	if(!cl->numtics) 
	{
		memset(cmd, 0, TICCMD_SIZE);
		return false;	
	}
	// There will be one less tic in the buffer after this.
	cl->numtics--;
	// There's at least one command in the buffer.
	memcpy(cmd, &cl->ticcmds[TICCMD_IDX(cl->firsttic++)], TICCMD_SIZE);
	// Make sure the firsttic iterator is in range.
	if(cl->firsttic >= BACKUPTICS) cl->firsttic = 0;
	return true;
}

void Net_Ticker(void)
{
	int		i;
	client_t *cl;
	static int monitorCount = 0;

	if(monitorSendQueue && monitorCount-- <= 0)
	{
		int num, bytes;
		monitorCount = MONITORTICS;
		num = jtNetCheckQueue(clients[!consoleplayer].jtNetNode, &bytes);
		CON_Printf( "Queue to %i: %i packets, %i bytes.\n",
			!consoleplayer, num, bytes);
	}

	// Check the pingers.
	for(i=0, cl=clients; i<MAXPLAYERS; i++, cl++)
	{
		// Clients can only ping the server.
		if(!server && i || i == consoleplayer) continue;
		if(cl->ping.sent)
		{
			// The pinger is active.
			if(I_GetRealTime() - cl->ping.sent > PING_TIMEOUT) // Timed out?
			{
				cl->ping.times[cl->ping.current] = -1;
				Net_SendPing(i, 0);
			}
		}
	}
	
	if(server) Sv_Ticker(); else Cl_Ticker();
}

/*
===============
=
= TryRunTics
=
===============
*/

extern  boolean advancedemo;

void TryRunTics (void)
{
	//int 			i;
	int 			entertic;
	int 			realtics, availabletics;
	int 			counts;
//	int 			numplaying;

//
// get real tics
//
	entertic = I_GetTime();
	if(firstRunTic)
	{
		firstRunTic = false;
		oldentertics = entertic;
	}
	realtics = entertic - oldentertics;
	oldentertics = entertic;

	if(!realtics)
	{
		// No tics passed since last time? We can sleep for a while.
		while(!realtics)
		{
			I_Sleep(5);
			entertic = I_GetTime();
			realtics = entertic - oldentertics;
			oldentertics = entertic;
		}
	}

//
// get available tics
//
	Net_Update();

/*	lowtic = DDMAXINT;
	numplaying = 0;
	for (i=0 ; i<doomcom->numnodes ; i++)
		if (nodeingame[i])
		{
			numplaying++;
			if (nettics[i] < lowtic)
				lowtic = nettics[i];
		}
	availabletics = lowtic - gametic/ticdup;*/

	//availabletics = clients[consoleplayer].tics - gametic;
	
	/*if(netgame && !server)*/ 
	availabletics = realtics;

//
// decide how many tics to run
//
/*	if (realtics < availabletics-1)
		counts = realtics+1;
	else if (realtics < availabletics)
		counts = realtics;
	else
		counts = availabletics;
	if (counts < 1)
		counts = 1;*/

	counts = availabletics;
	//if(realtics < counts) counts = realtics;

	if(counts > MAXRUNTICS)	// *EASY*, big fellah!
		counts = MAXRUNTICS;

	if (debugfile)
	{
		fprintf (debugfile,"=======real: %i  nettics: %i  avail: %i  game: %i\n",
			realtics, clients[0].numtics, availabletics, counts);
	}

//
// run the count * ticdup tics
//
	while (counts--)
	{
		gx.Ticker();
		CON_Ticker();

		Sv_FixLocalAngles();

		// The low-level code gets to tick as well.
		I_NetTicker();
		if(netgame) Net_Ticker();

		gametic++;
		systics++;

		Net_Update();		// check for new console commands		
	}
}

int CCmdSetName(int argc, char **argv)
{
	playerinfo_packet_t info;

	if(argc != 2)
	{
		CON_Printf( "Usage: %s (name)\n", argv[0]);
		CON_Printf( "Use quotes to include spaces in the name.\n");
		return true;
	}
	CvarSetString(CvarGet("n_plrname"), argv[1]);
	if(!netgame) return true;

	// In netgames, a notification is sent to other players.
	memset(&info, 0, sizeof(info));
	info.console = consoleplayer;
	strncpy(info.name, argv[1], PLAYERNAMELEN-1);	

	// Serverplayers can update their name right away.
	if(server) strcpy(clients[0].name, info.name);

	D_SendPacket(DDSP_RELIABLE | (server? DDSP_ALL_PLAYERS : 0),
		pkt_player_info, &info, sizeof(info));			
	return true;
}