
//**************************************************************************
//**
//** OGL_DYN.C
//**
//** Dynamic lights.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include "dd_def.h"
#include "gl_def.h"
#include "m_bams.h"
#include "gl_rl.h"
#include "i_win32.h"
#include "settings.h"

// MACROS ------------------------------------------------------------------

#define NOBLOCKS				// Don't use the blockmap.
//#define CLIP_TO_SUBSECTOR		// Clip lights with subsector edges.

#define ZMOD_FLAT		0 //.05f	// How much difference?
#define ZMOD_WALL		.15f		
#define DISTFACTOR		1		// Really...

#define X_TO_DLBX(cx)			( ((cx) - dlBlockOrig.x) >> (FRACBITS+7) )
#define Y_TO_DLBY(cy)			( ((cy) - dlBlockOrig.y) >> (FRACBITS+7) )
#define DLB_ROOT_DLBXY(bx, by)	(dlBlockLinks + bx + by*dlBlockWidth)

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

fvertex_t *edgeClipper(int *numpoints, fvertex_t *points, int numclippers, fdivline_t *clippers);

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

#ifndef NOBLOCKS
boolean DL_IsSectorRegistered(int lumNum, sector_t *sec);
#endif

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int			useDynLights;
extern subsector_t	*currentssec;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int			useDynLights = true, dlBlend = 1;
float		dlFactor = 0.4f;

lumobj_t	*luminousList = 0;
int			numLuminous = 0, maxLuminous = 0;
int			dlMaxRad = 128; // Dynamic lights maximum radius.
int			maxDynLights = 0;

#ifndef NOBLOCKS
byte		*lightRegister = 0;
int			lrEntrySize = 0;	// The size of one entry in the Register.
int			lrAllocedMem = 0;	// Memory allocated for the Light Register.
#endif

// PRIVATE DATA DEFINITIONS ------------------------------------------------

lumobj_t	**dlBlockLinks = 0;
vertex_t	dlBlockOrig;
int			dlBlockWidth, dlBlockHeight;	// In 128 blocks.

// CODE --------------------------------------------------------------------

// *VERY* similar to SegFacingDir(). Well, this is actually the same
// function, only a more general version.
__inline static int SegFacingPoint(float v1[2], float v2[2], float p[2])
{
	return((v1[VY]-v2[VY])*(v1[VX]-p[VX])+(v2[VX]-v1[VX])*(v1[VY]-p[VY]) > 0);
}

/*static void projectVector(float a[2], float b[2], float *a_on_b)
{
	float	factor = (a[0]*b[0] + a[1]*b[1]) / (b[0]*b[0] + b[1]*b[1]);
	a_on_b[0] = factor * b[0];
	a_on_b[1] = factor * b[1];
}*/

void thingDLRadius(lumobj_t *lum)
{
	int thingRad = lum->thing->radius >> FRACBITS, 
		thingHeight = spriteheights[lum->patch];
	int	rad = thingRad>thingHeight? thingRad : thingHeight;
	int bigRad;
	
	// No point to have a too small dynamic light. It would just be
	// a waste of time.
	bigRad = rad*5;
	rad *= 2;
	if(rad < 48) rad = 48;
	if(rad > dlMaxRad) rad = dlMaxRad;
	if(rad > bigRad) rad = bigRad;
	lum->radius = rad;

	rad = (7*thingHeight + thingRad)/8; // < thingRad? thingHeight : thingRad;
	//rad *= rad;
	//ST_Message( "%i\n", rad);
	lum->flaresize = rad;
}

void thingDLColor(lumobj_t *lum, DGLubyte *outRGB, float light)
{
	int		i;

	if(light < 0) light = 0;
	if(light > 1) light = 1;
	light *= dlFactor;
	// Multiply the light color.
	for(i=0; i<3; i++)
	{
		outRGB[i] = (DGLubyte) (light * lum->rgb[i]);
		//outRGB[i] += (255 - outRGB[i])/4; // more brightness, less color...
	}
}

void DL_InitBlockMap()
{
	byte		*ptr;
	vertex_t	*vtx;
	vertex_t	min, max;
	int			i;

	memcpy(&min, vertexes, sizeof(min));
	memcpy(&max, vertexes, sizeof(max)); 
	for(i=1, ptr=vertexes+VTXSIZE; i<numvertexes; i++, ptr += VTXSIZE)
	{
		vtx = (vertex_t*) ptr;
		if(vtx->x < min.x) min.x = vtx->x;
		if(vtx->x > max.x) max.x = vtx->x;
		if(vtx->y < min.y) min.y = vtx->y;
		if(vtx->y > max.y) max.y = vtx->y;
	}
	memcpy(&dlBlockOrig, &min, sizeof(min));
	max.x -= min.x;
	max.y -= min.y;
	dlBlockWidth = (max.x >> FRACBITS >> 7) + 1;
	dlBlockHeight = (max.y >> FRACBITS >> 7) + 1;
	// Blocklinks is a table of lumobj_t pointers.
	dlBlockLinks = realloc(dlBlockLinks, sizeof(lumobj_t*) * dlBlockWidth * dlBlockHeight);
}

// Create dynamic light quads, if necessary.
void DL_ProcessWall(rendquad_t *quad, float v1[2], float v2[2])
{
	int			i, c;
	rendquad_t	dlq;
	float		pntLight[2];
	float		uvecWall[2];	// Unit vector.
	float		dist;			// Distance between light source and wall.

	// We can't handle masked walls. The alpha...
	if(quad->flags & RQF_MASKED) return;

	for(i=0; i<numLuminous; i++)
	{
		lumobj_t *lum = luminousList + i;

		// If the light is not in use, we skip it.
		if(!(lum->flags & LUMF_USED)) continue;

		// First we must check orientation. Backfaces aren't lighted, naturally.
		pntLight[VX] = FIX2FLT(lum->thing->x);
		pntLight[VY] = FIX2FLT(lum->thing->y);
		if(!SegFacingPoint(v1, v2, pntLight)) continue;

		// Make a copy of the original.
		memcpy(&dlq, quad, sizeof(dlq));

		// Copy the given end points.
		memcpy(dlq.v1, v1, sizeof(v1));
		memcpy(dlq.v2, v2, sizeof(v2));
		dlq.flags |= RQF_LIGHT; // This is a light texture.
		dlq.texw = dlq.texh = lum->radius*2;

		// The wall vector.
		for(c=0; c<2; c++) uvecWall[c] = (quad->v2[c] - quad->v1[c]) / quad->len;
		
		dist = ((v1[VY] - pntLight[VY])*(v2[VX] - v1[VX]) - 
			(v1[VX] - pntLight[VX])*(v2[VY] - v1[VY])) / quad->len;
		if(dist > lum->radius*DISTFACTOR) continue; // Too far away.

		// Now we can calculate the color of the light.
		thingDLColor(lum, dlq.color[0].rgb, 1.5f - 1.5f*dist/lum->radius);

		// Do a scalar projection for the offset.
		dlq.texoffx = ((v1[VY] - pntLight[VY])*(v1[VY] - v2[VY]) - 
			(v1[VX] - pntLight[VX])*(v2[VX] - v1[VX])) / quad->len - lum->radius;

		// There is no need to draw the *whole* wall always. Adjust the start
		// and end points so that only a relevant portion is included.
		if(dlq.texoffx > dlq.len) continue; // Doesn't fit on the wall.
		if(dlq.texoffx < -dlq.texw) continue; // Ditto, but in the other direction.
		if(dlq.texoffx > 0)
		{
			for(c=0; c<2; c++) dlq.v1[c] += dlq.texoffx * uvecWall[c];
			if(dlq.texoffx+dlq.texw <= dlq.len) // Fits completely?
			{
				for(c=0; c<2; c++) dlq.v2[c] = dlq.v1[c] + dlq.texw*uvecWall[c];
				dlq.len = dlq.texw;
			}
			else // Doesn't fit.
				dlq.len -= dlq.texoffx;
			dlq.texoffx = 0;
		}
		else // It goes off to the left.
		{
			if(dlq.texoffx+dlq.texw <= dlq.len) // Fits completely?
			{
				for(c=0; c<2; c++) dlq.v2[c] = dlq.v1[c] + (dlq.texw+dlq.texoffx)*uvecWall[c];
				dlq.len = dlq.texw + dlq.texoffx;
			}
		}
		// The vertical offset is easy to determine.
		//dlq.texoffy = FIX2FLT(lum->thing->z) + lum->top - lum->height/2 + lum->radius - dlq.top;
		dlq.texoffy = FIX2FLT(lum->thing->z) + lum->center + lum->radius - dlq.top;
		if(dlq.texoffy < -dlq.top+dlq.bottom) continue;
		if(dlq.texoffy > dlq.texh) continue;
		if(dlq.top+dlq.texoffy-dlq.texh >= dlq.bottom) // Fits completely?
			dlq.bottom = dlq.top+dlq.texoffy-dlq.texh;
		if(dlq.texoffy < 0)
		{
			dlq.top += dlq.texoffy;
			dlq.texoffy = 0;
		}
		// Now we can give the new quad to the rendering engine.
		RL_AddQuad(&dlq, 0);
	}
}

void DL_Clear()
{
	if(luminousList) Z_Free(luminousList);
	luminousList = 0;
	maxLuminous = numLuminous = 0;

	free(dlBlockLinks);
	dlBlockLinks = 0;
	dlBlockOrig.x = dlBlockOrig.y = 0;
	dlBlockWidth = dlBlockHeight = 0;

#ifndef NOBLOCKS
	free(lightRegister);
	lightRegister = 0;
	lrEntrySize = 0;
	lrAllocedMem = 0;
#endif
}

void DL_AddLuminous(mobj_t *thing)
{
	if(thing->frame & FF_FULLBRIGHT && !(thing->ddflags&DDMF_DONTDRAW))
	{
		spritedef_t *sprdef;
		spriteframe_t *sprframe;
		int lump, i;
		lumobj_t *lum;

		// Only allocate memory when it's needed.
		if(++numLuminous > maxLuminous)
		{
			lumobj_t *newlist = Z_Malloc(sizeof(lumobj_t) * (maxLuminous+=16), PU_STATIC, 0);
			// Copy the old data over to the new list.
			if(luminousList)
			{
				memcpy(newlist, luminousList, sizeof(lumobj_t) * (numLuminous-1));
				Z_Free(luminousList);
			}
			luminousList = newlist;
		}
		lum = luminousList + numLuminous-1;
		lum->thing = thing;
		// We need to know how tall the thing currently is.
		sprdef = &sprites[thing->sprite];
		sprframe = &sprdef->spriteframes[ thing->frame & FF_FRAMEMASK ];
		if(sprframe->rotate)
			lump = sprframe->lump[(R_PointToAngle(thing->x, thing->y) - thing->angle + (unsigned)(ANG45/2)*9) >> 29];
		else
			lump = sprframe->lump[0];
		
		// This'll ensure we have up-to-date information about the texture.
		// Notice that the texture is not bound now.
		OGL_PrepareSprite(lump);
		lum->patch = lump;
		//lum->top = FIX2FLT(spritetopoffset[lump] - lum->thing->floorclip);
		//lum->height = spriteheights[lump];
		lum->center = FIX2FLT(spritetopoffset[lump] - lum->thing->floorclip) - spriteheights[lump]/2;

		// Sets the dynlight and flare radii.
		thingDLRadius(lum);

		// Does the mobj use a light offset?
		if(thing->ddflags & DDMF_LIGHTOFFSET)
		{
			float mul;
			i = (thing->ddflags & DDMF_LIGHTOFFSET) >> DDMF_LIGHTOFFSETSHIFT;
			if(i & 0x8) i = 8-i; // Negative?
			lum->center += spriteheights[lump] * i/14.0f;
			// Reduce the size of the flare if it's moved off-center.
			mul = (12 - (i>0? i : -i)) / 12.0f;
			lum->radius *= mul;			
			lum->flaresize *= mul;
		}
		lum->flags = 0;

		// The color, too.
		OGL_GetSpriteColor(lum->patch, lum->rgb);

		// Approximate the distance, if necessary.
		if(maxDynLights || flareFadeMax) 
			lum->distance = M_AproxDistance(thing->x-viewx, thing->y-viewy);
	}
}

void DL_LinkLuminous()
{
	lumobj_t	**root, *lum = luminousList;
	int			i, bx, by;

	memset(dlBlockLinks, 0, sizeof(lumobj_t*) * dlBlockWidth * dlBlockHeight);
	for(i=0; i<numLuminous; i++, lum++)
	{
		// Link this lumobj to the dlBlockLinks, if it can be.
		lum->next = NULL;
		bx = X_TO_DLBX(lum->thing->x);
		by = Y_TO_DLBY(lum->thing->y);
		if(bx >= 0 && by >= 0 && bx < dlBlockWidth && by < dlBlockHeight)
		{
			root = DLB_ROOT_DLBXY(bx, by);
			lum->next = *root;
			*root = lum;
		}
	}
}

// We want to know which luminous objects are close enough the subsector.
void DL_MarkForSubsector(subsector_t *sub)
{
	int		i;
	float	minx, miny, maxx, maxy;	// Subsector bounding box.
	float	x, y;
#ifndef NOBLOCKS
	byte	*lrStart = lightRegister;
#endif

	// Adjust the bounding box to include the maximum radius of the
	// dynamic lights.
	// Now check all the luminous objects and mark those that
	// are close enough.
	for(i=0; i<numLuminous; i++
#ifdef NOBLOCKS
	)
#else
	, lightRegister += lrEntrySize)
#endif
	{
		lumobj_t *lum = luminousList + i;
		// By default the luminous object isn't used.
		lum->flags &= ~LUMF_USED;

		if(lum->thing->subsector == sub) lum->flags |= LUMF_RENDERED;
		if(!useDynLights) continue; // If only using halos, for instance.

#ifndef NOBLOCKS
		if(!DL_IsSectorRegistered(i, sub->sector)) continue;
#endif
		
		x = Q_FIX2FLT(lum->thing->x); 
		y = Q_FIX2FLT(lum->thing->y);
		minx = sub->bbox[0].x - lum->radius;
		miny = sub->bbox[0].y - lum->radius;
		maxx = sub->bbox[1].x + lum->radius; 
		maxy = sub->bbox[1].y + lum->radius; 
		// Is inside the bounding box?
		if(x > minx && y > miny && x < maxx && y < maxy)
		{
			lum->flags |= LUMF_USED/* | LUMF_RENDERED*/;
		}
	}
#ifndef NOBLOCKS
	lightRegister = lrStart;
#endif
}

void DL_ProcessFlats(rendquad_t *triangle, subsector_t *ssec)
{
	boolean		lightCeiling = true, lightFloor = true;
	int			i;
	sector_t	*sect = ssec->sector;
	float		fceil = FIX2FLT(sect->ceilingheight) + sect->skyfix, 
				ffloor = FIX2FLT(sect->floorheight),
				cdiff, fdiff;	// Differences with floor and ceiling.
#ifdef CLIP_TO_SUBSECTOR
	fdivline_t	clippers[MAX_CC_SIDES];
#endif

	// Check if lighting can be skipped.
	if(sect->floorpic == skyflatnum) lightFloor = false;
	if(sect->ceilingpic == skyflatnum) lightCeiling = false;

	// See if we can leave. This case is pretty rare, though (!).
	if(!lightCeiling && !lightFloor) return;

#ifdef CLIP_TO_SUBSECTOR	
	// First set up the clippers.
	for(i=0; i<ssec->numedgeverts; i++)
	{
		int next = i+1;
		if(next == ssec->numedgeverts) next = 0; // Wraparound.
		clippers[i].x = ssec->origedgeverts[i].x;
		clippers[i].y = ssec->origedgeverts[i].y;
		clippers[i].dx = ssec->origedgeverts[next].x - ssec->origedgeverts[i].x;
		clippers[i].dy = ssec->origedgeverts[next].y - ssec->origedgeverts[i].y;
	}
#endif
	
	triangle->flags = RQF_FLAT | RQF_LIGHT;

	for(i=0; i<numLuminous; i++)
	{
#ifdef CLIP_TO_SUBSECTOR
		fvertex_t	*lverts;	// The vertices for the light.
		int			numlverts = 4;
		float		x, y;
#endif
		lumobj_t	*lum = luminousList + i;
		float		z;
		boolean		applyCeiling, applyFloor;	// Is the light on the right side?
			
		if(!(lum->flags & LUMF_USED)) continue; // The light is not in use.

		z = FIX2FLT(lum->thing->z);

		// Determine on which side the light is for both the floor and ceiling.
		if(lightCeiling && z+lum->center-lum->radius < fceil)
			applyCeiling = true;
		else
			applyCeiling = false;

		if(lightFloor && z+lum->center+lum->radius > ffloor)
			applyFloor = true;
		else
			applyFloor = false;

		// Center the Z.
		z += lum->center;

		// Check that the height differences are tolerable.
		if(applyCeiling)
		{
			cdiff = fabs(fceil - z);
			if(cdiff > lum->radius) applyCeiling = false;
		}
		if(applyFloor)
		{
			fdiff = fabs(ffloor - z);
			if(fdiff > lum->radius) applyFloor = false;
		}

		// Do we need to light with this light at all?
		if(!applyCeiling && !applyFloor) continue;

		/* 
		// Check that we're on the right side.
		if((floor && z+lum->top < triangle->top) ||
			(!floor && z+lum->top-lum->height > triangle->top)) continue; */
		
/*		// Check that the height difference isn't too big.
		z += lum->top - lum->height/2;	// Center Z.
		if((hdiff=fabs(triangle->top - z)) > lum->radius) continue;*/

#ifdef CLIP_TO_SUBSECTOR
		// Clip using the subsector edges.
		lverts = malloc(sizeof(fvertex_t)*numlverts);
		
		// Get the coordinates.
		x = FIX2FLT(lum->thing->x);
		y = FIX2FLT(lum->thing->y);
		// These are the initial coordinates for the light.
		lverts[0].x = x - lum->radius;
		lverts[0].y = y - lum->radius;
		lverts[1].x = x + lum->radius;
		lverts[1].y = y - lum->radius;
		lverts[2].x = x + lum->radius;
		lverts[2].y = y + lum->radius;
		lverts[3].x = x - lum->radius;
		lverts[3].y = y + lum->radius;

		// Clip the light to fit the subsector.
		lverts = edgeClipper(&numlverts, lverts, ssec->numedgeverts, clippers);
		if(!numlverts) continue;	// Nothing left to render.
#endif

		if(applyCeiling)
		{
			triangle->top = fceil - ZMOD_FLAT;
			thingDLColor(lum, triangle->color[0].rgb, 1.5f - 1.5f*cdiff/lum->radius);
			// We can add the light quads.
#ifdef CLIP_TO_SUBSECTOR
			RL_AddFlatQuads(triangle, (int)lum, numlverts, lverts, 0);
#else
			RL_AddFlatQuads(triangle, (int)lum, ssec->numedgeverts, ssec->origedgeverts, 1);
#endif
		}
		if(applyFloor)
		{
			triangle->top = ffloor + ZMOD_FLAT;
			thingDLColor(lum, triangle->color[0].rgb, 1.5f - 1.5f*fdiff/lum->radius);
			// Add the light quads.
#ifdef CLIP_TO_SUBSECTOR
			RL_AddFlatQuads(triangle, (int)lum, numlverts, lverts, 1);
#else
			RL_AddFlatQuads(triangle, (int)lum, ssec->numedgeverts, ssec->origedgeverts, 0);
#endif
		}

#ifdef CLIP_TO_SUBSECTOR
		// Now we can let the light vertices go.
		free(lverts);
#endif
	}
}

#ifndef NOBLOCKS
__inline boolean DL_IsSectorRegistered(int lumNum, sector_t *sec)
{
	register unsigned short secnum = sec-sectors;//) / sizeof(sector_t);
	register unsigned short byteoff = secnum/8;		// Which byte?

	//return( lightRegister[lumNum*lrEntrySize + byteoff] & (1<<(secnum - byteoff*8)) );
	return( lightRegister[byteoff] & (1<<(secnum - byteoff*8)) );
}

void DL_RegisterSector(sector_t *sec)
{
	int secnum = sec-sectors;//) / sizeof(sector_t);
	int byteoff = secnum/8;				// Which byte?

	lightRegister[byteoff] |= 1 << (secnum - byteoff*8); // always: secnum >= byteoff*8
}

boolean DL_LineOperator(line_t *line)
{
	if(line->frontsector) DL_RegisterSector(line->frontsector);
	if(line->backsector) DL_RegisterSector(line->backsector);
	return true;	// Carry on...
}
#endif

int lumobjSorter(const void *e1, const void *e2)
{
	lumobj_t *lum1 = (lumobj_t*) e1, *lum2 = (lumobj_t*) e2;

	if(lum1->distance > lum2->distance) return 1;
	if(lum1->distance < lum2->distance) return -1;
	return 0;
}

void DL_InitForNewFrame()
{
	sector_t	*seciter;//, *lastsec, *startsec;
	int			i, done = false;
#ifndef NOBLOCKS
	int			bx1, by1, bx2, by2, s, t;
	byte		*lrBeginning; // Pointer to the beginning of the light register.
	fixed_t		fixrad;
#endif

	numLuminous = 0;	// Clear the luminous object list.
			
	// Add all the luminous objects to the list.
	// First add the lights in the player's sector.
	//i = (players[consoleplayer].mo->subsector->sector - sectors) / sizeof(sector_t);
	//lastsec = sectors + numsectors-1;
//	seciter = startsec = players[displayplayer].mo->subsector->sector;
	//while(!done)

	// Clearly there are smarter ways to do this, but...
	for(i=0; i<numsectors; i++)
	{
		mobj_t *iter;
		//seciter = sectors + i;
		seciter = SECTOR_PTR(i);
		for(iter=seciter->thinglist; iter; iter=iter->snext)
		{
			DL_AddLuminous(iter);
			/*if(maxDynLights && ++dlCounter >= maxDynLights) 
			{
				done = true;
				break;
			}*/
		}
		// Advance to the next sector.
		/*if(seciter == lastsec) seciter = sectors; else seciter++;
		// Are we done?
		if(seciter == startsec) break; */
	}
	if(maxDynLights)
	{
		// Sort the lumobjs based on their distance.
		qsort(luminousList, numLuminous, sizeof(lumobj_t), lumobjSorter);
		// Force the maximum.
		if(numLuminous > maxDynLights) numLuminous = maxDynLights;
	}
	// Link the luminous objects into the blockmap.
	DL_LinkLuminous();

#ifndef NOBLOCKS
	// Now that we know how many luminous objects we have, we must
	// update the Light Register. The first thing is to see that
	// we have enough memory for the current number of luminous
	// objects.
	if(numLuminous*lrEntrySize > lrAllocedMem)
	{
		lrAllocedMem = numLuminous*lrEntrySize;
		lightRegister = realloc(lightRegister, lrAllocedMem);
	}

	// Save the light register beginning.
	lrBeginning = lightRegister;

	// Update the Register.
	for(i=0; i<numLuminous; i++, lightRegister += lrEntrySize)
	{
		lumobj_t *lum = luminousList + i;

		// Clear all the sector bits for this light.
		memset(lightRegister, 0, lrEntrySize);

		// Register the current sector by default.
		DL_RegisterSector(lum->thing->subsector->sector);

		// Now we must determine the blockmap coordinates for the
		// bounding box of this light.
		fixrad = lum->radius << FRACBITS;
		bx1 = (lum->thing->x - bmaporgx - fixrad) >> MAPBLOCKSHIFT;
		by1 = (lum->thing->y - bmaporgy - fixrad) >> MAPBLOCKSHIFT;
		bx2 = (lum->thing->x - bmaporgx + fixrad) >> MAPBLOCKSHIFT;
		by2 = (lum->thing->y - bmaporgy + fixrad) >> MAPBLOCKSHIFT;

		// Keep things up to speed.
		validcount++; 
		// Walk through the blocks.
		for(t=by1; t<=by2; t++)
			for(s=bx1; s<=bx2; s++)
				P_BlockLinesIterator(s, t, DL_LineOperator);								
	}
	
	// Restore the light register beginning.
	lightRegister = lrBeginning;
#endif
}

// Usually called in P_SetupLevel.
void DL_SetRegisterEntrySize()
{
#ifndef NOBLOCKS
	lrEntrySize = numsectors/8;	// Eight bits per byte (really??!).
	// Do we need an extra byte?
	if(numsectors % 8) lrEntrySize++;	// One more for the odd bits.
#endif
}

// Calls func for all luminous objects within the specified range from (x,y).
boolean DL_RadiusIterator(fixed_t x, fixed_t y, fixed_t radius, 
						  boolean (*func)(lumobj_t*,fixed_t))
{
	int		s, t, bx1, by1, bx2, by2;
	fixed_t	dist;
	lumobj_t *iter;
	
	bx1 = X_TO_DLBX(x-radius);
	bx2 = X_TO_DLBX(x+radius);
	by1 = Y_TO_DLBY(y-radius);
	by2 = Y_TO_DLBY(y+radius);
	// Walk through the blocks.
	for(t=by1; t<=by2; t++)
		for(s=bx1; s<=bx2; s++)
		{
			// We can't go outside the blockmap.
			if(s < 0 || t < 0 || s >= dlBlockWidth || t >= dlBlockHeight) continue;
			for(iter = *DLB_ROOT_DLBXY(s, t); iter; iter = iter->next)
			{
				dist = M_AproxDistance(iter->thing->x - x, iter->thing->y - y);
				if(dist <= radius)
				{
					if(!func(iter, dist)) return false;						
				}
			}
		}
	return true;
}