#if defined(USEA3D)

//**************************************************************************
//**
//** I_SFX.C
//**
//** Sound playing using A3D 3.0. 
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN

#include <stdlib.h>
#include <math.h>
#include <windows.h>
#include <mmsystem.h>
#include <objbase.h>
#include <initguid.h>
#include <ia3dapi.h>
#include <ia3dutil.h>

#include "h2def.h"
#include "p_local.h"
//#include "soundst.h"
#include "i_timer.h"
#include "i_mus.h"

// MACROS ------------------------------------------------------------------

#define SNDBUFPTR(handle)	(buffers+handle-1)
#define SNDBUFHANDLE(sbptr)	((sbptr-buffers)+1)

#define SBF_UNOWNED		0x1	// The buffer is not owned by anyone.

// TYPES -------------------------------------------------------------------

typedef struct
{
	IA3dSource	*source;	// Pointer to the source.
	int			flags;
} sndbuffer_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern HWND			hWndMain;	// The game window.
extern sfxinfo_t	S_sfx[];

// PUBLIC DATA DEFINITIONS -------------------------------------------------

int	snd_MusicDevice=1, snd_SfxDevice=0;      
int snd_3DAudio=0; // By default, the 3D audio is off.

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static boolean		initOk = false;
static IA3d4		*a3d = NULL;
static IA3dListener *a3dListener = NULL;

int					numBuffers = 0;
static sndbuffer_t	*buffers = 0;

// CODE --------------------------------------------------------------------

// Returns true if the buffer is playing (duh).
int audio_IsBufferPlaying(sndbuffer_t *sndbuf)
{
	DWORD	status;

	IA3dSource_GetStatus(sndbuf->source, &status);
	// Status == zero if buffer is *not* playing.
	return(status != 0);
}	

// Returns a free buffer.
sndbuffer_t *audio_GetInertBuffer()
{
	int		i;
	
	// First check if there are any free buffers in the list.
	for(i=0; i<numBuffers; i++)
	{
		if(buffers[i].source == NULL)
			return buffers + i;
		if(buffers[i].flags & SBF_UNOWNED && !audio_IsBufferPlaying(buffers+i))
			return buffers + i;
	}
	
	// Ah well. We need to allocate a new one.
	numBuffers++;
	buffers = realloc(buffers, sizeof(*buffers) * numBuffers);
	// Clear it.
	memset(buffers + numBuffers-1, 0, sizeof(*buffers));
	// Return a pointer to it.
	return buffers + numBuffers-1;
}

void audio_DestroyBuffers()
{
	int		i;

	for(i=0; i<numBuffers; i++)
		if(buffers[i].source) IA3dSource_Release(buffers[i].source);

	free(buffers);
	buffers = NULL;
	numBuffers = 0;
}

// Volume is linear, from 0 to 1.
void audio_SetBufferVolume(sndbuffer_t *sndbuf, float volume)
{
	float gain = 0;
	if(volume > 0) gain = pow(.5, (100*(1-volume))/6);
	IA3dSource_SetGain(sndbuf->source, gain);
}

void audio_SetPitch(sndbuffer_t *sndbuf, float pitch)
{
	IA3dSource_SetPitch(sndbuf->source, pitch);
}

// Set the position and velocity of the source.
void audio_UpdateBuffer(sndbuffer_t *sndbuf, channel_t *channel)
{
	float	pos[3] = {0,0,0}, vel[3] = {0,0,0};
	mobj_t	*mo = channel->mo;
	float	ang;

	if(mo)
	{
		// All mobjs have a position, naturally.
		pos[0] = FIX2FLT(mo->x);
		pos[1] = FIX2FLT(mo->z + mo->height/2);
		pos[2] = -FIX2FLT(mo->y);

		// Add a displacement caused by turning.
		ang = mo->angle / (float) ANGLE_MAX * 2*PI;
		pos[0] += cos(ang) * (mo->radius>>FRACBITS);
		pos[2] -= sin(ang) * (mo->radius>>FRACBITS);

		// Missiles have usually a good velocity, too.
		if(mo->flags & MF_MISSILE)
		{
			vel[0] = 35*FIX2FLT(mo->momx);
			vel[1] = 35*FIX2FLT(mo->momz);
			vel[2] = -35*FIX2FLT(mo->momy);

			// Reset values.
		/*IA3dSource_SetPosition3f(sndbuf->source, 0, 0, 0);
		IA3dSource_SetVelocity3f(sndbuf->source, 0, 0, 0);*/

		}
	}
	
	// Convert everything to a right-handed coordinate system.
	// Sounds eminate from the middle of mobjs.
	/*IA3dSource_SetPosition3f(sndbuf->source, 
		FIX2FLT(channel->mo->x), FIX2FLT(channel->mo->z+channel->mo->height/2), 
		-FIX2FLT(channel->mo->y));
	IA3dSource_SetVelocity3f(sndbuf->source,
		FIX2FLT(channel->mo->momx)*35, FIX2FLT(channel->mo->momz)*35, -FIX2FLT(channel->mo->momy)*35);*/

	IA3dSource_SetPosition3fv(sndbuf->source, pos);
	IA3dSource_SetVelocity3fv(sndbuf->source, vel);
}

//int I_StartSound(int id, void *data, int vol, int sep, int pitch, int priority)
int I_StartSound(channel_t *channel)
{
	void			*data = S_sfx[channel->sound_id].snd_ptr;
	unsigned short	*ptr = data;		// Easier access to the header.
	int				length = *(ptr+2), bits = 8, channels = 1, freq = *(ptr+1);
	int				hr;
	sndbuffer_t		*sndbuf;
	WAVEFORMATEX	form;
	void			*writePtr1=NULL, *writePtr2=NULL;
	DWORD			writeBytes1, writeBytes2, dataBytes = length * bits/8 * channels;

	// Can we play sounds?
	if(!initOk || data == NULL) return 0;	// Sorry...

	// Get a buffer that's doing nothing.
	sndbuf = audio_GetInertBuffer();

	// Create a new source.
	if(sndbuf->source == NULL)
	{
		// No old source, create a new one.
		if(FAILED(hr = IA3d4_NewSource(a3d, A3DSOURCE_TYPEDEFAULT, &sndbuf->source)))
		{
			I_Error("I_StartSound: couldn't create a new source (result = %d).\n", hr);
			return 0;
		}
	}
	else
	{
		// Free the old source's wave data.
		IA3dSource_FreeWaveData(sndbuf->source);
	}

	// We need to load the wave data into the source.
	// Prepare the format description.
	memset(&form, 0, sizeof(form));
	form.wFormatTag = WAVE_FORMAT_PCM;
	form.nChannels = channels;
	form.nSamplesPerSec = freq;
	form.nBlockAlign = channels * bits/8;
	form.nAvgBytesPerSec = form.nSamplesPerSec * form.nBlockAlign;
	form.wBitsPerSample = bits;

	if(FAILED(hr = IA3dSource_SetWaveFormat(sndbuf->source, (void*)&form)))
	{
		I_Error("I_StartSound: couldn't set wave format (result = %d).\n", hr);
		return 0;
	}

	// Allocate the resources.
	if(FAILED(hr = IA3dSource_AllocateWaveData(sndbuf->source, dataBytes)))
	{
		I_Error("I_StartSound: couldn't allocate wave data (result = %d).\n", hr);
		return 0;
	}

	// Lock the buffer.
	if(FAILED(hr = IA3dSource_Lock(sndbuf->source, 0, dataBytes, 
		&writePtr1, &writeBytes1, &writePtr2, &writeBytes2,	A3D_ENTIREBUFFER)))
	{
		I_Error("I_StartSound: couldn't lock source (result = %d).\n", hr);
		return 0;
	}

	// Copy the data over.
	memcpy(writePtr1, data, writeBytes1);
	if(writePtr2) memcpy(writePtr2, (char*)data + writeBytes1, writeBytes2);

	// Unlock the buffer.
	if(FAILED(hr = IA3dSource_Unlock(sndbuf->source, writePtr1, writeBytes1, 
		writePtr2, writeBytes2)))
	{
		I_Error("I_StartSound: couldn't unlock source (result = %d).\n", hr);
		return 0;
	}

	// Set the minimum and maximum distances.
	IA3dSource_SetMinMaxDistance(sndbuf->source, 256, MAX_SND_DIST, A3D_MUTE);
	// Priority.
	IA3dSource_SetPriority(sndbuf->source, channel->priority/512.0f);

	// Set the volume.
	audio_SetBufferVolume(sndbuf, channel->volume/127.0f);

	// The pitch.
	audio_SetPitch(sndbuf, channel->pitch/127.0f);

	// Set the source position and velocity.
	audio_UpdateBuffer(sndbuf, channel);

	// Start playing the buffer.
	if(FAILED(hr = IA3dSource_Play(sndbuf->source, A3D_SINGLE)))
	{
		I_Error("I_StartSound: couldn't start source (result = %d).\n", hr);
		return 0;
	}

	// Return the handle to the sound source.
	channel->handle = SNDBUFHANDLE(sndbuf);
	return channel->handle;
}

void I_StopSound(int handle)
{
	sndbuffer_t *sndbuf = SNDBUFPTR(handle);
	IA3dSource_Stop(sndbuf->source);
	sndbuf->flags |= SBF_UNOWNED;
}

int I_SoundIsPlaying(int handle)
{
	if(!handle) return 0;
	return audio_IsBufferPlaying(SNDBUFPTR(handle));
}

//void I_UpdateSoundParams(int handle, int vol, int sep, int pitch)
void I_UpdateSoundParams(channel_t *channel)
{
	audio_UpdateBuffer(SNDBUFPTR(channel->handle), channel);
}

void I_SetChannels(int channels)
{

}

// This'll mark the handle for reuse, the user has abandoned it.
void I_FreeHandle(int handle)
{
	if(!handle) return;
	SNDBUFPTR(handle)->flags |= SBF_UNOWNED;
}

void I_UpdateListener(player_t *plr)
{
	float yaw, pitch, xyz[3], vel[3];

	if(plr->mo == NULL)
	{
		// The player has no mo?
		xyz[0] = xyz[1] = xyz[2] = 0;
		vel[0] = vel[1] = vel[2] = 0;
		yaw = pitch = 0;
	}
	else
	{
		xyz[0] = FIX2FLT(plr->mo->x);
		xyz[1] = FIX2FLT(plr->mo->z+plr->mo->height)-10;
		xyz[2] = -FIX2FLT(plr->mo->y);
		
		vel[0] = 35*FIX2FLT(plr->mo->momx);
		vel[1] = 35*FIX2FLT(plr->mo->momz);
		vel[2] = -35*FIX2FLT(plr->mo->momy);

		yaw = plr->mo->angle / (float)ANGLE_MAX * 360 - 90;
		pitch = LOOKDIR2DEG(plr->lookdir);
	}
	// Set them!
	IA3dListener_SetPosition3fv(a3dListener, xyz);
	IA3dListener_SetVelocity3fv(a3dListener, vel);
	IA3dListener_SetOrientationAngles3f(a3dListener, yaw, pitch, 0);
}

void I_BeginSoundFrame()
{
	IA3d4_Clear(a3d);
}

void I_EndSoundFrame()
{
	IA3d4_Flush(a3d);
}


/*
**  Startup/Exit Stuff
*/

// Attempts to start up A3D 2.0 using the specified hwnd.
int audio_Startup(HWND hWndMain)
{
	int	hr;

	// First, get us the IA3d4 object.
	if(FAILED(hr = A3dCreate(NULL, (void**)&a3d, NULL, 0)))
	{
		ST_Message( "audioStartup: Couldn't create a3d (result = %d).\n", hr);
		return 0;
	}
	
	// Set the cooperative level.
	IA3d4_SetCooperativeLevel(a3d, hWndMain, A3D_CL_NORMAL);

	// Give me the listener!
	if(FAILED(hr = IA3d4_QueryInterface(a3d, &IID_IA3dListener, (void**)&a3dListener)))
	{
		ST_Message( "audioStartup: Couldn't get a listener (result = %d).\n", hr);
		return 0;
	}

	IA3d4_SetUnitsPerMeter(a3d, 64);
	IA3d4_SetDopplerScale(a3d, 2);

	// We're successful.
	initOk = true;	// No sound can be played without this flag.
	return 1;				
}

void audio_Exit()
{
	audio_DestroyBuffers();
	if(a3d) IA3d4_Shutdown(a3d); // This'll get rid of everything else.
	a3d = NULL;
	a3dListener = NULL;
	A3dUninitialize();
	initOk = false;
}

void I_StartupSound (void)
{
	snd_SfxDevice = 1;
	if(snd_MusicDevice > 2) snd_MusicDevice = 1; // Old settings revert to midi.

	// Check command-line parameters -- overrides config file.
	if (M_CheckParm("-nosound")) snd_MusicDevice = snd_SfxDevice = 0;
	if (M_CheckParm("-nosfx")) snd_SfxDevice = 0;
	if (M_CheckParm("-nomusic")) snd_MusicDevice = 0;

/*	if (debugmode)
		ST_Message("I_StartupSound: Hope you hear a pop.\n");*/

	I_StartupTimer();

	if(snd_SfxDevice)
	{
		if(!audio_Startup(hWndMain))
			ST_Message("I_StartupSound: Error initializing sound.\n");
	}
	else
		ST_Message("I_StartupSound: Sound is disabled.\n");

	// Initialize the MUS player.
	if(MUS_Init())		// True when everything OK.
		ST_Message("I_StartupSound: Error initializing music.\n");

	// If the music device isn't MIDI, the Mus player should be disabled.
	if(snd_MusicDevice != 1) MUS_Disable(true);
}

// shuts down all sound stuff

void I_ShutdownSound (void)
{
	MUS_Shutdown();
	audio_Exit();
	I_ShutdownTimer();
}

#endif