
//**************************************************************************
//**
//** OGL_TEX.C
//**
//** A bit of a mess, if you ask me...
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include <stdlib.h>
#include "gl_s3tc.h"
#include "dd_def.h"
#include "i_win32.h"
#include "ogl_def.h"

// MACROS ------------------------------------------------------------------

#define NUM_FLARES	3

#define RGB15(r,g,b) ((r)+((g)<<5)+((b)<<10))

// TYPES -------------------------------------------------------------------

// A translated sprite.
typedef struct
{
	int				patch;
	GLuint			tex;
	unsigned char	*table;
} transspr_t;

// Sprite color (for dynamic lights, if the sprite is fullbright).
typedef struct
{
	unsigned char	rgb[3];
} rgbcol_t;

// Sky texture topline colors.
typedef struct
{
	int				texidx;
	unsigned char	rgb[3];
} skycol_t;

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern int maxTexSize;		// Maximum supported texture size.
extern int ratioLimit;
extern int test3dfx;
extern boolean palettedTextureExtAvailable;
extern boolean s3tcAvailable;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

boolean paletted = false;	// Use GL_EXT_paletted_texture (cvar)
//boolean usePalTex = false;	// GL_EXT_paletted_texture is inited and in use
boolean r_s3tc = false;		// Use GL_S3_s3tc (cvar)
boolean useS3TC = false;	// GL_S3_s3tc is inited and in use

// Convert a 15-bit RGB value to a playpal index. 
byte pal15to8[32768];

int mipmapping=3, linearRaw=1;

int pallump;
int mynumlumps;

// Properties of the current texture.
float	texw=1, texh=1;
int		texmask=0;	
GLuint	curtex = 0;

texsize_t *lumptexsizes;	// Sizes for all the lumps. 
unsigned short *spriteheights;

skycol_t *skytop_colors = NULL;
int num_skytop_colors = 0;

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// Texture names for all lumps. Although the list will contain names for 
// ALL lumps, only the graphics entries that aren't flats, wall textures or 
// sprites are used.
static DGLuint *lumptexnames, *lumptexnames2;	// Support for two parts.
static int *rawlumps, numrawlumps;	// Raw screen lumps (just lump numbers).
	
static DGLuint *flattexnames, *texnames, *spritenames;
static rgbcol_t *spritecolors;
static char *texmasked;		// 1 if the texture is masked.

// The translated sprites.
static transspr_t *transsprites;
static int numtranssprites;

static DGLuint dltexname;	// Name of the dynamic light texture.
static DGLuint flaretexnames[NUM_FLARES];

static int glmode[6] = // Indexed by 'mipmapping'.
{ 
	DGL_NEAREST,
	DGL_LINEAR,
	DGL_NEAREST_MIPMAP_NEAREST,
	DGL_LINEAR_MIPMAP_NEAREST,
	DGL_NEAREST_MIPMAP_LINEAR,
	DGL_LINEAR_MIPMAP_LINEAR
};


// CODE --------------------------------------------------------------------

int FindNextPower2(int num)
{
	int cumul;
	for(cumul=1; num > cumul; cumul <<= 1);
	return cumul;
}

float NextPower2Ratio(int num)
{
	return num/(float)FindNextPower2(num);
}

// Copies a rectangular region of the source buffer to the destination
// buffer. Doesn't perform clipping, so be careful.
// Yeah, 13 parameters.
static void pixBlt(byte *src, int srcWidth, int srcHeight, byte *dest, 
				   int destWidth, int destHeight,
				   int alpha, int srcRegX, int srcRegY, int destRegX, int destRegY,
				   int regWidth, int regHeight)
{
	int	y;	// Y in the copy region.
	int srcNumPels = srcWidth*srcHeight;
	int destNumPels = destWidth*destHeight;

	for(y=0; y<regHeight; y++) // Copy line by line.
	{
		// The color index data.
		memcpy(dest + destRegX + (y+destRegY)*destWidth,
			src + srcRegX + (y+srcRegY)*srcWidth, 
			regWidth);

		if(alpha)
		{
			// Alpha channel data.
			memcpy(dest + destNumPels + destRegX + (y+destRegY)*destWidth,
				src + srcNumPels + srcRegX + (y+srcRegY)*srcWidth,	
				regWidth);
		}
	}
}

// Prepare the pal15to8 table.
static void LookupPal15to8(byte *palette)
{
	int				r, g, b, i;
	byte			palRGB[3];
	unsigned int	diff, smallestDiff, closestIndex;
		
	// This is a time-consuming operation.
	for(r=0; r<32; r++)
		for(g=0; g<32; g++)
			for(b=0; b<32; b++)
			{
				// We must find the color index that most closely
				// resembles this RGB combination.
				smallestDiff = -1;
				for(i=0; i<256; i++)
				{
					memcpy(palRGB, palette + 3*i, 3);
					diff = (palRGB[0]-(r<<3))*(palRGB[0]-(r<<3)) 
						+ (palRGB[1]-(g<<3))*(palRGB[1]-(g<<3)) 
						+ (palRGB[2]-(b<<3))*(palRGB[2]-(b<<3));
					if(diff < smallestDiff) 
					{
						smallestDiff = diff;
						closestIndex = i;
					}
				}
				pal15to8[RGB15(r,g,b)] = closestIndex;
			}

	if(M_CheckParm("-dump_pal15to8"))
	{
		FILE *file = fopen("Pal15to8.lmp", "wb");
		fwrite(pal15to8, sizeof(pal15to8), 1, file);								
		fclose(file);
	}
}

static void LoadPalette()
{
	byte	*playpal = W_CacheLumpNum(pallump=W_GetNumForName("PLAYPAL"), PU_CACHE);
	byte	*paldata = _alloca(256 * 3);
	int		i, c;

	memset(paldata, 0, 256*3);
	// Prepare the color table.
	for(i=0; i<256; i++)
	{
		for(c=0; c<3; c++) paldata[i*3 + c] = gammatable[usegamma][playpal[i*3 + c]];
	}
/*	glColorTableEXT(GL_SHARED_TEXTURE_PALETTE_EXT, GL_RGB, 256, GL_RGB, GL_UNSIGNED_BYTE, 
		paldata);*/
	gl.Palette(DGL_RGB, paldata);
}

// Initializes the paletted texture extension.
// Returns true if successful.
int OGL_InitPalettedTexture()
{
	int		i;
	// This is the default setting.
//	usePalTex = false;
	
	if(M_CheckParm("-dump_pal15to8"))
		LookupPal15to8(W_CacheLumpName("PLAYPAL", PU_CACHE));

	// Should the extension be used?
	if(!paletted && !M_CheckParm("-paltex")) 
		return true;

	gl.Enable(DGL_PALETTED_TEXTURES);
	gl.GetIntegerv(DGL_PALETTED_TEXTURES, &i);
	if(i == DGL_FALSE)
	{
		ST_Message("Paletted textures init failed!\n");
		return false;
	}

/*	if(!palettedTextureExtAvailable) 
	{
		ST_Message("GL_EXT_shared_texture_palette is not available.\n");
		return false;
	}

	if((glColorTableEXT = (PFNGLCOLORTABLEEXTPROC) wglGetProcAddress("glColorTableEXT")) == NULL 
		|| (glGetColorTableParameterivEXT = (PFNGLGETCOLORTABLEPARAMETERIVEXTPROC)
		wglGetProcAddress("glGetColorTableParameterivEXT")) == NULL 
		|| (glGetColorTableEXT = (PFNGLGETCOLORTABLEEXTPROC)
		wglGetProcAddress("glGetColorTableEXT")) == NULL)
	{
		ST_Message("getProcAddress failed (EXT_paletted_texture).\n");
		return false;
	}
	else
	{
		int		i;
		byte	*playpal = W_CacheLumpNum(pallump=W_GetNumForName("PLAYPAL"), PU_CACHE);

		glEnable(GL_SHARED_TEXTURE_PALETTE_EXT);
		LoadPalette();

		// Load the pal15to8 table from the lump PAL15TO8.
		if((i=W_CheckNumForName("PAL15TO8")) == -1)
		{
			LookupPal15to8(playpal);
		}
		else memcpy(pal15to8, W_CacheLumpNum(i, PU_CACHE), sizeof(pal15to8));
	}
	// LoadTexture will use the appropriate data formats.
	usePalTex = true;*/
	return true;
}

/*int OGL_InitS3TC()
{
	// This is the default setting.
	useS3TC = false;
	
	// Should the extension be used?
	if(!r_s3tc && !M_CheckParm("-s3tc")) 
		return true;

	if(!s3tcAvailable) 
	{
		ST_Message("GL_S3_s3tc is not available.\n");
		return false;
	}

	// LoadTexture will use the appropriate data formats.
	useS3TC = true;
	return true;
}*/

void OGL_TexInit()
{
	int		i;

	// Allocate memory for the flat texture numbers.
	flattexnames = Z_Malloc(sizeof(GLuint)*numflats, PU_OPENGL, 0);	
	memset(flattexnames, 0, sizeof(GLuint)*numflats);

	texnames = Z_Malloc(sizeof(GLuint)*numtextures, PU_OPENGL, 0);
	memset(texnames, 0, sizeof(GLuint)*numtextures);

	texmasked = Z_Malloc(numtextures, PU_OPENGL, 0);
	memset(texmasked, 0, numtextures);

	// Sprites.
	spritenames = Z_Malloc(sizeof(GLuint)*numspritelumps, PU_OPENGL, 0);
	memset(spritenames, 0, sizeof(GLuint)*numspritelumps);	
	spriteheights = Z_Malloc(sizeof(short)*numspritelumps, PU_OPENGL, 0);
	memset(spriteheights, 0, sizeof(short)*numspritelumps);
	spritecolors = Z_Malloc(sizeof(rgbcol_t)*numspritelumps, PU_OPENGL, 0);
	memset(spritecolors, 0, sizeof(rgbcol_t)*numspritelumps);
	
	transsprites = 0;
	numtranssprites = 0;

	// Standard lump textures (raw images and other gfx).
	// Numlumps changes when data is loaded at runtime.
	mynumlumps = numlumps;
	// First parts.
	lumptexnames = Z_Malloc(sizeof(GLuint)*numlumps, PU_OPENGL, 0);
	memset(lumptexnames, 0, sizeof(GLuint)*numlumps);
	// Second parts.
	lumptexnames2 = Z_Malloc(sizeof(GLuint)*numlumps, PU_OPENGL, 0);
	memset(lumptexnames2, 0, sizeof(GLuint)*numlumps);
	// Size data.
	lumptexsizes = Z_Malloc(sizeof(texsize_t)*numlumps, PU_OPENGL, 0);
	memset(lumptexsizes, 0, sizeof(texsize_t)*numlumps);

	// Raw screen lump book keeping.
	rawlumps = 0;
	numrawlumps = 0;

	// The dynamic light map.
	dltexname = 0;
	memset(flaretexnames, 0, sizeof(flaretexnames));

	// The palette lump, for color information (really??!!?!?!).
	pallump = W_GetNumForName("PLAYPAL");

	OGL_InitPalettedTexture();
	LoadPalette();
	// Load the pal15to8 table from the lump PAL15TO8.
	if((i=W_CheckNumForName("PAL15TO8")) == -1)
	{
		LookupPal15to8(W_CacheLumpNum(pallump, PU_CACHE));
	}
	else
	{
		memcpy(pal15to8, W_CacheLumpNum(i, PU_CACHE), sizeof(pal15to8));
	}

	dltexname = OGL_PrepareLightTexture();
	gl.SetInteger(DGL_LIGHT_TEXTURE, dltexname);

//	OGL_InitS3TC();
}

void OGL_TexDestroy()
{
	OGL_TexReset();
	OGL_ResetLumpTexData();
	Z_FreeTags(PU_OPENGL, PU_OPENGL);
}

void OGL_TexReset(void)
{
	int					i;

	gl.DeleteTextures(numflats, flattexnames);
	memset(flattexnames, 0, sizeof(GLuint)*numflats);

	gl.DeleteTextures(numtextures, texnames);
	memset(texnames, 0, sizeof(GLuint)*numtextures);
	
	memset(texmasked, 0, numtextures);

	gl.DeleteTextures(numspritelumps, spritenames);
	memset(spritenames, 0, sizeof(GLuint)*numspritelumps);
	memset(spriteheights, 0, sizeof(short)*numspritelumps);
	memset(spritecolors, 0, sizeof(rgbcol_t)*numspritelumps);

	// Delete the translated sprite textures.
	for(i=0; i<numtranssprites; i++)
	{
		gl.DeleteTextures(1, &transsprites[i].tex);
	}
	free(transsprites);
	transsprites = 0;
	numtranssprites = 0;

	free(skytop_colors);
	skytop_colors = 0;
	num_skytop_colors = 0;

	gl.DeleteTextures(1, &dltexname);
	dltexname = 0;
	gl.SetInteger(DGL_LIGHT_TEXTURE, dltexname = OGL_PrepareLightTexture());
	
	gl.DeleteTextures(NUM_FLARES, flaretexnames);
	memset(flaretexnames, 0, sizeof(flaretexnames));
}

void OGL_ResetLumpTexData()
{
	gl.DeleteTextures(mynumlumps, lumptexnames);
	gl.DeleteTextures(mynumlumps, lumptexnames2);
	memset(lumptexnames, 0, sizeof(GLuint)*mynumlumps);
	memset(lumptexnames2, 0, sizeof(GLuint)*mynumlumps);
	memset(lumptexsizes, 0, sizeof(texsize_t)*mynumlumps);

	// Free the raw lumps book keeping table.
	free(rawlumps);
	rawlumps = 0;
	numrawlumps = 0;
}

void OGL_ClearTextureMemory(void)
{
	OGL_TexReset();
	OGL_ResetLumpTexData();
}

void OGL_UpdateGamma(void)
{
	LoadPalette();
	OGL_ClearTextureMemory();
}

// Binds the texture if necessary.
void OGL_BindTexture(GLuint texname)
{
	if(curtex != texname) 
	{
		gl.Bind(texname);
		curtex = texname;
	}
}

static void PalToRGB(byte *palidx, byte *rgb)
{
	int i;
	for(i=0; i<3; i++) 
	{
		*(rgb+i) = gammatable[usegamma][*(palidx+i)];
	}
}

void PalIdxToRGB(byte *pal, int idx, byte *rgb)
{
	int c;
	for(c=0; c<3; c++) // Red, green and blue.
	{
		rgb[c] = gammatable[usegamma][pal[idx*3 + c]];
	}
}

// This ignores the alpha channel!
// This certainly isn't a very optimized algorithm.
static void ScaleBuffer8(byte *in, int inWidth, int inHeight,
						 byte *out, int outWidth, int outHeight, int alpha)
{
	byte	*palette = W_CacheLumpName("playpal", PU_CACHE);
	int		x, y, outIdx, inIdx, inX, inY;
	float	xscale, yscale;		// outcoord * scale = incoord
	byte	*inAlpha = 0, *outAlpha = 0;

	if(inWidth == outWidth && inHeight == outHeight)
	{
		// Just copy it.
		memcpy(out, in, inWidth * inHeight * (alpha? 2 : 1));
		return;
	}

	if(alpha)
	{
		inAlpha = in + inWidth * inHeight;
		outAlpha = out + outWidth * outHeight;
	}

	xscale = inWidth / (float) outWidth;
	yscale = inHeight / (float) outHeight;
	for(y=0; y<outHeight; y++)
		for(x=0; x<outWidth; x++)
		{
			outIdx = y*outWidth + x;
			// Do we need to box filter?
			if(xscale > 1 || yscale > 1)
			{
				int i, k, inBox[4], r, g, b, a, count;
				byte inRGB[3];
				inBox[0] = (int) x * xscale;
				inBox[1] = (int) y * yscale;
				inBox[2] = (int) (x+1) * xscale;
				inBox[3] = (int) (y+1) * yscale;
				if(inBox[2] > inBox[0]) inBox[2]--;
				if(inBox[3] > inBox[1]) inBox[3]--;

				// Now we know what pixels to check out.
				for(count=r=g=b=a=0, i=inBox[0]; i<=inBox[2]; i++)
					for(k=inBox[1]; k<=inBox[3]; k++)
					{
						memcpy(inRGB, palette + 3*in[i + k*inWidth], 3);
						r += inRGB[0];
						g += inRGB[1];
						b += inRGB[2];
						if(alpha) a += inAlpha[i + k*inWidth];
						count++;
					}
				// Calculate the average.
				r /= count;
				g /= count;
				b /= count;
				out[outIdx] = pal15to8[RGB15(r>>3, g>>3, b>>3)];
				if(alpha)
				{
					outAlpha[outIdx] = a/count;
				}
			}
			else
			{
				inX = (int) x * xscale;
				inY = (int) y * yscale;
				inIdx = inY*inWidth + inX;
				out[outIdx] = in[inIdx];
				if(alpha)
				{
					outAlpha[outIdx] = inAlpha[inIdx];
				}
			}
		}
}

// Takes maxTexSize and the ratio limit into consideration.
/*void CalcSingleTextureSize(int width, int height, int *outWidth, int *outHeight)
{
	width = FindNextPower2(width);
	height = FindNextPower2(height);

	if(width > maxTexSize) width = maxTexSize;
	if(height > maxTexSize) height = maxTexSize;
	
	if(ratioLimit)
	{
		if(width > height)
		{
			if(height < width/ratioLimit)
				height = width/ratioLimit;
		}
		else
		{
			if(width < height/ratioLimit)
				width = height/ratioLimit;
		}
	}

	*outWidth = width;
	*outHeight = height;
}*/

static void BuildMipmaps8(byte *data, int width, int height, int alpha)
{
	int		i, levelWidth, levelHeight;	// width and height at the top level
	byte	*buffer;
	
	// We can't have a wider texture than maxTexSize.
	// FIXME: Height is not checked; it shouldn't matter, not with 
	// the original artwork.
	if(width > maxTexSize)
		levelWidth = maxTexSize;
	else
		levelWidth = FindNextPower2(width);
	levelHeight = FindNextPower2(height);

	buffer = _alloca(levelWidth * levelHeight * (alpha? 2 : 1));
	for(i=0; levelWidth || levelHeight; i++)
	{
		if(!levelWidth) levelWidth = 1;
		if(!levelHeight) levelHeight = 1;

		ScaleBuffer8(data, width, height, buffer, levelWidth, levelHeight, alpha);

		// Upload it.
		/*glTexImage2D(GL_TEXTURE_2D, i, GL_COLOR_INDEX8_EXT, levelWidth, 
			levelHeight, 0, GL_COLOR_INDEX,	GL_UNSIGNED_BYTE, buffer);*/

		if( gl.TexImage(alpha? DGL_COLOR_INDEX_8_PLUS_A8 : DGL_COLOR_INDEX_8, 
			levelWidth, levelHeight, i, buffer) != DGL_OK)
		{
			I_Error("BuildMipmaps8: TexImage failed (%i x %i), alpha:%i\n",
				levelWidth, levelHeight, alpha);
		}			

		// Move on.
		levelWidth >>= 1;
		levelHeight >>= 1;
	}
}

// Returns the name of the texture. 
// The texture parameters will NOT be set here.
// 'data' contains indices to the playpal. If 'alphachannel' is true,
// 'data' also contains the alpha values (after the indices).
static GLuint LoadTexture(byte *data, int width, int height, boolean alphachannel, 
						  boolean nomipmap)
{
	int		format = (alphachannel? DGL_COLOR_INDEX_8_PLUS_A8 : DGL_COLOR_INDEX_8);
	DGLuint	texName;

	// Generate a new texture name and bind it.
	texName = gl.NewTexture();

	if(nomipmap)
	{
		if(width != FindNextPower2(width))
			I_Error("LoadTexture: width not a power of 2.\n");
		if(height != FindNextPower2(height))
			I_Error("LoadTexture: height not a power of 2.\n");

		/*glTexImage2D(GL_TEXTURE_2D, 0, GL_COLOR_INDEX8_EXT, width, height, 
			0, GL_COLOR_INDEX, GL_UNSIGNED_BYTE, data);*/

		if( gl.TexImage(format, width, height, 0, data) != DGL_OK)
			I_Error("LoadTexture: TexImage failed.\n");
	}
	else
	{
		// We'll use a simple mipmap reduction algorithm of our own.
		BuildMipmaps8(data, width, height, alphachannel);
	}

/*	}
	else // Use true color textures.
	{
		int i, colorComps = alphachannel? 4 : 3;
		int numPixels = width * height;
		byte *buffer = _alloca(numPixels * colorComps);
		byte *palette = W_CacheLumpNum(pallump=W_GetNumForName("PLAYPAL"), PU_CACHE);
		byte *pixel;

		// The conversion is very straightforward.
		for(i=0, pixel=buffer; i<numPixels; i++, pixel+=colorComps)
		{
			PalIdxToRGB(palette, data[i], pixel);
			if(alphachannel)
			{
				pixel[3] = data[numPixels+i];
			}
		}

		// Load the texture.
		if(nomipmap)
		{
			if(width != FindNextPower2(width))
				I_Error("LoadTexture: width not a power of 2.\n");
			if(height != FindNextPower2(height))
				I_Error("LoadTexture: height not a power of 2.\n");

			glTexImage2D(GL_TEXTURE_2D, 0, colorComps, width, height, 0, 
				alphachannel? GL_RGBA : GL_RGB, GL_UNSIGNED_BYTE, buffer);
		}
		else
		{
			// Write ScaleBuffer32 and BuildMipmaps32 if you don't 
			// want to use this.
			gluBuild2DMipmaps(GL_TEXTURE_2D, colorComps, width, height, 
				alphachannel? GL_RGBA : GL_RGB, GL_UNSIGNED_BYTE, buffer);
		}
	}*/
	return texName;
}

unsigned int OGL_BindTexFlat(int lump)
{
	DGLuint name;
	byte *flatptr;
	
	if(lumpinfo[lump].size < 4096)
	{
		//ST_Message("OGL_BindTexFlat: lumpinfo.size %d\n", lumpinfo[lump].size);
		OGL_BindTexture(0);	
		return 0;
	}
	// Get a pointer to the texture data.
	flatptr = W_CacheLumpNum(lump, PU_STATIC);
	
	// Load the texture.
	name = LoadTexture(flatptr, 64, 64, false, test3dfx);
	
	// Set the parameters.
/*	glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, glmode[mipmapping]);
	glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);*/

	gl.TexParameter(DGL_MIN_FILTER, glmode[mipmapping]);
	gl.TexParameter(DGL_MAG_FILTER, DGL_LINEAR);
	
	Z_ChangeTag(flatptr, PU_CACHE);
	// The name of the texture is returned.
	return name;
}

// Returns the OpenGL name of the texture.
unsigned int OGL_PrepareFlat(int idx)
{
	if(idx < 0 || idx > numflats-1) idx = 0;

	idx = flattranslation[idx];
	if(!flattexnames[idx])
	{
		// The flat isn't yet bound with OpenGL.
		flattexnames[idx] = OGL_BindTexFlat(firstflat+idx);
	}
	texw = texh = 64;
	texmask = 0;
	return flattexnames[idx];
}

void OGL_SetFlat(int idx)
{
	gl.Bind(OGL_PrepareFlat(idx));
}

// Return GL_RGB or GL_RGBA.
/*static int DrawRealPatch(byte *rgbflat, byte *rgbaflat, byte *palette, int texwidth,
						 int texheight, patch_t *patch, boolean maskZero,
						 unsigned char *transtable)*/

// If maskZero is true, the buffer must have room for the alpha values.
// Returns true if the buffer really has alpha information.
static int DrawRealPatch(byte *buffer, byte *palette, int texwidth,
						 int texheight, patch_t *patch, boolean maskZero,
						 unsigned char *transtable)
{
	int count;
	int col;
	column_t *column;
	//byte *desttop1, *desttop2;
	byte *destTop, *destAlphaTop = NULL;
	byte *dest1, *dest2;
	byte *source;
	int w, i;

	col = 0;
	/*desttop1 = rgbflat;
	desttop2 = rgbaflat;*/
	destTop = buffer;
	destAlphaTop = buffer + texwidth*texheight;
	w = SHORT(patch->width);
	for(; col < w; col++, /*desttop1+=3, desttop2+=4*/ destTop++, destAlphaTop++)
	{
		column = (column_t *)((byte *)patch+LONG(patch->columnofs[col]));
		// Step through the posts in a column
		while(column->topdelta != 0xff)
		{
			source = (byte *)column+3;
			
			/*if(rgbflat) dest1 = desttop1 + column->topdelta*texwidth*3;
			if(rgbaflat) dest2 = desttop2 + column->topdelta*texwidth*4;*/

			dest1 = destTop + column->topdelta*texwidth;
			dest2 = destAlphaTop + column->topdelta*texwidth;

			count = column->length;
			while(count--)
			{
				unsigned char palidx = *source++;
				// Do we need to make a translation?
				if(transtable) palidx = transtable[palidx];
				/*if(rgbflat)	
				{
					if(!maskZero || palidx)	PalIdxToRGB(palette, palidx, dest1);
					dest1 += texwidth*3;
				}
				if(rgbaflat)
				{
					if(!maskZero || palidx)
					{
						PalIdxToRGB(palette, palidx, dest2);
						*(dest2+3) = 0xff;	// This pixel is not clear.
					}
					dest2 += texwidth*4;
				}*/
				if(!maskZero || palidx)
					*dest1 = palidx;
				dest1 += texwidth; // One row down.
				
				if(maskZero)
					*dest2 = palidx? 0xff : 0;
				else
					*dest2 = 0xff;
				dest2 += texwidth; // One row down.
			}
			column = (column_t *)((byte *)column+column->length+4);
		}
	}
	// Scan through the RGBA buffer and check for sub-0xff alpha.
	source = buffer + texwidth*texheight;
	for(i=0; i<texwidth*texheight; i++)
		if(source[i] < 0xff) return true; // Has alpha data.
	return false; // Doesn't have alpha data.
}

// Returns the OpenGL texture name.
unsigned int OGL_PrepareTexture(int idx)
{
	if(idx == 0)//|| idx > numtextures-1)
	{
		// No texture?
		//glBindTexture(GL_TEXTURE_2D, 0);
		//curtex = 0;
		texw = 1;
		texh = 1;
		texmask = 0;
		return 0;
	}
	idx = texturetranslation[idx];
	if(!texnames[idx])
	{
		// The texture must be given to OpenGL.
		int i, k;//, textype;
		boolean alphaChannel = false;
		texture_t *tex = textures[idx];
		byte *palette = W_CacheLumpNum(pallump, PU_CACHE);
		//byte *rgbflat = _alloca(3*tex->width*tex->height);
		byte *buffer = _alloca(2*tex->width*tex->height);
		byte *colptr;
	
		if(tex->patchcount > 1)
		{
			//textype = GL_RGB;
			// The texture will have no alpha information.
			for(i=0; i<tex->width; i++)
			{
				byte *pixel = buffer + i;
				colptr = R_GetColumn(idx,i);
				for(k=0; k<tex->height; k++, pixel+=tex->width)
				{
					*pixel = colptr[k];
				}
					//PalIdxToRGB(palette, colptr[k], rgbflat+(k*tex->width+i)*3);
			}
		}
		else 
		{
			// This texture has only only one patch. It might be masked.
			//byte *rgbaflat = _alloca(4*tex->width*tex->height);
			//memset(rgbaflat, 0, 4*tex->width*tex->height);

			memset(buffer, 0, 2*tex->width*tex->height);

			/*textype = DrawRealPatch(rgbflat, rgbaflat, palette,	tex->width, tex->height,
				W_CacheLumpNum(tex->patches[0].patch, PU_CACHE), false, 0);
			if(textype == GL_RGBA) rgbflat = rgbaflat;*/

			alphaChannel = DrawRealPatch(buffer, palette, tex->width, tex->height,
				W_CacheLumpNum(tex->patches[0].patch, PU_CACHE), true, 0);
		}
						
		// Generate and bind the texture.
/*		glGenTextures(1, texnames+idx);
		glBindTexture(GL_TEXTURE_2D, texnames[idx]);
		gluBuild2DMipmaps(GL_TEXTURE_2D, (textype==GL_RGB)?3:4, tex->width, 
			tex->height, textype, GL_UNSIGNED_BYTE, rgbflat);*/

		texnames[idx] = LoadTexture(buffer, tex->width, tex->height, alphaChannel, false);

		//glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
		
		/*glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, glmode[mipmapping]);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);*/
		gl.TexParameter(DGL_MIN_FILTER, glmode[mipmapping]);
		gl.TexParameter(DGL_MAG_FILTER, DGL_LINEAR);

		/*if(textype == GL_RGBA)*/
		texmasked[idx] = alphaChannel==true;	// Yes it is.
	}
	texw = textures[idx]->width;
	texh = textures[idx]->height;
	texmask = texmasked[idx];
	return texnames[idx];
}

void OGL_SetTexture(int idx)
{
	gl.Bind(OGL_PrepareTexture(idx));
}

int LineAverageRGB(byte *imgdata, int width, int height, int line, byte *rgb, byte *palette)
{
	byte	*start = imgdata + width*line;
	byte	*alphaStart = start + width*height;
	int		i, c, count = 0;
	int		integerRGB[3] = {0,0,0};
	byte	col[3];

	for(i=0; i<width; i++)
	{
		//byte *pixpos = start + i;
		//if(*(pixpos+3) > 0)	
		// Not transparent?
		if(alphaStart[i] > 0)
		{
			count++;
			PalIdxToRGB(palette, start[i], col);
			for(c=0; c<3; c++) integerRGB[c] += col[c];
		}
	}
	// All transparent? Sorry...
	if(!count) return 0;

	// We're going to make it!
	for(c=0; c<3; c++) rgb[c] = integerRGB[c]/count;
	return 1;	// Successful.
}

// The imgdata must have alpha info, too.
void ImageAverageRGB(byte *imgdata, int width, int height, byte *rgb, byte *palette)
{
	int	i, c, integerRGB[3] = {0,0,0}, count = 0;

	for(i=0; i<height; i++)
	{
		if(LineAverageRGB(imgdata, width, height, i, rgb, palette))
		{
			count++;
			for(c=0; c<3; c++) integerRGB[c] += rgb[c];
		}
	}
	if(count)	// If there were pixels...
		for(c=0; c<3; c++) 
			rgb[c] = integerRGB[c]/count;
}

// Return a skycol_t for texidx. 
// More housekeeping...
skycol_t *OGL_GetSkyColor(int texidx)
{
	int			i;
	skycol_t	*skycol;

	if(texidx < 0) return NULL;

	// Try to find a skytop color for this.
	for(i=0; i<num_skytop_colors; i++)
		if(skytop_colors[i].texidx == texidx)
			return skytop_colors + i;

	// There was no skycol for the specified texidx!
	skytop_colors = realloc(skytop_colors, sizeof(skycol_t) * ++num_skytop_colors);
	skycol = skytop_colors + num_skytop_colors-1;
	memset(skycol, 0, sizeof(*skycol));
	skycol->texidx = texidx;
	return skycol;
}

// Fills the empty pixels with reasonable color indices.
// This gets rid of the black outlines.
// Not a very efficient algorithm...
static void ColorOutlines(byte *buffer, int width, int height)
{
	int		numpels = width*height;
	byte	*ptr;
	int		i, k, a, b;

	for(k=0; k<height; k++)
		for(i=0; i<width; i++)
			// Solid pixels spread around...
			if(buffer[numpels + i + k*width])
			{
				for(b=-1; b<=1; b++)
					for(a=-1; a<=1; a++)
					{
						// First check that the pixel is OK.
						if(!a && !b || i+a < 0 || k+b < 0 || i+a >= width || k+b >= height) 
							continue;

						ptr = buffer + i+a + (k+b)*width;
						if(!ptr[numpels]) // An alpha pixel?
							*ptr = buffer[i + k*width];
					}
			}
}

// Sky textures are ALWAYS 256 pixels wide!!
unsigned int OGL_PrepareSky(int idx, boolean zeroMask)
{
	if(idx > numtextures-1) return 0;

	if(idx != texturetranslation[idx])
		I_Error("Skytex: %d, translated: %d\n", idx, texturetranslation[idx]);
	
	idx = texturetranslation[idx];
	
	if(!texnames[idx])
	{
		skycol_t	*skycol = OGL_GetSkyColor(idx);
		int			i, k;
		byte		*palette = W_CacheLumpNum(pallump, PU_CACHE);
		texture_t	*tex = textures[idx];
		int			h2 = FindNextPower2(tex->height);
		int			numpels = 256*h2;
		byte		*imgdata = _alloca(2*numpels), *colptr;

		// Choose the right format.
		/*if(zeroMask)
		{
			textype = GL_RGBA;
			imgdata = _alloca(4*numpels);
			memset(imgdata, 0, 4*numpels);
		}
		else
		{
			textype = GL_RGB;
			imgdata = _alloca(3*numpels);
			memset(imgdata, 0, 3*numpels);
		}*/
		memset(imgdata, 0, 2*numpels);
		if(tex->patchcount > 1)
		{
			for(i=0; i<tex->width; i++)
			{
				colptr = R_GetColumn(idx, i);
				for(k=0; k<tex->height; k++)
				{
					//if(textype == GL_RGB)
					if(!zeroMask)
						//PalIdxToRGB(palette, colptr[k], imgdata+(k*256+i)*3);
						imgdata[k*256 + i] = colptr[k];
					//else if(textype == GL_RGBA && colptr[k])
					else if(colptr[k])
					{
						//byte *imgpos = imgdata+(k*256+i)*4;
						byte *imgpos = imgdata+(k*256+i);
						//PalIdxToRGB(palette, colptr[k], imgpos);
						*imgpos = colptr[k];
						*(imgpos+numpels) = 0xff;	// Not transparent, this pixel.
						//*(imgpos+3) = 0xff;	// Not transparent, this pixel.
					}
				}
			}
		}
		else 
		{
			// This texture has only only one patch.
			/*if(textype == GL_RGB)
				DrawRealPatch(imgdata, 0, palette, 256, tex->height, 
					W_CacheLumpNum(tex->patches[0].patch, PU_CACHE), false, 0);
			else if(textype == GL_RGBA) // Mask out zeros.
				DrawRealPatch(0, imgdata, palette, 256, tex->height,
					W_CacheLumpNum(tex->patches[0].patch, PU_CACHE), true, 0);*/
			DrawRealPatch(imgdata, palette, 256, h2, 
				W_CacheLumpNum(tex->patches[0].patch, PU_CACHE),
				zeroMask, 0);
		}
		//if(textype == GL_RGBA)	// For masked data, calculate the alpha-fill color.
		
		if(zeroMask)				// For masked data, calculate the alpha-fill color.
		{
			ColorOutlines(imgdata, 256, h2);
			/*byte rgbAverage[3], *alphaStart = imgdata + numpels;
			//boolean in

			ImageAverageRGB(imgdata, 256, h2, rgbAverage, palette);
			// Fill all the transparent places.
			for(i=0; i<256*tex->height; i++)
			{
				//byte *pixel = imgdata + i;
				if(!alphaStart[i]) //memcpy(pixel, rgbAverage, 3);
					imgdata[i] = 
			}*/
		}
		// Calculate the topline RGB for sky top fadeouts.
		//LineAverageRGB(imgdata, (textype==GL_RGB)?3:4, 256, 0, skycol->rgb, palette);
		LineAverageRGB(imgdata, 256, h2, 0, skycol->rgb, palette);

		// Generate and bind the texture.
		/*glGenTextures(1, texnames+idx);
		glBindTexture(GL_TEXTURE_2D, texnames[idx]);
		if(test3dfx)
		{
			glTexImage2D(GL_TEXTURE_2D, 0, (textype==GL_RGB)?3:4, 256, h2, 0, GL_RGB, 
				GL_UNSIGNED_BYTE, imgdata);
		}
		else
			gluBuild2DMipmaps(GL_TEXTURE_2D, (textype==GL_RGB)?3:4, 256, h2, textype, 
				GL_UNSIGNED_BYTE, imgdata);*/

		texnames[idx] = LoadTexture(imgdata, 256, h2, zeroMask, false);

//		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
/*		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, glmode[mipmapping]);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);*/
		gl.TexParameter(DGL_MIN_FILTER, glmode[mipmapping]);
		gl.TexParameter(DGL_MAG_FILTER, DGL_LINEAR);

		// Do we have a masked texture?
		if(zeroMask)//textype == GL_RGBA) 
			texmasked[idx] = 1;
		else
			texmasked[idx] = 0;
	}
	texw = textures[idx]->width;
	texh = textures[idx]->height;
	texmask = texmasked[idx];
	return texnames[idx];
}

void OGL_GetSkyTopColor(int texidx, byte *rgb)
{
	skycol_t *skycol = OGL_GetSkyColor(texidx);

	if(!skycol) return;
	memcpy(rgb, skycol->rgb, 3);
}

transspr_t *OGL_NewTranslatedSprite(int pnum, unsigned char *table)
{
	transspr_t *news;
	
	transsprites = realloc(transsprites, sizeof(transspr_t)* ++numtranssprites);
	news = transsprites + numtranssprites-1;
	news->patch = pnum;
	news->tex = 0;
	news->table = table;
	return news;
}

transspr_t *OGL_GetTranslatedSprite(int pnum, unsigned char *table)
{
	int		i;

	for(i=0; i<numtranssprites; i++)
		if(transsprites[i].patch == pnum && transsprites[i].table == table)
			return transsprites + i;
	return 0;
}

// Data must have alpha info.
void averageColorIdx(rgbcol_t *sprcol, byte *data, int w, int h, byte *palette)
{
	int				i;
	unsigned int	r, g, b, count;
	byte			*alphaStart = data + w*h, rgb[3];

	// First clear them.
	r = g = b = count = 0;
	for(i=0; i<w*h; i++)
	{
		//byte *pixel = rgbaData + 4*i;
		//if(pixel[3]) // Skip pixels with no visible color (alpha==0).
		if(alphaStart[i])
		{
			count++;
			PalIdxToRGB(palette, data[i], rgb);
			r += rgb[0];
			g += rgb[1];
			b += rgb[2];
		}
	}
	sprcol->rgb[0] = r/count;
	sprcol->rgb[1] = g/count;
	sprcol->rgb[2] = b/count;
}

void averageColorRGBA(rgbcol_t *sprcol, byte *rgbaData, int w, int h)
{
	int				i;
	unsigned int	r, g, b, count;

	// First clear them.
	r = g = b = count = 0;
	for(i=0; i<w*h; i++)
	{
		byte *pixel = rgbaData + 4*i;
		if(pixel[3]) // Skip pixels with no visible color (alpha==0).
		{
			count++;
			r += pixel[0];
			g += pixel[1];
			b += pixel[2];
		}
	}
	sprcol->rgb[0] = r/count;
	sprcol->rgb[1] = g/count;
	sprcol->rgb[2] = b/count;
}

unsigned int OGL_PrepareTranslatedSprite(int pnum, unsigned char *table)
{
	transspr_t	*tspr = OGL_GetTranslatedSprite(pnum, table);

	if(!tspr)
	{
		// There's no name for this patch, load it in.
		patch_t *patch = W_CacheLumpNum(firstspritelump+pnum, PU_CACHE);
		int p2width = FindNextPower2(patch->width), 
			p2height = OGL_ValidTexHeight2(patch->width, patch->height);// FindNextPower2(patch->height);
		int bufsize = 2 * /*4*/ p2width*p2height;
		byte *palette = W_CacheLumpNum(pallump, PU_CACHE);
		//byte *rgbaflat = _alloca(flatsize);
		byte *buffer = _alloca(bufsize);
		
		//printf( "orig: %d x %d => %d x %d\n", patch->width, patch->height, p2width, p2height);

		//memset(rgbaflat, 0, flatsize);
		memset(buffer, 0, bufsize);

		/*DrawRealPatch(0, rgbaflat, W_CacheLumpNum(pallump,PU_CACHE),
			p2width, p2height, patch, false, table);*/
		DrawRealPatch(buffer, W_CacheLumpNum(pallump, PU_CACHE), p2width, p2height,
			patch, false, table);

		averageColorIdx(spritecolors+pnum, buffer, p2width, p2height, palette);

		tspr = OGL_NewTranslatedSprite(pnum, table);

		// Generate and bind the texture.
		/*glGenTextures(1, &tspr->tex);
		glBindTexture(GL_TEXTURE_2D, tspr->tex);
		gluBuild2DMipmaps(GL_TEXTURE_2D, 4, p2width, p2height, GL_RGBA, 
			GL_UNSIGNED_BYTE, rgbaflat);*/

		tspr->tex = LoadTexture(buffer, p2width, p2height, true, false);

/*		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, glmode[mipmapping]);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);*/

		gl.TexParameter(DGL_MIN_FILTER, glmode[mipmapping]);
		gl.TexParameter(DGL_MAG_FILTER, DGL_LINEAR);
		gl.TexParameter(DGL_WRAP_S, DGL_CLAMP);
		gl.TexParameter(DGL_WRAP_T, DGL_CLAMP);

		// Also here the height is updated.
		spriteheights[pnum] = patch->height;
	}
	return tspr->tex;
}

unsigned int OGL_PrepareSprite(int pnum)
{
//	if(pnum > numspritelumps-1) return 0;

	if(!spritenames[pnum])
	{
		// There's no name for this patch, load it in.
		byte	*palette = W_CacheLumpNum(pallump, PU_CACHE);
		patch_t *patch = W_CacheLumpNum(firstspritelump+pnum, PU_CACHE);
		int		p2width = FindNextPower2(patch->width), 
				p2height = OGL_ValidTexHeight2(patch->width, patch->height);// FindNextPower2(patch->height);
		/*int flatsize = 4*p2width*p2height;
		byte *rgbaflat = _alloca(flatsize);*/
		int		bufsize;
		byte	*buffer;

		// This'll allow us to the resize the image when its loaded.
		if(p2width > maxTexSize) p2width = patch->width;

		//p2width = FindNextPower2(patch->width);
		//p2height = FindNextPower2(patch->height);
		bufsize = 2 * p2width * p2height;
		buffer = _alloca(bufsize);
		//printf( "orig: %d x %d => %d x %d\n", patch->width, patch->height, p2width, p2height);

		//memset(rgbaflat, 0, flatsize);
		memset(buffer, 0, bufsize);
		/*DrawRealPatch(0, rgbaflat, W_CacheLumpNum(pallump,PU_CACHE),
			p2width, p2height, patch, false, 0);*/
		DrawRealPatch(buffer, W_CacheLumpNum(pallump, PU_CACHE), p2width, p2height,
			patch, false, 0);

		ColorOutlines(buffer, p2width, p2height);
		
		averageColorIdx(spritecolors+pnum, buffer, p2width, p2height, palette);

		// Generate and bind the texture.
		/*glGenTextures(1, spritenames+pnum);
		glBindTexture(GL_TEXTURE_2D, spritenames[pnum]);
		gluBuild2DMipmaps(GL_TEXTURE_2D, 4, p2width, p2height, GL_RGBA, 
			GL_UNSIGNED_BYTE, rgbaflat);*/

		spritenames[pnum] = LoadTexture(buffer, p2width, p2height, true, false);

/*		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, glmode[mipmapping]);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);*/

		gl.TexParameter(DGL_MIN_FILTER, glmode[mipmapping]);
		gl.TexParameter(DGL_MAG_FILTER, DGL_LINEAR);
		gl.TexParameter(DGL_WRAP_S, DGL_CLAMP);
		gl.TexParameter(DGL_WRAP_T, DGL_CLAMP);

		spriteheights[pnum] = patch->height;
	}
	return spritenames[pnum];
}

void OGL_GetSpriteColor(int pnum, unsigned char *rgb)
{
	if(pnum > numspritelumps-1) return;
	memcpy(rgb, spritecolors+pnum, 3);
}
 
void OGL_SetSprite(int pnum)
{
	OGL_BindTexture(OGL_PrepareSprite(pnum));
}

void OGL_SetTranslatedSprite(int pnum, unsigned char *trans)
{
	OGL_BindTexture(OGL_PrepareTranslatedSprite(pnum, trans));
}

void OGL_NewRawLump(int lump)
{
	rawlumps = realloc(rawlumps, sizeof(int) * ++numrawlumps);
	rawlumps[numrawlumps-1] = lump;
}

GLuint OGL_GetOtherPart(int lump)
{
	return lumptexnames2[lump];
}

// Part is either 1 or 2. Part 0 means only the left side is loaded.
// No splittex is created in that case. Once a raw image is loaded
// as part 0 it must be deleted before the other part is loaded at the
// next loading.
void OGL_SetRawImage(int lump, int part)
{
	if(part < 0 || part > 2 || lump > numlumps-1) return;	// Check the part.

	if(!lumptexnames[lump])
	{
		// Load the raw texture data (320x200).
		// We'll create two textures (256x256 and 64x256).
		byte *raw = W_CacheLumpNum(lump, PU_CACHE);
		byte *dat1 = _alloca(256*256);	
		byte *dat2 = _alloca(64*256);
/*		byte *dat1 = _alloca(3*256*256);	// Let's hope there's enough stack!
		byte *dat2 = _alloca(3*64*256);
*/
		byte *palette = W_CacheLumpNum(pallump, PU_CACHE);
		int i,k;

		memset(dat1, 0, 256*256);	// Why must this be done??
		for(k=0; k<200; k++)
			for(i=0; i<256; i++)
			{
				// We can't go over the end.
				if(k*320+i > lumpinfo[lump].size-1) break;
				// Part one.
				//PalIdxToRGB(palette, *(raw+(k*320+i)), dat1+3*(k*256+i));
				dat1[k*256 + i] = raw[k*320 + i];
				if(i<64 && part) // Part two?
					//PalIdxToRGB(palette, *(raw+(k*320+i+256)), dat2+3*(k*64+i));
					dat2[k*64 + i] = raw[k*320 + i + 256];
			}

		// Do a special fill for textures with h<200 (part 0).
		if(/*lumpinfo[lump].size/320 < 200 &&*/ !part)
		{
			int lines = lumpinfo[lump].size/320;
			// Copy the missing data from the beginning.
			//memcpy(dat1+lines*256*3, dat1, 3*256*(256-lines));
			memcpy(dat1+lines*256, dat1, 256*(256-lines));
		}
			
		// Generate and load the textures.
		/*glGenTextures(1, lumptexnames+lump);
		glBindTexture(GL_TEXTURE_2D, lumptexnames[lump]);
		glTexImage2D(GL_TEXTURE_2D, 0, 3, 256, 256, 0, GL_RGB, GL_UNSIGNED_BYTE, dat1);*/

		lumptexnames[lump] = LoadTexture(dat1, 256, 256, false, true);
/*		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, linearRaw? GL_LINEAR : GL_NEAREST);*/

		gl.TexParameter(DGL_MIN_FILTER, DGL_NEAREST);
		gl.TexParameter(DGL_MAG_FILTER, linearRaw? DGL_LINEAR : DGL_NEAREST);
		
		if(part)
		{
			// And the other part.
			/*glGenTextures(1, lumptexnames2+lump);
			glBindTexture(GL_TEXTURE_2D, lumptexnames2[lump]);
			glTexImage2D(GL_TEXTURE_2D, 0, 3, 64, 256, 0, GL_RGB, GL_UNSIGNED_BYTE, dat2);*/
			lumptexnames2[lump] = LoadTexture(dat2, 64, 256, false, true);
/*			glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, linearRaw? GL_LINEAR : GL_NEAREST);*/
		
			gl.TexParameter(DGL_MIN_FILTER, DGL_NEAREST);
			gl.TexParameter(DGL_MAG_FILTER, linearRaw? DGL_LINEAR : DGL_NEAREST);

			//glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			gl.TexParameter(DGL_WRAP_S, DGL_CLAMP);

			// Add it to the list.
			OGL_NewRawLump(lump);
		}

		lumptexsizes[lump].w = 256;
		lumptexsizes[lump].w2 = 64;
		lumptexsizes[lump].h = 200;
	}
	// Bind the correct part.
	if(part <= 1) gl.Bind(lumptexnames[lump]);
	if(part == 2) gl.Bind(lumptexnames2[lump]);
	// We don't track the current texture with raw images.
	curtex = 0;
}

// Notices ratiolimit.
int OGL_ValidTexHeight2(int width, int height)
{
	int p2w, p2h = FindNextPower2(height);
	int minheight;

	if(!ratioLimit) return p2h;

	p2w = FindNextPower2(width);
	// Do we have to notice that the texture will be split?
	if(p2w > maxTexSize)
	{
		int part2width = FindNextPower2(width - maxTexSize);
		int height1 = maxTexSize/ratioLimit,
			height2 = part2width/ratioLimit;
		minheight = height1 > height2? height1 : height2;
	}
	else minheight = p2w/ratioLimit;
	// Do the test.
	if(minheight > p2h) return minheight;
	return p2h;
}

void OGL_SetPatch(int lump)	// No mipmaps are generated.
{
	if(lump > numlumps-1) return;
	if(!lumptexnames[lump])
	{
		// Load the patch.
		patch_t	*patch = W_CacheLumpNum(lump, PU_CACHE);
		int		p2width = FindNextPower2(patch->width), 
				p2height = OGL_ValidTexHeight2(patch->width, patch->height);
		int		numpels = p2width*p2height;
		byte	*buffer = _alloca(2*numpels);
		int		alphaChannel;
		
		memset(buffer, 0, 2*numpels);
		
		alphaChannel = DrawRealPatch(buffer, W_CacheLumpNum(pallump, PU_CACHE),
			p2width, p2height, patch, false, 0);

		ColorOutlines(buffer, p2width, p2height);

		// See if we have to split the patch into two parts.
		if(p2width > maxTexSize) // Nothing about the height...
		{
			// Notice: This is only vertical splitting, p2height 
			// applies to both parts. 

			// The width of the first part is maxTexSize.
			int part2width = FindNextPower2(patch->width - maxTexSize);
			byte *tempbuff = _alloca(2*maxTexSize*p2height);
			
			if(part2width > maxTexSize)
				I_Error("OGL_SetPatch: Too wide texture (really: %i, pow2: %i).\n", patch->width, p2width);
			
			// We'll use a temporary buffer for doing to splitting.
			// First, part one.
			pixBlt(buffer, p2width, p2height, tempbuff, maxTexSize, p2height, 
				alphaChannel, 0, 0, 0, 0, maxTexSize, p2height);

			// Generate a texture.
			lumptexnames[lump] = LoadTexture(tempbuff, maxTexSize, p2height, alphaChannel, true);

			/*glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);*/

			gl.TexParameter(DGL_MIN_FILTER, DGL_NEAREST);
			gl.TexParameter(DGL_MAG_FILTER, DGL_LINEAR);
			gl.TexParameter(DGL_WRAP_S, DGL_CLAMP);
			gl.TexParameter(DGL_WRAP_T, DGL_CLAMP);

			// Then part two.
			pixBlt(buffer, p2width, p2height, tempbuff, part2width, p2height,
				alphaChannel, maxTexSize, 0, 0, 0, part2width, p2height);

			// Generate a texture.
			lumptexnames2[lump] = LoadTexture(tempbuff, part2width, p2height, alphaChannel, true);

/*			glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);*/

			gl.TexParameter(DGL_MIN_FILTER, DGL_NEAREST);
			gl.TexParameter(DGL_MAG_FILTER, DGL_LINEAR);
			gl.TexParameter(DGL_WRAP_S, DGL_CLAMP);
			gl.TexParameter(DGL_WRAP_T, DGL_CLAMP);

			OGL_BindTexture(lumptexnames[lump]);

			lumptexsizes[lump].w = maxTexSize;
			lumptexsizes[lump].w2 = patch->width - maxTexSize;
		}
		else // We can use the normal one-part method.
		{
			// Generate a texture.
			lumptexnames[lump] = LoadTexture(buffer, p2width, p2height, alphaChannel, true);

			/*glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);*/
			//glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			//glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

			gl.TexParameter(DGL_MIN_FILTER, DGL_NEAREST);
			gl.TexParameter(DGL_MAG_FILTER, DGL_LINEAR);

			lumptexsizes[lump].w = patch->width;
			lumptexsizes[lump].w2 = 0;
		}
		// The rest of the size information.
		lumptexsizes[lump].h = patch->height;
		lumptexsizes[lump].offx = -patch->leftoffset;
		lumptexsizes[lump].offy = -patch->topoffset;
	}
	else
	{
		OGL_BindTexture(lumptexnames[lump]);
	}
	curtex = lumptexnames[lump];
}

void OGL_SetNoTexture()
{
	gl.Bind(0);
	curtex = 0;
}

GLuint OGL_PrepareLightTexture()
{
	if(!dltexname)
	{
		// We need to generate the texture, I see.
		byte *image = W_CacheLumpName("DLIGHT", PU_CACHE);
		if(!image)
			I_Error("OGL_SetLightTexture: no dlight texture.\n");
		// The dynamic light map is a 64x64 grayscale 8-bit image.
		/*glGenTextures(1, &dltexname);
		glBindTexture(GL_TEXTURE_2D, dltexname);*/

		dltexname = gl.NewTexture();

		/*glTexImage2D(GL_TEXTURE_2D, 0, 1, 64, 64, 0, GL_LUMINANCE, 
			GL_UNSIGNED_BYTE, image);*/
		
		gl.TexImage(DGL_LUMINANCE, 64, 64, 0, image);

		gl.TexParameter(DGL_MIN_FILTER, DGL_NEAREST);
		gl.TexParameter(DGL_MAG_FILTER, DGL_LINEAR);
		gl.TexParameter(DGL_WRAP_S, DGL_CLAMP);
		gl.TexParameter(DGL_WRAP_T, DGL_CLAMP);

/*		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);*/
	}
	// Set the info.
	texw = texh = 64;
	texmask = 0;
	return dltexname;
}

GLuint OGL_PrepareFlareTexture(int flare)
{
	int		w, h;

	// There are three flare textures.
	if(flare < 0 || flare >= NUM_FLARES) return 0;

	// What size?
	w = h = flare==2? 128 : 64;

	if(!flaretexnames[flare])
	{
		byte *image = W_CacheLumpName(flare==0? "FLARE" : flare==1? "BRFLARE" : "BIGFLARE", PU_CACHE);
		if(!image)
			I_Error("OGL_PrepareFlareTexture: flare texture %i not found!\n", flare);

		/*glGenTextures(1, flaretexnames + flare);
		glBindTexture(GL_TEXTURE_2D, flaretexnames[flare]);*/

		flaretexnames[flare] = gl.NewTexture();

/*		glTexImage2D(GL_TEXTURE_2D, 0, 1, w, h, 0, GL_LUMINANCE,
			GL_UNSIGNED_BYTE, image);*/

		gl.TexImage(DGL_LUMINANCE, w, h, 0, image);

/*		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);*/

		gl.TexParameter(DGL_MIN_FILTER, DGL_NEAREST);
		gl.TexParameter(DGL_MAG_FILTER, DGL_LINEAR);
		gl.TexParameter(DGL_WRAP_S, DGL_CLAMP);
		gl.TexParameter(DGL_WRAP_T, DGL_CLAMP);
	}
	texmask = 0;
	texw = w;
	texh = h;
	return flaretexnames[flare];
}

int OGL_GetLumpTexWidth(int lump)
{
	return lumptexsizes[lump].w;
}

int OGL_GetLumpTexHeight(int lump)
{
	return lumptexsizes[lump].h;
}

// Updates the textures, flats and sprites (gameTex) or the user 
// interface textures (patches and raw screens).
void OGL_SetTextureParams(int minMode, int magMode, int gameTex, int uiTex)
{
	int	i;
	
	if(gameTex)
	{
		// Textures.
		for(i=0; i<numtextures; i++)
			if(texnames[i])	// Is the texture loaded?
			{
				gl.Bind(texnames[i]);
				/*glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minMode);
				glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magMode);*/
				gl.TexParameter(DGL_MIN_FILTER, minMode);
				gl.TexParameter(DGL_MAG_FILTER, magMode);
			}
		// Flats.
		for(i=0; i<numflats; i++)
			if(flattexnames[i]) // Is the texture loaded?
			{
				gl.Bind(flattexnames[i]);
				/*glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minMode);
				glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magMode);*/
				gl.TexParameter(DGL_MIN_FILTER, minMode);
				gl.TexParameter(DGL_MAG_FILTER, magMode);
			}
		// Sprites.
		for(i=0; i<numspritelumps; i++)
			if(spritenames[i])
			{
				gl.Bind(spritenames[i]);
				/*glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minMode);
				glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magMode);*/
				gl.TexParameter(DGL_MIN_FILTER, minMode);
				gl.TexParameter(DGL_MAG_FILTER, magMode);
			}
		// Translated sprites.
		for(i=0; i<numtranssprites; i++)
		{	
			gl.Bind(transsprites[i].tex);
			/*glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minMode);
			glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magMode);*/
			gl.TexParameter(DGL_MIN_FILTER, minMode);
			gl.TexParameter(DGL_MAG_FILTER, magMode);
		}
	}
	if(uiTex)
	{
		for(i=0; i<numlumps; i++)
		{
			if(lumptexnames[i])
			{
				gl.Bind(lumptexnames[i]);
				/*glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minMode);
				glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magMode);*/
				gl.TexParameter(DGL_MIN_FILTER, minMode);
				gl.TexParameter(DGL_MAG_FILTER, magMode);
			}
			if(lumptexnames2[i])
			{
				gl.Bind(lumptexnames[i]);
				/*glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minMode);
				glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magMode);*/
				gl.TexParameter(DGL_MIN_FILTER, minMode);
				gl.TexParameter(DGL_MAG_FILTER, magMode);
			}
		}
	}
}

void OGL_UpdateTexParams(int mipmode)
{
	mipmapping = mipmode;
	OGL_SetTextureParams(glmode[mipmode], DGL_LINEAR, true, false);
}

void OGL_LowRes()
{
	// Set everything as low as they go.
	OGL_SetTextureParams(DGL_NEAREST, DGL_NEAREST, true, true);
//	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
}

// Updates the raw screen smoothing (linear magnification).
// This is the main reason for having the rawlumps table.
void OGL_UpdateRawScreenParams(int smoothing)
{
	int		i;
	int		glmode = (smoothing)? DGL_LINEAR : DGL_NEAREST;

	linearRaw = smoothing;
	for(i=0; i<numrawlumps; i++)
	{
		// First part 1.
		gl.Bind(lumptexnames[rawlumps[i]]);
		//glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, glmode);
		gl.TexParameter(DGL_MAG_FILTER, glmode);
		// Then part 2.
		gl.Bind(lumptexnames2[rawlumps[i]]);
		//glTexParametereteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, glmode);
		gl.TexParameter(DGL_MAG_FILTER, glmode);
	}
}

void OGL_TextureFilterMode(int target, int parm)
{
	if(target == DD_TEXTURES) OGL_UpdateTexParams(parm);
	if(target == DD_RAWSCREENS) OGL_UpdateRawScreenParams(parm);
}

// Deletes a texture. Only for textures (not for sprites, flats, etc.).
void OGL_DeleteTexture(int texidx)
{
	if(texidx < 0) return;

	if(texnames[texidx])
	{
		gl.DeleteTextures(1, texnames+texidx);
		texnames[texidx] = 0;
	}
}

unsigned int OGL_GetTextureName(int texidx)
{
	return texnames[texidx];
}