
//**************************************************************************
//**
//** SV_FRAME.C
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "dd_def.h"
#include "dd_net.h"
#include "i_net.h"
#include "i_win32.h"
#include "i_timer.h"
#include "sv_def.h"
#include "console.h"

// MACROS ------------------------------------------------------------------

#define FIXED8_8(x)					(((x) & 0xffff00) >> 8)

#define MAX_MOBJ_LEN			23
#define MAX_PLAYER_LEN			20

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

void Sv_RefreshClient(int playernum);

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

extern boolean net_timerefresh;

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static int last_transmit = -1;

// CODE --------------------------------------------------------------------

// Sends all the relevant information to each client.
void Sv_TransmitFrame(void)
{
	int		i;

	// Naturally single-player games or clients don't transmit anything.
	if(!netgame || !server) return;			

	// If the server is fast, we can get here several times 
	// per tic. We don't want that.
	if(gametic == last_transmit) return;
	last_transmit = gametic;

	// Scan for missiles to send.
	Sv_ScanForMissiles();

	for(i=1; i<MAXPLAYERS; i++)
		if(clients[i].ready && clients[i].updateCount > 0)
		{
			// Don't allow packets to pile up.
			if(I_CheckSendQueue(i)) Sv_RefreshClient(i);
		}
}

void Sv_PackMobj(byte flags, mobj_t *mobj)
{
	byte temp;

	// Data isn't sent for nothing.
	if(!mobj->frame) flags &= ~MDP_FRAME;
	if(!mobj->momx && !mobj->momy) flags &= ~MDP_MOVEMENT_XY;
	if(!mobj->momz) flags &= ~MDP_MOVEMENT_Z;
	if(!(mobj->ddflags & DDMF_PACK_MASK)) flags &= ~MDP_EXTRA_FLAGS;
	if(mobj->floorclip) flags |= MDP_EXTRA_FLAGS;

	// Combine flags and spritenum into the first word.
	Msg_WriteShort((flags<<8) | (mobj->sprite & MDP_SPRITE_MASK));

	// Position is sent always.
	if(flags & MDP_ACCURATE)
	{
		Msg_WriteShort(mobj->x >> 16);
		Msg_WriteByte((mobj->x >> 8) & 0xff);
		Msg_WriteShort(mobj->y >> 16);
		Msg_WriteByte((mobj->y >> 8) & 0xff);
		if(flags & MDP_POSITION_Z)
		{
			Msg_WriteShort(mobj->z >> 16);
			Msg_WriteByte((mobj->z >> 8) & 0xff);
		}
	}
	else
	{
		Msg_WriteShort(mobj->x >> 16);
		Msg_WriteShort(mobj->y >> 16);
		if(flags & MDP_POSITION_Z)
		{
			Msg_WriteShort(mobj->z >> 16);
		}
	}
	if(flags & MDP_MOVEMENT_XY)
	{
		Msg_WriteShort(FIXED8_8(mobj->momx));
		Msg_WriteShort(FIXED8_8(mobj->momy));
	}
	if(flags & MDP_MOVEMENT_Z)
	{
		Msg_WriteShort(FIXED8_8(mobj->momz));
	}
	if(flags & MDP_DIRECTION)
	{
		if(flags & MDP_ACCURATE)
		{
			Msg_WriteShort(mobj->angle >> 16);
		}
		else
		{
			Msg_WriteByte(mobj->angle >> 24);
		}
	}
	if(flags & MDP_EXTRA_FLAGS)
	{
		unsigned short fl = 0;
		temp = 0;
		if(mobj->ddflags & DDMF_SHADOW) fl |= MDPE_SHADOW;
		if(mobj->ddflags & DDMF_ALTSHADOW) fl |= MDPE_ALTSHADOW;
		if(mobj->ddflags & DDMF_BRIGHTSHADOW) fl |= MDPE_BRIGHTSHADOW;
		//if(mobj->ddflags & DDMF_FITTOP) fl |= MDPE_FITTOP;
		if(mobj->floorclip) fl |= MDPE_FLOORCLIP;
		if(mobj->ddflags & DDMF_NOFITBOTTOM) fl |= MDPE_NOFITBOTTOM;
		if(mobj->ddflags & DDMF_VIEWALIGN) fl |= MDPE_VIEWALIGN;
		if(mobj->ddflags & DDMF_SOLID) fl |= MDPE_SOLID;
		fl |= (mobj->ddflags & DDMF_LIGHTOFFSET) >> DDMF_LIGHTOFFSETSHIFT
			<< MDPE_LIGHTOFFSETSHIFT;
		fl |= (mobj->ddflags & DDMF_RESERVED) >> DDMF_CLASSTRSHIFT
			<< MDPE_CLASSSHIFT;
		fl |= (mobj->ddflags & DDMF_TRANSLATION) >> DDMF_TRANSSHIFT
			<< MDPE_TRANSSHIFT;
		temp |= (mobj->ddflags & DDMF_LIGHTSCALE) >> DDMF_LIGHTSCALESHIFT 
			<< MDPE_LIGHTSCALESHIFT;
		if(temp) fl |= MDPE_EXTENDED;
		Msg_WriteShort(fl);
		if(temp) Msg_WriteByte(temp);
		// Floorclip.
		if(mobj->floorclip) Msg_WriteByte(mobj->floorclip >> 14);
	}
	if(flags & MDP_FRAME)
	{
		temp = mobj->frame & MDP_FRAME_MASK;
		if(mobj->frame & FF_FULLBRIGHT) temp |= MDPF_FULLBRIGHT;
		Msg_WriteByte(temp);
		if(mobj->frame & FF_FULLBRIGHT)
		{
			// For fullbright objects, it's also necessary to sent
			// the radius of the mobj. 
			Msg_WriteByte(mobj->radius >> FRACBITS);
		}
	}
}

void Sv_PackCamera(int plrNum)
{
	ddplayer_t *plr = &players[plrNum];
	boolean fixangles = (plr->flags & DDPF_FIXANGLES) != 0;

	// First, the filter + flags.
	Msg_WriteByte((fixangles? CDP_FIX_ANGLES : 0) 
		| (plr->filter? CDP_FILTER : 0)
		| (plr->extralight? CDP_EXTRA_LIGHT : 0));
	Sv_PackMobj(MDP_ACCURATE | MDP_POSITION_Z | MDP_MOVEMENT 
		| (fixangles? MDP_DIRECTION : 0), plr->mo);
	// Look direction?
	if(fixangles) Msg_WriteShort(plr->lookdir/110 * DDMAXSHORT); 
	Msg_WriteShort(FIXED8_8(plr->viewheight));	// View height.
	// Filter?
	if(plr->filter) Msg_WriteLong(plr->filter);
	// Extra light?
	if(plr->extralight) Msg_WriteByte(plr->extralight);

	// Sending once will have to suffice.
	plr->flags &= ~DDPF_FIXANGLES;
}

void Sv_PackPlayer(int plrnum)
{
	ddplayer_t *plr = &players[plrnum];

	// The first byte contains the player number and some flags.
	Msg_WriteByte(plrnum);	// No flags yet.
	Sv_PackMobj(MDP_ALL, plr->mo);
}

// Send all necessary data to the client.
void Sv_RefreshClient(int plrNum)
{
	ddplayer_t	*player = &players[plrNum], *other;
	client_t	*cl = &clients[plrNum];
	int			i, count; 
	sector_t	*sec, *plrsec;
	mobj_t		*iter;
	byte		flags;
	spritedef_t *sprdef;
	spriteframe_t *sprframe;
	int			valid = ++validcount;
	byte		*headerPtr;

	int			refresh_started_at = I_GetRealTime();

	// If this hits zero, the client will stop having updates.
	cl->updateCount--;

	if(!player->mo) 
	{
		// Interesting... we don't know where the client is.
		return;		
	}

	player->mo->validcount = valid;

	Msg_Begin(psv_frame);
	Msg_WriteByte(gametic);	// Frame time, lowest byte of gametic.

	// Send the player itself as a camera.
	Sv_PackCamera(plrNum);

	// The recipient needs to know how many players there will be.
	headerPtr = netbuffer.cursor;
	Msg_WriteByte(0);

	// Other players are always sent. This is to allow such things
	// as player markers on the automap and client side weaponfire
	// simulation.
	for(i=0, other=players, count=0; i<MAXPLAYERS; i++, other++)
	{
		if(!other->ingame || other == player || !other->mo) 
			continue;	
		other->mo->validcount = valid;
		Sv_PackPlayer(i);
		count++;
	}
	// Set the count.
	*headerPtr = count;

	// Send all mobjs in view.
	//CON_Printf( "begin: %i / ", ticcount);
	plrsec = player->mo->subsector->sector;

	headerPtr = netbuffer.cursor;
	Msg_WriteByte(0);

	for(i=0, sec=(sector_t*)sectors, count=0; i<numsectors; i++, 
		((byte*)sec)+=SECTSIZE)
	{
		if(!P_CheckReject(plrsec, sec)) continue;
		for(iter=sec->thinglist; iter; iter=iter->snext)
		{
			// Invisible mobjs aren't sent.
			if(iter->ddflags & (DDMF_LOCAL | DDMF_MISSILE | DDMF_DONTDRAW)) continue;
			if(iter->validcount == valid) continue;

			// Check view frustum (too far, behind, off to the sides?).
			if(!P_CheckFrustum(plrNum, iter)) continue;
												
			// LOS check.
			if(!P_CheckSight(player->mo, iter)) continue;
			
			// This mobj will be sent.
			iter->validcount = valid;
			flags = MDP_MOVEMENT | MDP_FRAME | MDP_POSITION_Z | MDP_EXTRA_FLAGS;
			sprdef = &sprites[iter->sprite];
			sprframe = &sprdef->spriteframes[ iter->frame & FF_FRAMEMASK ];
			if(sprframe->rotate) flags |= MDP_DIRECTION;
			if(iter->z == sec->floorheight) flags &= ~MDP_POSITION_Z;
			
			// Add the mobj into the buffer.
			Sv_PackMobj(flags, iter);
			count++;

			if(count == MAX_CLMOBJS) 
			{
				// Enough already, stop this.
				i = numsectors;
				break;
			}
		}
	}
	*headerPtr = count;

	// Send the missile spawn requests suitable for this player.
	Sv_PackMissiles(plrNum);

	// Send the frame packet.
	Net_SendPacket(plrNum, SPF_FRAME);

	if(net_timerefresh)
	{
		CON_Printf( "refresh %i: %i ms\n", plrNum, I_GetRealTime()-refresh_started_at);
	}
}
