
//**************************************************************************
//**
//** SV_MAIN.C
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "dd_def.h"
#include "dd_net.h"
#include "sv_def.h"
#include "i_net.h"
#include "i_win32.h"
#include "i_timer.h"
#include "console.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

// PRIVATE DATA DEFINITIONS ------------------------------------------------

// CODE --------------------------------------------------------------------

// For single-player / servers.
void Sv_FixLocalAngles()
{
	ddplayer_t *pl = &players[0];

	// This is not for clients.
	if(consoleplayer || isDedicated) return;
	
	if(pl->flags & DDPF_FIXANGLES)
	{
		pl->flags &= ~DDPF_FIXANGLES;
		pl->clAngle = pl->mo->angle;
		pl->clLookDir = pl->lookdir;
	}
}

void Sv_HandlePacket(void)
{
	int i, mask, from = netbuffer.player;
	ddplayer_t *pl = &players[from];
	client_t *sender = &clients[from];
	playerinfo_packet_t info;

	switch(netbuffer.msg.type)
	{
	case pcl_hello:
		CON_Printf( "Sv_HandlePacket: Hello from client %i.\n", from);
		// The client requests a handshake.
		if(!pl->ingame && !sender->handshake)
		{
			// This'll be true until the client says it's ready.
			sender->handshake = true;

			// The player is now in the game.
			players[from].ingame = true;

			// Tell the game about this.
			gx.NetPlayerEvent(from, DDPE_ARRIVAL, 0);

			// Send the handshake packets.
			Sv_Handshake(from);
		}
		break;

	case pkt_ok:
		if(sender->handshake)
		{
			// Not any more. The client has acknowledged the handshake
			// and sends its regards.
			sender->handshake = false;
			sender->ready = true;
			Sv_SendText(from, SV_CONSOLE_FLAGS, SV_WELCOME_STRING"\n");
		}
		break;

	case pkt_chat:
		// The first byte contains the sender, but we don't need it.
		Msg_ReadByte();	
		// Is the message for us?
		mask = Msg_ReadShort();		
		if(mask & 1) 
		{
			D_ShowChatMessage();
			gx.NetPlayerEvent(netbuffer.msg.data[0], DDPE_CHAT_MESSAGE,
				netbuffer.msg.data+3);
		}
		// Set the cursor to the end of the message.
		netbuffer.cursor = netbuffer.msg.data + netbuffer.length;
		// Servers relay chat messages to all the recipients.
		for(i=1; i<MAXPLAYERS; i++)
			if(players[i].ingame && mask & (1<<i) && i != from)
				Net_SendPacket(i, SPF_RELIABLE);	
		break;

	case pkt_player_info:
		Msg_Read(&info, sizeof(info));
		CON_FPrintf(CBLF_TRANSMIT | SV_CONSOLE_FLAGS,
			"%s renamed to %s.\n", sender->name, info.name);
		strcpy(sender->name, info.name);
		D_SendPacket(DDSP_RELIABLE | DDSP_ALL_PLAYERS, pkt_player_info,
			&info, sizeof(info));
		break;
	}
}

//========================================================================
//
// Sv_GetPackets
//
// Server's packet handler.
//
//========================================================================

void Sv_GetPackets (void)
{
	int             netconsole;
	int				start, num, i;
	client_t		*sender;

	while(Net_GetPacket())
	{
		switch(netbuffer.msg.type)
		{
		case pkt_ticcmd:
			// Determine who sent this packet.
			netconsole = netbuffer.player;
			if(netconsole < 0 || netconsole >= MAXPLAYERS) continue; 
			
			sender = &clients[netconsole];

			// If the client isn't ready, don't accept any cmds.
			if(!clients[netconsole].ready) continue;

			// Now we know this client is alive, update the frame send count.
			// Clients will only be refreshed if their updateCount is greater
			// than zero.
			sender->updateCount = UPDATECOUNT;
			
			// Add the tics into the client's ticcmd buffer, if there is room.
			// If the buffer overflows, the rest of the cmds will be forgotten.
			num = Msg_ReadByte();
			if(sender->numtics + num > BACKUPTICS) 
				num = BACKUPTICS - sender->numtics;
			start = sender->firsttic + sender->numtics;
			// Increase the counter.
			sender->numtics += num;
			// Copy as many as fits.
			for(i=start; num > 0; num--, i++)
			{
				if(i >= BACKUPTICS) i -= BACKUPTICS;
				Msg_Read(&sender->ticcmds[TICCMD_IDX(i)], TICCMD_SIZE);
			}
			break;

		case pkt_ping:
			Net_PingResponse();
			break;

		case pcl_hello:
		case pkt_ok:
		case pkt_chat:
		case pkt_player_info:
			Sv_HandlePacket();
			break;

		default:
			if(netbuffer.msg.type >= pkt_game_marker)
			{
				// A client has sent a game specific packet.
				gx.HandlePacket(netbuffer.player,
					netbuffer.msg.type, netbuffer.msg.data,
					netbuffer.length);
			}
		}
	}
}

// Called by jtNet.
void Sv_PlayerArrives(int newnode)
{
	int		i;
	char	**nameList, newname[PLAYERNAMELEN];
	unsigned int nodeids[MAXPLAYERS];

	if(!netgame)
	{
		ST_Message( "Sv_NetPlayerArrives called when not in netgame!\n");
		return;
	}
	if(!server) return;

	nameList = jtNetGetStringList(JTNET_PLAYER_NAME_LIST, NULL);
	strncpy(newname, nameList[newnode], PLAYERNAMELEN-1);
	newname[PLAYERNAMELEN-1] = 0;	
	CON_Printf( "%s has arrived.\n", newname);

	I_NetUpdateNodes();
	
	jtNetGetPlayerIDs(nodeids);

	// We need to find the new player a client entry.
	for(i=1; i<MAXPLAYERS; i++)
		if(!clients[i].connected)
		{
			// This'll do.
			clients[i].connected = true;
			clients[i].ready = false;
			clients[i].jtNetNode = newnode;
			clients[i].nodeID = nodeids[newnode];
			strcpy(clients[i].name, newname);

			CON_Printf( "Sv_PlayerArrives: %s designated plr %i, node %i (%x)\n",
				clients[i].name, i, newnode, nodeids[newnode]);

			// In order to get in the game, the client must first
			// shake hands. It'll request this by sending a Hello packet.
			clients[i].handshake = false;
			clients[i].updateCount = UPDATECOUNT;
			break;
		}
}

// Called by jtNet.
void Sv_PlayerLeaves(jtnetplayer_t *plrdata)
{
	int i, pNumber = -1;

	if(!server) return;

	// First let's find out the player number.
	for(i=0; i<MAXPLAYERS; i++)
		if(clients[i].nodeID == plrdata->id)
		{
			pNumber = i;
			break;
		}
	if(pNumber == -1) return; // Bogus?

	// Print something in the console.
	CON_Printf( "Sv_PlayerLeaves: %s (plr %i) has left.\n", plrdata->name, pNumber);
	players[pNumber].ingame = false;
	clients[pNumber].connected = false;
	clients[pNumber].ready = false;
	clients[pNumber].updateCount = 0;
	clients[pNumber].handshake = false;
	I_NetUpdateNodes();
	
	// Inform the DLL about this.
	gx.NetPlayerEvent(pNumber, DDPE_EXIT, NULL);

	// Inform other clients about this.
	Msg_Begin(psv_player_exit);
	Msg_WriteByte(pNumber);
	Net_SendPacket(NSP_BROADCAST, SPF_RELIABLE);
}

// The player will be sent the introductory handshake packets.
void Sv_Handshake(int playernum)
{
	int						i;
	handshake_packet_t		shake;
	playerinfo_packet_t		info;

	CON_Printf( "Sv_Handshake: shaking hands with player %i.\n", playernum);

	shake.version = SV_VERSION;
	shake.yourConsole = playernum;
	shake.playerMask = 0;
	for(i=0; i<MAXPLAYERS; i++)
		if(clients[i].connected)
			shake.playerMask |= 1<<i;
	D_SendPacket(playernum | DDSP_RELIABLE, psv_handshake, &shake, 
		sizeof(shake));

	// Propagate client information.
	for(i=0; i<MAXPLAYERS; i++)
	{
		if(clients[i].connected)
		{
			info.console = i;
			strcpy(info.name, clients[i].name);
			D_SendPacket(playernum | DDSP_RELIABLE, pkt_player_info, 
				&info, sizeof(info));
		}
		// Send the new player's info to other players.
		if(i != 0 && i != playernum && clients[i].connected)
		{
			info.console = playernum;
			strcpy(info.name, clients[playernum].name);
			D_SendPacket(i | DDSP_RELIABLE, pkt_player_info, 
				&info, sizeof(info));
		}
	}

	// The game DLL wants to shake hands as well?
 	gx.NetWorldEvent(DDWE_HANDSHAKE, playernum, NULL);

	Sv_SendWorldUpdate(playernum);
}

void Sv_StartNetGame()
{
	int		i;

	// Reset all the counters and other data.
	// We must make it appear as if the program had just been started.
	for(i=0; i<MAXPLAYERS; i++)
	{
		players[i].ingame = false;
		clients[i].connected = false;
		clients[i].ready = false;
		clients[i].numtics = 0;
		clients[i].firsttic = 0;
		clients[i].updateCount = UPDATECOUNT;
		clients[i].fov = 90;
		strcpy(clients[i].name, "");
		resend_start = 0;
		resend_count = 0;
	}
	gametic = maketic = 0;
	firstNetUpdate = firstRunTic = true;

	// The server is always player number zero.
	consoleplayer = displayplayer = 0; 

	netgame = true;
	server = true;
	
	if(!isDedicated)
	{
		players[consoleplayer].ingame = true;
		clients[consoleplayer].connected = true;
		clients[consoleplayer].ready = true;
		strcpy(clients[consoleplayer].name, playerName);	
	}
}

void Sv_SendText(int to, int con_flags, char *text)
{
	Msg_Begin(psv_console_text);
	Msg_WriteLong(con_flags & ~CBLF_TRANSMIT);
	Msg_Write(text, strlen(text)+1);
	Net_SendPacket(to, SPF_RELIABLE);
}

void Sv_Ticker(void)
{
	// Update master every 3 minutes.
	if(masterAware && !(systics % (180*TICRATE)))
	{
		jtNetUpdateServerInfo();
	}
}